#include "dri3.h"

#include <xcb/xcb.h>
#include <xcb/xcbext.h>

#include "base/logging.h"
#include "ui/gfx/x/xproto_internal.h"

namespace x11 {

Dri3::Dri3(x11::Connection* connection,
    const x11::QueryExtensionReply& info)
    : connection_(connection), info_(info) {}

Future<Dri3::QueryVersionReply>
Dri3::QueryVersion(
    const Dri3::QueryVersionRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& major_version = request.major_version;
  auto& minor_version = request.minor_version;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 0;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // major_version
  Write(&major_version, &buf);

  // minor_version
  Write(&minor_version, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Dri3::QueryVersionReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Dri3::QueryVersionReply>
detail::ReadReply<Dri3::QueryVersionReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Dri3::QueryVersionReply>();

  auto& sequence = (*reply).sequence;
  auto& major_version = (*reply).major_version;
  auto& minor_version = (*reply).minor_version;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // major_version
  Read(&major_version, &buf);

  // minor_version
  Read(&minor_version, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Dri3::OpenReply>
Dri3::Open(
    const Dri3::OpenRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& drawable = request.drawable;
  auto& provider = request.provider;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 1;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // provider
  Write(&provider, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Dri3::OpenReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Dri3::OpenReply>
detail::ReadReply<Dri3::OpenReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Dri3::OpenReply>();

  auto& nfd = (*reply).nfd;
  auto& sequence = (*reply).sequence;
  auto& device_fd = (*reply).device_fd;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // nfd
  Read(&nfd, &buf);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // device_fd
  Read(&device_fd, &buf);

  // pad0
  Pad(&buf, 24);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
Dri3::PixmapFromBuffer(
    const Dri3::PixmapFromBufferRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& pixmap = request.pixmap;
  auto& drawable = request.drawable;
  auto& size = request.size;
  auto& width = request.width;
  auto& height = request.height;
  auto& stride = request.stride;
  auto& depth = request.depth;
  auto& bpp = request.bpp;
  auto& pixmap_fd = request.pixmap_fd;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 2;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // pixmap
  Write(&pixmap, &buf);

  // drawable
  Write(&drawable, &buf);

  // size
  Write(&size, &buf);

  // width
  Write(&width, &buf);

  // height
  Write(&height, &buf);

  // stride
  Write(&stride, &buf);

  // depth
  Write(&depth, &buf);

  // bpp
  Write(&bpp, &buf);

  // pixmap_fd
  Write(&pixmap_fd, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<Dri3::BufferFromPixmapReply>
Dri3::BufferFromPixmap(
    const Dri3::BufferFromPixmapRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& pixmap = request.pixmap;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 3;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // pixmap
  Write(&pixmap, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Dri3::BufferFromPixmapReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Dri3::BufferFromPixmapReply>
detail::ReadReply<Dri3::BufferFromPixmapReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Dri3::BufferFromPixmapReply>();

  auto& nfd = (*reply).nfd;
  auto& sequence = (*reply).sequence;
  auto& size = (*reply).size;
  auto& width = (*reply).width;
  auto& height = (*reply).height;
  auto& stride = (*reply).stride;
  auto& depth = (*reply).depth;
  auto& bpp = (*reply).bpp;
  auto& pixmap_fd = (*reply).pixmap_fd;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // nfd
  Read(&nfd, &buf);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // size
  Read(&size, &buf);

  // width
  Read(&width, &buf);

  // height
  Read(&height, &buf);

  // stride
  Read(&stride, &buf);

  // depth
  Read(&depth, &buf);

  // bpp
  Read(&bpp, &buf);

  // pixmap_fd
  Read(&pixmap_fd, &buf);

  // pad0
  Pad(&buf, 12);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
Dri3::FenceFromFD(
    const Dri3::FenceFromFDRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& drawable = request.drawable;
  auto& fence = request.fence;
  auto& initially_triggered = request.initially_triggered;
  auto& fence_fd = request.fence_fd;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 4;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // fence
  Write(&fence, &buf);

  // initially_triggered
  Write(&initially_triggered, &buf);

  // pad0
  Pad(&buf, 3);

  // fence_fd
  Write(&fence_fd, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<Dri3::FDFromFenceReply>
Dri3::FDFromFence(
    const Dri3::FDFromFenceRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& drawable = request.drawable;
  auto& fence = request.fence;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 5;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // fence
  Write(&fence, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Dri3::FDFromFenceReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Dri3::FDFromFenceReply>
detail::ReadReply<Dri3::FDFromFenceReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Dri3::FDFromFenceReply>();

  auto& nfd = (*reply).nfd;
  auto& sequence = (*reply).sequence;
  auto& fence_fd = (*reply).fence_fd;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // nfd
  Read(&nfd, &buf);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // fence_fd
  Read(&fence_fd, &buf);

  // pad0
  Pad(&buf, 24);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Dri3::GetSupportedModifiersReply>
Dri3::GetSupportedModifiers(
    const Dri3::GetSupportedModifiersRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& window = request.window;
  auto& depth = request.depth;
  auto& bpp = request.bpp;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 6;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  // depth
  Write(&depth, &buf);

  // bpp
  Write(&bpp, &buf);

  // pad0
  Pad(&buf, 2);

  Align(&buf, 4);

  return x11::SendRequest<Dri3::GetSupportedModifiersReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Dri3::GetSupportedModifiersReply>
detail::ReadReply<Dri3::GetSupportedModifiersReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Dri3::GetSupportedModifiersReply>();

  auto& sequence = (*reply).sequence;
  uint32_t num_window_modifiers{};
  uint32_t num_screen_modifiers{};
  auto& window_modifiers = (*reply).window_modifiers;
  size_t window_modifiers_len = window_modifiers.size();
  auto& screen_modifiers = (*reply).screen_modifiers;
  size_t screen_modifiers_len = screen_modifiers.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // num_window_modifiers
  Read(&num_window_modifiers, &buf);

  // num_screen_modifiers
  Read(&num_screen_modifiers, &buf);

  // pad1
  Pad(&buf, 16);

  // window_modifiers
  window_modifiers.resize(num_window_modifiers);
  for (auto& window_modifiers_elem : window_modifiers) {
    // window_modifiers_elem
    Read(&window_modifiers_elem, &buf);

  }

  // screen_modifiers
  screen_modifiers.resize(num_screen_modifiers);
  for (auto& screen_modifiers_elem : screen_modifiers) {
    // screen_modifiers_elem
    Read(&screen_modifiers_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
Dri3::PixmapFromBuffers(
    const Dri3::PixmapFromBuffersRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& pixmap = request.pixmap;
  auto& window = request.window;
  uint8_t num_buffers{};
  auto& width = request.width;
  auto& height = request.height;
  auto& stride0 = request.stride0;
  auto& offset0 = request.offset0;
  auto& stride1 = request.stride1;
  auto& offset1 = request.offset1;
  auto& stride2 = request.stride2;
  auto& offset2 = request.offset2;
  auto& stride3 = request.stride3;
  auto& offset3 = request.offset3;
  auto& depth = request.depth;
  auto& bpp = request.bpp;
  auto& modifier = request.modifier;
  auto& buffers = request.buffers;
  size_t buffers_len = buffers.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 7;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // pixmap
  Write(&pixmap, &buf);

  // window
  Write(&window, &buf);

  // num_buffers
  num_buffers = buffers.size();
  Write(&num_buffers, &buf);

  // pad0
  Pad(&buf, 3);

  // width
  Write(&width, &buf);

  // height
  Write(&height, &buf);

  // stride0
  Write(&stride0, &buf);

  // offset0
  Write(&offset0, &buf);

  // stride1
  Write(&stride1, &buf);

  // offset1
  Write(&offset1, &buf);

  // stride2
  Write(&stride2, &buf);

  // offset2
  Write(&offset2, &buf);

  // stride3
  Write(&stride3, &buf);

  // offset3
  Write(&offset3, &buf);

  // depth
  Write(&depth, &buf);

  // bpp
  Write(&bpp, &buf);

  // pad1
  Pad(&buf, 2);

  // modifier
  Write(&modifier, &buf);

  // buffers
  DCHECK_EQ(static_cast<size_t>(num_buffers), buffers.size());
  for (auto& buffers_elem : buffers) {
    // buffers_elem
    NOTIMPLEMENTED();

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<Dri3::BuffersFromPixmapReply>
Dri3::BuffersFromPixmap(
    const Dri3::BuffersFromPixmapRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& pixmap = request.pixmap;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 8;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // pixmap
  Write(&pixmap, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Dri3::BuffersFromPixmapReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Dri3::BuffersFromPixmapReply>
detail::ReadReply<Dri3::BuffersFromPixmapReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Dri3::BuffersFromPixmapReply>();

  uint8_t nfd{};
  auto& sequence = (*reply).sequence;
  auto& width = (*reply).width;
  auto& height = (*reply).height;
  auto& modifier = (*reply).modifier;
  auto& depth = (*reply).depth;
  auto& bpp = (*reply).bpp;
  auto& strides = (*reply).strides;
  size_t strides_len = strides.size();
  auto& offsets = (*reply).offsets;
  size_t offsets_len = offsets.size();
  auto& buffers = (*reply).buffers;
  size_t buffers_len = buffers.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // nfd
  Read(&nfd, &buf);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // width
  Read(&width, &buf);

  // height
  Read(&height, &buf);

  // pad0
  Pad(&buf, 4);

  // modifier
  Read(&modifier, &buf);

  // depth
  Read(&depth, &buf);

  // bpp
  Read(&bpp, &buf);

  // pad1
  Pad(&buf, 6);

  // strides
  strides.resize(nfd);
  for (auto& strides_elem : strides) {
    // strides_elem
    Read(&strides_elem, &buf);

  }

  // offsets
  offsets.resize(nfd);
  for (auto& offsets_elem : offsets) {
    // offsets_elem
    Read(&offsets_elem, &buf);

  }

  // buffers
  buffers.resize(nfd);
  for (auto& buffers_elem : buffers) {
    // buffers_elem
    NOTIMPLEMENTED();

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

}  // namespace x11
