#include "glx.h"

#include <xcb/xcb.h>
#include <xcb/xcbext.h>

#include "base/logging.h"
#include "ui/gfx/x/xproto_internal.h"

namespace x11 {

Glx::Glx(x11::Connection* connection,
    const x11::QueryExtensionReply& info)
    : connection_(connection), info_(info) {}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<Glx::PbufferClobberEvent>(
    Glx::PbufferClobberEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& event_type = (*event_).event_type;
  auto& draw_type = (*event_).draw_type;
  auto& drawable = (*event_).drawable;
  auto& b_mask = (*event_).b_mask;
  auto& aux_buffer = (*event_).aux_buffer;
  auto& x = (*event_).x;
  auto& y = (*event_).y;
  auto& width = (*event_).width;
  auto& height = (*event_).height;
  auto& count = (*event_).count;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // event_type
  Read(&event_type, &buf);

  // draw_type
  Read(&draw_type, &buf);

  // drawable
  Read(&drawable, &buf);

  // b_mask
  Read(&b_mask, &buf);

  // aux_buffer
  Read(&aux_buffer, &buf);

  // x
  Read(&x, &buf);

  // y
  Read(&y, &buf);

  // width
  Read(&width, &buf);

  // height
  Read(&height, &buf);

  // count
  Read(&count, &buf);

  // pad1
  Pad(&buf, 4);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<Glx::BufferSwapCompleteEvent>(
    Glx::BufferSwapCompleteEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& event_type = (*event_).event_type;
  auto& drawable = (*event_).drawable;
  auto& ust_hi = (*event_).ust_hi;
  auto& ust_lo = (*event_).ust_lo;
  auto& msc_hi = (*event_).msc_hi;
  auto& msc_lo = (*event_).msc_lo;
  auto& sbc = (*event_).sbc;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // event_type
  Read(&event_type, &buf);

  // pad1
  Pad(&buf, 2);

  // drawable
  Read(&drawable, &buf);

  // ust_hi
  Read(&ust_hi, &buf);

  // ust_lo
  Read(&ust_lo, &buf);

  // msc_hi
  Read(&msc_hi, &buf);

  // msc_lo
  Read(&msc_lo, &buf);

  // sbc
  Read(&sbc, &buf);

}

Future<void>
Glx::Render(
    const Glx::RenderRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& data = request.data;
  size_t data_len = data.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 1;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // data
  DCHECK_EQ(static_cast<size_t>(data_len), data.size());
  for (auto& data_elem : data) {
    // data_elem
    Write(&data_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::RenderLarge(
    const Glx::RenderLargeRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& request_num = request.request_num;
  auto& request_total = request.request_total;
  uint32_t data_len{};
  auto& data = request.data;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 2;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // request_num
  Write(&request_num, &buf);

  // request_total
  Write(&request_total, &buf);

  // data_len
  data_len = data.size();
  Write(&data_len, &buf);

  // data
  DCHECK_EQ(static_cast<size_t>(data_len), data.size());
  for (auto& data_elem : data) {
    // data_elem
    Write(&data_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::CreateContext(
    const Glx::CreateContextRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context = request.context;
  auto& visual = request.visual;
  auto& screen = request.screen;
  auto& share_list = request.share_list;
  auto& is_direct = request.is_direct;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 3;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context
  Write(&context, &buf);

  // visual
  Write(&visual, &buf);

  // screen
  Write(&screen, &buf);

  // share_list
  Write(&share_list, &buf);

  // is_direct
  Write(&is_direct, &buf);

  // pad0
  Pad(&buf, 3);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::DestroyContext(
    const Glx::DestroyContextRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context = request.context;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 4;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context
  Write(&context, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<Glx::MakeCurrentReply>
Glx::MakeCurrent(
    const Glx::MakeCurrentRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& drawable = request.drawable;
  auto& context = request.context;
  auto& old_context_tag = request.old_context_tag;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 5;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // context
  Write(&context, &buf);

  // old_context_tag
  Write(&old_context_tag, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::MakeCurrentReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::MakeCurrentReply>
detail::ReadReply<Glx::MakeCurrentReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::MakeCurrentReply>();

  auto& sequence = (*reply).sequence;
  auto& context_tag = (*reply).context_tag;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // context_tag
  Read(&context_tag, &buf);

  // pad1
  Pad(&buf, 20);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::IsDirectReply>
Glx::IsDirect(
    const Glx::IsDirectRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context = request.context;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 6;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context
  Write(&context, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::IsDirectReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::IsDirectReply>
detail::ReadReply<Glx::IsDirectReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::IsDirectReply>();

  auto& sequence = (*reply).sequence;
  auto& is_direct = (*reply).is_direct;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // is_direct
  Read(&is_direct, &buf);

  // pad1
  Pad(&buf, 23);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::QueryVersionReply>
Glx::QueryVersion(
    const Glx::QueryVersionRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& major_version = request.major_version;
  auto& minor_version = request.minor_version;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 7;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // major_version
  Write(&major_version, &buf);

  // minor_version
  Write(&minor_version, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::QueryVersionReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::QueryVersionReply>
detail::ReadReply<Glx::QueryVersionReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::QueryVersionReply>();

  auto& sequence = (*reply).sequence;
  auto& major_version = (*reply).major_version;
  auto& minor_version = (*reply).minor_version;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // major_version
  Read(&major_version, &buf);

  // minor_version
  Read(&minor_version, &buf);

  // pad1
  Pad(&buf, 16);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
Glx::WaitGL(
    const Glx::WaitGLRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 8;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::WaitX(
    const Glx::WaitXRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 9;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::CopyContext(
    const Glx::CopyContextRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& src = request.src;
  auto& dest = request.dest;
  auto& mask = request.mask;
  auto& src_context_tag = request.src_context_tag;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 10;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // src
  Write(&src, &buf);

  // dest
  Write(&dest, &buf);

  // mask
  Write(&mask, &buf);

  // src_context_tag
  Write(&src_context_tag, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::SwapBuffers(
    const Glx::SwapBuffersRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& drawable = request.drawable;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 11;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // drawable
  Write(&drawable, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::UseXFont(
    const Glx::UseXFontRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& font = request.font;
  auto& first = request.first;
  auto& count = request.count;
  auto& list_base = request.list_base;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 12;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // font
  Write(&font, &buf);

  // first
  Write(&first, &buf);

  // count
  Write(&count, &buf);

  // list_base
  Write(&list_base, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::CreateGLXPixmap(
    const Glx::CreateGLXPixmapRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& screen = request.screen;
  auto& visual = request.visual;
  auto& pixmap = request.pixmap;
  auto& glx_pixmap = request.glx_pixmap;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 13;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // screen
  Write(&screen, &buf);

  // visual
  Write(&visual, &buf);

  // pixmap
  Write(&pixmap, &buf);

  // glx_pixmap
  Write(&glx_pixmap, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<Glx::GetVisualConfigsReply>
Glx::GetVisualConfigs(
    const Glx::GetVisualConfigsRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& screen = request.screen;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 14;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // screen
  Write(&screen, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetVisualConfigsReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetVisualConfigsReply>
detail::ReadReply<Glx::GetVisualConfigsReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetVisualConfigsReply>();

  auto& sequence = (*reply).sequence;
  auto& num_visuals = (*reply).num_visuals;
  auto& num_properties = (*reply).num_properties;
  auto& property_list = (*reply).property_list;
  size_t property_list_len = property_list.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // num_visuals
  Read(&num_visuals, &buf);

  // num_properties
  Read(&num_properties, &buf);

  // pad1
  Pad(&buf, 16);

  // property_list
  property_list.resize(length);
  for (auto& property_list_elem : property_list) {
    // property_list_elem
    Read(&property_list_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
Glx::DestroyGLXPixmap(
    const Glx::DestroyGLXPixmapRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& glx_pixmap = request.glx_pixmap;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 15;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // glx_pixmap
  Write(&glx_pixmap, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::VendorPrivate(
    const Glx::VendorPrivateRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& vendor_code = request.vendor_code;
  auto& context_tag = request.context_tag;
  auto& data = request.data;
  size_t data_len = data.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 16;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // vendor_code
  Write(&vendor_code, &buf);

  // context_tag
  Write(&context_tag, &buf);

  // data
  DCHECK_EQ(static_cast<size_t>(data_len), data.size());
  for (auto& data_elem : data) {
    // data_elem
    Write(&data_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<Glx::VendorPrivateWithReplyReply>
Glx::VendorPrivateWithReply(
    const Glx::VendorPrivateWithReplyRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& vendor_code = request.vendor_code;
  auto& context_tag = request.context_tag;
  auto& data = request.data;
  size_t data_len = data.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 17;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // vendor_code
  Write(&vendor_code, &buf);

  // context_tag
  Write(&context_tag, &buf);

  // data
  DCHECK_EQ(static_cast<size_t>(data_len), data.size());
  for (auto& data_elem : data) {
    // data_elem
    Write(&data_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<Glx::VendorPrivateWithReplyReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::VendorPrivateWithReplyReply>
detail::ReadReply<Glx::VendorPrivateWithReplyReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::VendorPrivateWithReplyReply>();

  auto& sequence = (*reply).sequence;
  auto& retval = (*reply).retval;
  auto& data1 = (*reply).data1;
  size_t data1_len = data1.size();
  auto& data2 = (*reply).data2;
  size_t data2_len = data2.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // retval
  Read(&retval, &buf);

  // data1
  for (auto& data1_elem : data1) {
    // data1_elem
    Read(&data1_elem, &buf);

  }

  // data2
  data2.resize((length) * (4));
  for (auto& data2_elem : data2) {
    // data2_elem
    Read(&data2_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::QueryExtensionsStringReply>
Glx::QueryExtensionsString(
    const Glx::QueryExtensionsStringRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& screen = request.screen;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 18;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // screen
  Write(&screen, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::QueryExtensionsStringReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::QueryExtensionsStringReply>
detail::ReadReply<Glx::QueryExtensionsStringReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::QueryExtensionsStringReply>();

  auto& sequence = (*reply).sequence;
  auto& n = (*reply).n;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // pad2
  Pad(&buf, 16);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::QueryServerStringReply>
Glx::QueryServerString(
    const Glx::QueryServerStringRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& screen = request.screen;
  auto& name = request.name;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 19;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // screen
  Write(&screen, &buf);

  // name
  Write(&name, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::QueryServerStringReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::QueryServerStringReply>
detail::ReadReply<Glx::QueryServerStringReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::QueryServerStringReply>();

  auto& sequence = (*reply).sequence;
  uint32_t str_len{};
  auto& string = (*reply).string;
  size_t string_len = string.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // str_len
  Read(&str_len, &buf);

  // pad2
  Pad(&buf, 16);

  // string
  string.resize(str_len);
  for (auto& string_elem : string) {
    // string_elem
    Read(&string_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
Glx::ClientInfo(
    const Glx::ClientInfoRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& major_version = request.major_version;
  auto& minor_version = request.minor_version;
  uint32_t str_len{};
  auto& string = request.string;
  size_t string_len = string.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 20;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // major_version
  Write(&major_version, &buf);

  // minor_version
  Write(&minor_version, &buf);

  // str_len
  str_len = string.size();
  Write(&str_len, &buf);

  // string
  DCHECK_EQ(static_cast<size_t>(str_len), string.size());
  for (auto& string_elem : string) {
    // string_elem
    Write(&string_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<Glx::GetFBConfigsReply>
Glx::GetFBConfigs(
    const Glx::GetFBConfigsRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& screen = request.screen;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 21;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // screen
  Write(&screen, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetFBConfigsReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetFBConfigsReply>
detail::ReadReply<Glx::GetFBConfigsReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetFBConfigsReply>();

  auto& sequence = (*reply).sequence;
  auto& num_FB_configs = (*reply).num_FB_configs;
  auto& num_properties = (*reply).num_properties;
  auto& property_list = (*reply).property_list;
  size_t property_list_len = property_list.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // num_FB_configs
  Read(&num_FB_configs, &buf);

  // num_properties
  Read(&num_properties, &buf);

  // pad1
  Pad(&buf, 16);

  // property_list
  property_list.resize(length);
  for (auto& property_list_elem : property_list) {
    // property_list_elem
    Read(&property_list_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
Glx::CreatePixmap(
    const Glx::CreatePixmapRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& screen = request.screen;
  auto& fbconfig = request.fbconfig;
  auto& pixmap = request.pixmap;
  auto& glx_pixmap = request.glx_pixmap;
  auto& num_attribs = request.num_attribs;
  auto& attribs = request.attribs;
  size_t attribs_len = attribs.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 22;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // screen
  Write(&screen, &buf);

  // fbconfig
  Write(&fbconfig, &buf);

  // pixmap
  Write(&pixmap, &buf);

  // glx_pixmap
  Write(&glx_pixmap, &buf);

  // num_attribs
  Write(&num_attribs, &buf);

  // attribs
  DCHECK_EQ(static_cast<size_t>((num_attribs) * (2)), attribs.size());
  for (auto& attribs_elem : attribs) {
    // attribs_elem
    Write(&attribs_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::DestroyPixmap(
    const Glx::DestroyPixmapRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& glx_pixmap = request.glx_pixmap;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 23;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // glx_pixmap
  Write(&glx_pixmap, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::CreateNewContext(
    const Glx::CreateNewContextRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context = request.context;
  auto& fbconfig = request.fbconfig;
  auto& screen = request.screen;
  auto& render_type = request.render_type;
  auto& share_list = request.share_list;
  auto& is_direct = request.is_direct;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 24;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context
  Write(&context, &buf);

  // fbconfig
  Write(&fbconfig, &buf);

  // screen
  Write(&screen, &buf);

  // render_type
  Write(&render_type, &buf);

  // share_list
  Write(&share_list, &buf);

  // is_direct
  Write(&is_direct, &buf);

  // pad0
  Pad(&buf, 3);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<Glx::QueryContextReply>
Glx::QueryContext(
    const Glx::QueryContextRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context = request.context;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 25;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context
  Write(&context, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::QueryContextReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::QueryContextReply>
detail::ReadReply<Glx::QueryContextReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::QueryContextReply>();

  auto& sequence = (*reply).sequence;
  auto& num_attribs = (*reply).num_attribs;
  auto& attribs = (*reply).attribs;
  size_t attribs_len = attribs.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // num_attribs
  Read(&num_attribs, &buf);

  // pad1
  Pad(&buf, 20);

  // attribs
  attribs.resize((num_attribs) * (2));
  for (auto& attribs_elem : attribs) {
    // attribs_elem
    Read(&attribs_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::MakeContextCurrentReply>
Glx::MakeContextCurrent(
    const Glx::MakeContextCurrentRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& old_context_tag = request.old_context_tag;
  auto& drawable = request.drawable;
  auto& read_drawable = request.read_drawable;
  auto& context = request.context;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 26;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // old_context_tag
  Write(&old_context_tag, &buf);

  // drawable
  Write(&drawable, &buf);

  // read_drawable
  Write(&read_drawable, &buf);

  // context
  Write(&context, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::MakeContextCurrentReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::MakeContextCurrentReply>
detail::ReadReply<Glx::MakeContextCurrentReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::MakeContextCurrentReply>();

  auto& sequence = (*reply).sequence;
  auto& context_tag = (*reply).context_tag;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // context_tag
  Read(&context_tag, &buf);

  // pad1
  Pad(&buf, 20);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
Glx::CreatePbuffer(
    const Glx::CreatePbufferRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& screen = request.screen;
  auto& fbconfig = request.fbconfig;
  auto& pbuffer = request.pbuffer;
  auto& num_attribs = request.num_attribs;
  auto& attribs = request.attribs;
  size_t attribs_len = attribs.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 27;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // screen
  Write(&screen, &buf);

  // fbconfig
  Write(&fbconfig, &buf);

  // pbuffer
  Write(&pbuffer, &buf);

  // num_attribs
  Write(&num_attribs, &buf);

  // attribs
  DCHECK_EQ(static_cast<size_t>((num_attribs) * (2)), attribs.size());
  for (auto& attribs_elem : attribs) {
    // attribs_elem
    Write(&attribs_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::DestroyPbuffer(
    const Glx::DestroyPbufferRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& pbuffer = request.pbuffer;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 28;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // pbuffer
  Write(&pbuffer, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<Glx::GetDrawableAttributesReply>
Glx::GetDrawableAttributes(
    const Glx::GetDrawableAttributesRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& drawable = request.drawable;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 29;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetDrawableAttributesReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetDrawableAttributesReply>
detail::ReadReply<Glx::GetDrawableAttributesReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetDrawableAttributesReply>();

  auto& sequence = (*reply).sequence;
  auto& num_attribs = (*reply).num_attribs;
  auto& attribs = (*reply).attribs;
  size_t attribs_len = attribs.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // num_attribs
  Read(&num_attribs, &buf);

  // pad1
  Pad(&buf, 20);

  // attribs
  attribs.resize((num_attribs) * (2));
  for (auto& attribs_elem : attribs) {
    // attribs_elem
    Read(&attribs_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
Glx::ChangeDrawableAttributes(
    const Glx::ChangeDrawableAttributesRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& drawable = request.drawable;
  auto& num_attribs = request.num_attribs;
  auto& attribs = request.attribs;
  size_t attribs_len = attribs.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 30;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // num_attribs
  Write(&num_attribs, &buf);

  // attribs
  DCHECK_EQ(static_cast<size_t>((num_attribs) * (2)), attribs.size());
  for (auto& attribs_elem : attribs) {
    // attribs_elem
    Write(&attribs_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::CreateWindow(
    const Glx::CreateWindowRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& screen = request.screen;
  auto& fbconfig = request.fbconfig;
  auto& window = request.window;
  auto& glx_window = request.glx_window;
  auto& num_attribs = request.num_attribs;
  auto& attribs = request.attribs;
  size_t attribs_len = attribs.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 31;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // screen
  Write(&screen, &buf);

  // fbconfig
  Write(&fbconfig, &buf);

  // window
  Write(&window, &buf);

  // glx_window
  Write(&glx_window, &buf);

  // num_attribs
  Write(&num_attribs, &buf);

  // attribs
  DCHECK_EQ(static_cast<size_t>((num_attribs) * (2)), attribs.size());
  for (auto& attribs_elem : attribs) {
    // attribs_elem
    Write(&attribs_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::DeleteWindow(
    const Glx::DeleteWindowRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& glxwindow = request.glxwindow;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 32;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // glxwindow
  Write(&glxwindow, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::SetClientInfoARB(
    const Glx::SetClientInfoARBRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& major_version = request.major_version;
  auto& minor_version = request.minor_version;
  auto& num_versions = request.num_versions;
  uint32_t gl_str_len{};
  uint32_t glx_str_len{};
  auto& gl_versions = request.gl_versions;
  size_t gl_versions_len = gl_versions.size();
  auto& gl_extension_string = request.gl_extension_string;
  size_t gl_extension_string_len = gl_extension_string.size();
  auto& glx_extension_string = request.glx_extension_string;
  size_t glx_extension_string_len = glx_extension_string.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 33;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // major_version
  Write(&major_version, &buf);

  // minor_version
  Write(&minor_version, &buf);

  // num_versions
  Write(&num_versions, &buf);

  // gl_str_len
  gl_str_len = gl_extension_string.size();
  Write(&gl_str_len, &buf);

  // glx_str_len
  glx_str_len = glx_extension_string.size();
  Write(&glx_str_len, &buf);

  // gl_versions
  DCHECK_EQ(static_cast<size_t>((num_versions) * (2)), gl_versions.size());
  for (auto& gl_versions_elem : gl_versions) {
    // gl_versions_elem
    Write(&gl_versions_elem, &buf);

  }

  // gl_extension_string
  DCHECK_EQ(static_cast<size_t>(gl_str_len), gl_extension_string.size());
  for (auto& gl_extension_string_elem : gl_extension_string) {
    // gl_extension_string_elem
    Write(&gl_extension_string_elem, &buf);

  }

  // glx_extension_string
  DCHECK_EQ(static_cast<size_t>(glx_str_len), glx_extension_string.size());
  for (auto& glx_extension_string_elem : glx_extension_string) {
    // glx_extension_string_elem
    Write(&glx_extension_string_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::CreateContextAttribsARB(
    const Glx::CreateContextAttribsARBRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context = request.context;
  auto& fbconfig = request.fbconfig;
  auto& screen = request.screen;
  auto& share_list = request.share_list;
  auto& is_direct = request.is_direct;
  auto& num_attribs = request.num_attribs;
  auto& attribs = request.attribs;
  size_t attribs_len = attribs.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 34;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context
  Write(&context, &buf);

  // fbconfig
  Write(&fbconfig, &buf);

  // screen
  Write(&screen, &buf);

  // share_list
  Write(&share_list, &buf);

  // is_direct
  Write(&is_direct, &buf);

  // pad0
  Pad(&buf, 3);

  // num_attribs
  Write(&num_attribs, &buf);

  // attribs
  DCHECK_EQ(static_cast<size_t>((num_attribs) * (2)), attribs.size());
  for (auto& attribs_elem : attribs) {
    // attribs_elem
    Write(&attribs_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::SetClientInfo2ARB(
    const Glx::SetClientInfo2ARBRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& major_version = request.major_version;
  auto& minor_version = request.minor_version;
  auto& num_versions = request.num_versions;
  uint32_t gl_str_len{};
  uint32_t glx_str_len{};
  auto& gl_versions = request.gl_versions;
  size_t gl_versions_len = gl_versions.size();
  auto& gl_extension_string = request.gl_extension_string;
  size_t gl_extension_string_len = gl_extension_string.size();
  auto& glx_extension_string = request.glx_extension_string;
  size_t glx_extension_string_len = glx_extension_string.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 35;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // major_version
  Write(&major_version, &buf);

  // minor_version
  Write(&minor_version, &buf);

  // num_versions
  Write(&num_versions, &buf);

  // gl_str_len
  gl_str_len = gl_extension_string.size();
  Write(&gl_str_len, &buf);

  // glx_str_len
  glx_str_len = glx_extension_string.size();
  Write(&glx_str_len, &buf);

  // gl_versions
  DCHECK_EQ(static_cast<size_t>((num_versions) * (3)), gl_versions.size());
  for (auto& gl_versions_elem : gl_versions) {
    // gl_versions_elem
    Write(&gl_versions_elem, &buf);

  }

  // gl_extension_string
  DCHECK_EQ(static_cast<size_t>(gl_str_len), gl_extension_string.size());
  for (auto& gl_extension_string_elem : gl_extension_string) {
    // gl_extension_string_elem
    Write(&gl_extension_string_elem, &buf);

  }

  // glx_extension_string
  DCHECK_EQ(static_cast<size_t>(glx_str_len), glx_extension_string.size());
  for (auto& glx_extension_string_elem : glx_extension_string) {
    // glx_extension_string_elem
    Write(&glx_extension_string_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::NewList(
    const Glx::NewListRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& list = request.list;
  auto& mode = request.mode;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 101;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // list
  Write(&list, &buf);

  // mode
  Write(&mode, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::EndList(
    const Glx::EndListRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 102;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::DeleteLists(
    const Glx::DeleteListsRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& list = request.list;
  auto& range = request.range;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 103;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // list
  Write(&list, &buf);

  // range
  Write(&range, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<Glx::GenListsReply>
Glx::GenLists(
    const Glx::GenListsRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& range = request.range;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 104;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // range
  Write(&range, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GenListsReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GenListsReply>
detail::ReadReply<Glx::GenListsReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GenListsReply>();

  auto& sequence = (*reply).sequence;
  auto& ret_val = (*reply).ret_val;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // ret_val
  Read(&ret_val, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
Glx::FeedbackBuffer(
    const Glx::FeedbackBufferRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& size = request.size;
  auto& type = request.type;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 105;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // size
  Write(&size, &buf);

  // type
  Write(&type, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::SelectBuffer(
    const Glx::SelectBufferRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& size = request.size;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 106;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // size
  Write(&size, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<Glx::RenderModeReply>
Glx::RenderMode(
    const Glx::RenderModeRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& mode = request.mode;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 107;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // mode
  Write(&mode, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::RenderModeReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::RenderModeReply>
detail::ReadReply<Glx::RenderModeReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::RenderModeReply>();

  auto& sequence = (*reply).sequence;
  auto& ret_val = (*reply).ret_val;
  uint32_t n{};
  auto& new_mode = (*reply).new_mode;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // ret_val
  Read(&ret_val, &buf);

  // n
  Read(&n, &buf);

  // new_mode
  Read(&new_mode, &buf);

  // pad1
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::FinishReply>
Glx::Finish(
    const Glx::FinishRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 108;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::FinishReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::FinishReply>
detail::ReadReply<Glx::FinishReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::FinishReply>();

  auto& sequence = (*reply).sequence;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
Glx::PixelStoref(
    const Glx::PixelStorefRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& pname = request.pname;
  auto& datum = request.datum;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 109;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // pname
  Write(&pname, &buf);

  // datum
  Write(&datum, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Glx::PixelStorei(
    const Glx::PixelStoreiRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& pname = request.pname;
  auto& datum = request.datum;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 110;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // pname
  Write(&pname, &buf);

  // datum
  Write(&datum, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<Glx::ReadPixelsReply>
Glx::ReadPixels(
    const Glx::ReadPixelsRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& x = request.x;
  auto& y = request.y;
  auto& width = request.width;
  auto& height = request.height;
  auto& format = request.format;
  auto& type = request.type;
  auto& swap_bytes = request.swap_bytes;
  auto& lsb_first = request.lsb_first;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 111;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // x
  Write(&x, &buf);

  // y
  Write(&y, &buf);

  // width
  Write(&width, &buf);

  // height
  Write(&height, &buf);

  // format
  Write(&format, &buf);

  // type
  Write(&type, &buf);

  // swap_bytes
  Write(&swap_bytes, &buf);

  // lsb_first
  Write(&lsb_first, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::ReadPixelsReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::ReadPixelsReply>
detail::ReadReply<Glx::ReadPixelsReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::ReadPixelsReply>();

  auto& sequence = (*reply).sequence;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 24);

  // data
  data.resize((length) * (4));
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetBooleanvReply>
Glx::GetBooleanv(
    const Glx::GetBooleanvRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 112;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetBooleanvReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetBooleanvReply>
detail::ReadReply<Glx::GetBooleanvReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetBooleanvReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 15);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetClipPlaneReply>
Glx::GetClipPlane(
    const Glx::GetClipPlaneRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& plane = request.plane;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 113;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // plane
  Write(&plane, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetClipPlaneReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetClipPlaneReply>
detail::ReadReply<Glx::GetClipPlaneReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetClipPlaneReply>();

  auto& sequence = (*reply).sequence;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 24);

  // data
  data.resize((length) / (2));
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetDoublevReply>
Glx::GetDoublev(
    const Glx::GetDoublevRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 114;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetDoublevReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetDoublevReply>
detail::ReadReply<Glx::GetDoublevReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetDoublevReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 8);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetErrorReply>
Glx::GetError(
    const Glx::GetErrorRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 115;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetErrorReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetErrorReply>
detail::ReadReply<Glx::GetErrorReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetErrorReply>();

  auto& sequence = (*reply).sequence;
  auto& error = (*reply).error;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // error
  Read(&error, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetFloatvReply>
Glx::GetFloatv(
    const Glx::GetFloatvRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 116;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetFloatvReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetFloatvReply>
detail::ReadReply<Glx::GetFloatvReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetFloatvReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetIntegervReply>
Glx::GetIntegerv(
    const Glx::GetIntegervRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 117;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetIntegervReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetIntegervReply>
detail::ReadReply<Glx::GetIntegervReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetIntegervReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetLightfvReply>
Glx::GetLightfv(
    const Glx::GetLightfvRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& light = request.light;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 118;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // light
  Write(&light, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetLightfvReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetLightfvReply>
detail::ReadReply<Glx::GetLightfvReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetLightfvReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetLightivReply>
Glx::GetLightiv(
    const Glx::GetLightivRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& light = request.light;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 119;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // light
  Write(&light, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetLightivReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetLightivReply>
detail::ReadReply<Glx::GetLightivReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetLightivReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetMapdvReply>
Glx::GetMapdv(
    const Glx::GetMapdvRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& query = request.query;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 120;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // query
  Write(&query, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetMapdvReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetMapdvReply>
detail::ReadReply<Glx::GetMapdvReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetMapdvReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 8);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetMapfvReply>
Glx::GetMapfv(
    const Glx::GetMapfvRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& query = request.query;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 121;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // query
  Write(&query, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetMapfvReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetMapfvReply>
detail::ReadReply<Glx::GetMapfvReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetMapfvReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetMapivReply>
Glx::GetMapiv(
    const Glx::GetMapivRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& query = request.query;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 122;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // query
  Write(&query, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetMapivReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetMapivReply>
detail::ReadReply<Glx::GetMapivReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetMapivReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetMaterialfvReply>
Glx::GetMaterialfv(
    const Glx::GetMaterialfvRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& face = request.face;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 123;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // face
  Write(&face, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetMaterialfvReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetMaterialfvReply>
detail::ReadReply<Glx::GetMaterialfvReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetMaterialfvReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetMaterialivReply>
Glx::GetMaterialiv(
    const Glx::GetMaterialivRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& face = request.face;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 124;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // face
  Write(&face, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetMaterialivReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetMaterialivReply>
detail::ReadReply<Glx::GetMaterialivReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetMaterialivReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetPixelMapfvReply>
Glx::GetPixelMapfv(
    const Glx::GetPixelMapfvRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& map = request.map;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 125;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // map
  Write(&map, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetPixelMapfvReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetPixelMapfvReply>
detail::ReadReply<Glx::GetPixelMapfvReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetPixelMapfvReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetPixelMapuivReply>
Glx::GetPixelMapuiv(
    const Glx::GetPixelMapuivRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& map = request.map;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 126;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // map
  Write(&map, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetPixelMapuivReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetPixelMapuivReply>
detail::ReadReply<Glx::GetPixelMapuivReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetPixelMapuivReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetPixelMapusvReply>
Glx::GetPixelMapusv(
    const Glx::GetPixelMapusvRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& map = request.map;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 127;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // map
  Write(&map, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetPixelMapusvReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetPixelMapusvReply>
detail::ReadReply<Glx::GetPixelMapusvReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetPixelMapusvReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 16);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetPolygonStippleReply>
Glx::GetPolygonStipple(
    const Glx::GetPolygonStippleRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& lsb_first = request.lsb_first;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 128;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // lsb_first
  Write(&lsb_first, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetPolygonStippleReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetPolygonStippleReply>
detail::ReadReply<Glx::GetPolygonStippleReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetPolygonStippleReply>();

  auto& sequence = (*reply).sequence;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 24);

  // data
  data.resize((length) * (4));
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetStringReply>
Glx::GetString(
    const Glx::GetStringRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& name = request.name;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 129;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // name
  Write(&name, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetStringReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetStringReply>
detail::ReadReply<Glx::GetStringReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetStringReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& string = (*reply).string;
  size_t string_len = string.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // pad2
  Pad(&buf, 16);

  // string
  string.resize(n);
  for (auto& string_elem : string) {
    // string_elem
    Read(&string_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetTexEnvfvReply>
Glx::GetTexEnvfv(
    const Glx::GetTexEnvfvRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 130;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetTexEnvfvReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetTexEnvfvReply>
detail::ReadReply<Glx::GetTexEnvfvReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetTexEnvfvReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetTexEnvivReply>
Glx::GetTexEnviv(
    const Glx::GetTexEnvivRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 131;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetTexEnvivReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetTexEnvivReply>
detail::ReadReply<Glx::GetTexEnvivReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetTexEnvivReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetTexGendvReply>
Glx::GetTexGendv(
    const Glx::GetTexGendvRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& coord = request.coord;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 132;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // coord
  Write(&coord, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetTexGendvReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetTexGendvReply>
detail::ReadReply<Glx::GetTexGendvReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetTexGendvReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 8);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetTexGenfvReply>
Glx::GetTexGenfv(
    const Glx::GetTexGenfvRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& coord = request.coord;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 133;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // coord
  Write(&coord, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetTexGenfvReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetTexGenfvReply>
detail::ReadReply<Glx::GetTexGenfvReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetTexGenfvReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetTexGenivReply>
Glx::GetTexGeniv(
    const Glx::GetTexGenivRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& coord = request.coord;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 134;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // coord
  Write(&coord, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetTexGenivReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetTexGenivReply>
detail::ReadReply<Glx::GetTexGenivReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetTexGenivReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetTexImageReply>
Glx::GetTexImage(
    const Glx::GetTexImageRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& level = request.level;
  auto& format = request.format;
  auto& type = request.type;
  auto& swap_bytes = request.swap_bytes;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 135;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // level
  Write(&level, &buf);

  // format
  Write(&format, &buf);

  // type
  Write(&type, &buf);

  // swap_bytes
  Write(&swap_bytes, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetTexImageReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetTexImageReply>
detail::ReadReply<Glx::GetTexImageReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetTexImageReply>();

  auto& sequence = (*reply).sequence;
  auto& width = (*reply).width;
  auto& height = (*reply).height;
  auto& depth = (*reply).depth;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 8);

  // width
  Read(&width, &buf);

  // height
  Read(&height, &buf);

  // depth
  Read(&depth, &buf);

  // pad2
  Pad(&buf, 4);

  // data
  data.resize((length) * (4));
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetTexParameterfvReply>
Glx::GetTexParameterfv(
    const Glx::GetTexParameterfvRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 136;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetTexParameterfvReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetTexParameterfvReply>
detail::ReadReply<Glx::GetTexParameterfvReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetTexParameterfvReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetTexParameterivReply>
Glx::GetTexParameteriv(
    const Glx::GetTexParameterivRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 137;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetTexParameterivReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetTexParameterivReply>
detail::ReadReply<Glx::GetTexParameterivReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetTexParameterivReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetTexLevelParameterfvReply>
Glx::GetTexLevelParameterfv(
    const Glx::GetTexLevelParameterfvRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& level = request.level;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 138;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // level
  Write(&level, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetTexLevelParameterfvReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetTexLevelParameterfvReply>
detail::ReadReply<Glx::GetTexLevelParameterfvReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetTexLevelParameterfvReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetTexLevelParameterivReply>
Glx::GetTexLevelParameteriv(
    const Glx::GetTexLevelParameterivRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& level = request.level;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 139;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // level
  Write(&level, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetTexLevelParameterivReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetTexLevelParameterivReply>
detail::ReadReply<Glx::GetTexLevelParameterivReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetTexLevelParameterivReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::IsEnabledReply>
Glx::IsEnabled(
    const Glx::IsEnabledRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& capability = request.capability;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 140;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // capability
  Write(&capability, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::IsEnabledReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::IsEnabledReply>
detail::ReadReply<Glx::IsEnabledReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::IsEnabledReply>();

  auto& sequence = (*reply).sequence;
  auto& ret_val = (*reply).ret_val;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // ret_val
  Read(&ret_val, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::IsListReply>
Glx::IsList(
    const Glx::IsListRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& list = request.list;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 141;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // list
  Write(&list, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::IsListReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::IsListReply>
detail::ReadReply<Glx::IsListReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::IsListReply>();

  auto& sequence = (*reply).sequence;
  auto& ret_val = (*reply).ret_val;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // ret_val
  Read(&ret_val, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
Glx::Flush(
    const Glx::FlushRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 142;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<Glx::AreTexturesResidentReply>
Glx::AreTexturesResident(
    const Glx::AreTexturesResidentRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  int32_t n{};
  auto& textures = request.textures;
  size_t textures_len = textures.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 143;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // n
  n = textures.size();
  Write(&n, &buf);

  // textures
  DCHECK_EQ(static_cast<size_t>(n), textures.size());
  for (auto& textures_elem : textures) {
    // textures_elem
    Write(&textures_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<Glx::AreTexturesResidentReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::AreTexturesResidentReply>
detail::ReadReply<Glx::AreTexturesResidentReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::AreTexturesResidentReply>();

  auto& sequence = (*reply).sequence;
  auto& ret_val = (*reply).ret_val;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // ret_val
  Read(&ret_val, &buf);

  // pad1
  Pad(&buf, 20);

  // data
  data.resize((length) * (4));
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
Glx::DeleteTextures(
    const Glx::DeleteTexturesRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  int32_t n{};
  auto& textures = request.textures;
  size_t textures_len = textures.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 144;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // n
  n = textures.size();
  Write(&n, &buf);

  // textures
  DCHECK_EQ(static_cast<size_t>(n), textures.size());
  for (auto& textures_elem : textures) {
    // textures_elem
    Write(&textures_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<Glx::GenTexturesReply>
Glx::GenTextures(
    const Glx::GenTexturesRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& n = request.n;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 145;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // n
  Write(&n, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GenTexturesReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GenTexturesReply>
detail::ReadReply<Glx::GenTexturesReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GenTexturesReply>();

  auto& sequence = (*reply).sequence;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 24);

  // data
  data.resize(length);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::IsTextureReply>
Glx::IsTexture(
    const Glx::IsTextureRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& texture = request.texture;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 146;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // texture
  Write(&texture, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::IsTextureReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::IsTextureReply>
detail::ReadReply<Glx::IsTextureReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::IsTextureReply>();

  auto& sequence = (*reply).sequence;
  auto& ret_val = (*reply).ret_val;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // ret_val
  Read(&ret_val, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetColorTableReply>
Glx::GetColorTable(
    const Glx::GetColorTableRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& format = request.format;
  auto& type = request.type;
  auto& swap_bytes = request.swap_bytes;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 147;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // format
  Write(&format, &buf);

  // type
  Write(&type, &buf);

  // swap_bytes
  Write(&swap_bytes, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetColorTableReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetColorTableReply>
detail::ReadReply<Glx::GetColorTableReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetColorTableReply>();

  auto& sequence = (*reply).sequence;
  auto& width = (*reply).width;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 8);

  // width
  Read(&width, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize((length) * (4));
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetColorTableParameterfvReply>
Glx::GetColorTableParameterfv(
    const Glx::GetColorTableParameterfvRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 148;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetColorTableParameterfvReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetColorTableParameterfvReply>
detail::ReadReply<Glx::GetColorTableParameterfvReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetColorTableParameterfvReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetColorTableParameterivReply>
Glx::GetColorTableParameteriv(
    const Glx::GetColorTableParameterivRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 149;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetColorTableParameterivReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetColorTableParameterivReply>
detail::ReadReply<Glx::GetColorTableParameterivReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetColorTableParameterivReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetConvolutionFilterReply>
Glx::GetConvolutionFilter(
    const Glx::GetConvolutionFilterRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& format = request.format;
  auto& type = request.type;
  auto& swap_bytes = request.swap_bytes;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 150;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // format
  Write(&format, &buf);

  // type
  Write(&type, &buf);

  // swap_bytes
  Write(&swap_bytes, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetConvolutionFilterReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetConvolutionFilterReply>
detail::ReadReply<Glx::GetConvolutionFilterReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetConvolutionFilterReply>();

  auto& sequence = (*reply).sequence;
  auto& width = (*reply).width;
  auto& height = (*reply).height;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 8);

  // width
  Read(&width, &buf);

  // height
  Read(&height, &buf);

  // pad2
  Pad(&buf, 8);

  // data
  data.resize((length) * (4));
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetConvolutionParameterfvReply>
Glx::GetConvolutionParameterfv(
    const Glx::GetConvolutionParameterfvRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 151;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetConvolutionParameterfvReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetConvolutionParameterfvReply>
detail::ReadReply<Glx::GetConvolutionParameterfvReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetConvolutionParameterfvReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetConvolutionParameterivReply>
Glx::GetConvolutionParameteriv(
    const Glx::GetConvolutionParameterivRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 152;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetConvolutionParameterivReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetConvolutionParameterivReply>
detail::ReadReply<Glx::GetConvolutionParameterivReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetConvolutionParameterivReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetSeparableFilterReply>
Glx::GetSeparableFilter(
    const Glx::GetSeparableFilterRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& format = request.format;
  auto& type = request.type;
  auto& swap_bytes = request.swap_bytes;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 153;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // format
  Write(&format, &buf);

  // type
  Write(&type, &buf);

  // swap_bytes
  Write(&swap_bytes, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetSeparableFilterReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetSeparableFilterReply>
detail::ReadReply<Glx::GetSeparableFilterReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetSeparableFilterReply>();

  auto& sequence = (*reply).sequence;
  auto& row_w = (*reply).row_w;
  auto& col_h = (*reply).col_h;
  auto& rows_and_cols = (*reply).rows_and_cols;
  size_t rows_and_cols_len = rows_and_cols.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 8);

  // row_w
  Read(&row_w, &buf);

  // col_h
  Read(&col_h, &buf);

  // pad2
  Pad(&buf, 8);

  // rows_and_cols
  rows_and_cols.resize((length) * (4));
  for (auto& rows_and_cols_elem : rows_and_cols) {
    // rows_and_cols_elem
    Read(&rows_and_cols_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetHistogramReply>
Glx::GetHistogram(
    const Glx::GetHistogramRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& format = request.format;
  auto& type = request.type;
  auto& swap_bytes = request.swap_bytes;
  auto& reset = request.reset;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 154;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // format
  Write(&format, &buf);

  // type
  Write(&type, &buf);

  // swap_bytes
  Write(&swap_bytes, &buf);

  // reset
  Write(&reset, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetHistogramReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetHistogramReply>
detail::ReadReply<Glx::GetHistogramReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetHistogramReply>();

  auto& sequence = (*reply).sequence;
  auto& width = (*reply).width;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 8);

  // width
  Read(&width, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize((length) * (4));
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetHistogramParameterfvReply>
Glx::GetHistogramParameterfv(
    const Glx::GetHistogramParameterfvRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 155;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetHistogramParameterfvReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetHistogramParameterfvReply>
detail::ReadReply<Glx::GetHistogramParameterfvReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetHistogramParameterfvReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetHistogramParameterivReply>
Glx::GetHistogramParameteriv(
    const Glx::GetHistogramParameterivRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 156;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetHistogramParameterivReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetHistogramParameterivReply>
detail::ReadReply<Glx::GetHistogramParameterivReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetHistogramParameterivReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetMinmaxReply>
Glx::GetMinmax(
    const Glx::GetMinmaxRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& format = request.format;
  auto& type = request.type;
  auto& swap_bytes = request.swap_bytes;
  auto& reset = request.reset;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 157;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // format
  Write(&format, &buf);

  // type
  Write(&type, &buf);

  // swap_bytes
  Write(&swap_bytes, &buf);

  // reset
  Write(&reset, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetMinmaxReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetMinmaxReply>
detail::ReadReply<Glx::GetMinmaxReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetMinmaxReply>();

  auto& sequence = (*reply).sequence;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 24);

  // data
  data.resize((length) * (4));
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetMinmaxParameterfvReply>
Glx::GetMinmaxParameterfv(
    const Glx::GetMinmaxParameterfvRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 158;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetMinmaxParameterfvReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetMinmaxParameterfvReply>
detail::ReadReply<Glx::GetMinmaxParameterfvReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetMinmaxParameterfvReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetMinmaxParameterivReply>
Glx::GetMinmaxParameteriv(
    const Glx::GetMinmaxParameterivRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 159;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetMinmaxParameterivReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetMinmaxParameterivReply>
detail::ReadReply<Glx::GetMinmaxParameterivReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetMinmaxParameterivReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetCompressedTexImageARBReply>
Glx::GetCompressedTexImageARB(
    const Glx::GetCompressedTexImageARBRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& level = request.level;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 160;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // level
  Write(&level, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetCompressedTexImageARBReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetCompressedTexImageARBReply>
detail::ReadReply<Glx::GetCompressedTexImageARBReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetCompressedTexImageARBReply>();

  auto& sequence = (*reply).sequence;
  auto& size = (*reply).size;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 8);

  // size
  Read(&size, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize((length) * (4));
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
Glx::DeleteQueriesARB(
    const Glx::DeleteQueriesARBRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  int32_t n{};
  auto& ids = request.ids;
  size_t ids_len = ids.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 161;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // n
  n = ids.size();
  Write(&n, &buf);

  // ids
  DCHECK_EQ(static_cast<size_t>(n), ids.size());
  for (auto& ids_elem : ids) {
    // ids_elem
    Write(&ids_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<Glx::GenQueriesARBReply>
Glx::GenQueriesARB(
    const Glx::GenQueriesARBRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& n = request.n;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 162;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // n
  Write(&n, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GenQueriesARBReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GenQueriesARBReply>
detail::ReadReply<Glx::GenQueriesARBReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GenQueriesARBReply>();

  auto& sequence = (*reply).sequence;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 24);

  // data
  data.resize(length);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::IsQueryARBReply>
Glx::IsQueryARB(
    const Glx::IsQueryARBRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& id = request.id;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 163;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // id
  Write(&id, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::IsQueryARBReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::IsQueryARBReply>
detail::ReadReply<Glx::IsQueryARBReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::IsQueryARBReply>();

  auto& sequence = (*reply).sequence;
  auto& ret_val = (*reply).ret_val;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // ret_val
  Read(&ret_val, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetQueryivARBReply>
Glx::GetQueryivARB(
    const Glx::GetQueryivARBRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& target = request.target;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 164;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // target
  Write(&target, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetQueryivARBReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetQueryivARBReply>
detail::ReadReply<Glx::GetQueryivARBReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetQueryivARBReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetQueryObjectivARBReply>
Glx::GetQueryObjectivARB(
    const Glx::GetQueryObjectivARBRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& id = request.id;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 165;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // id
  Write(&id, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetQueryObjectivARBReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetQueryObjectivARBReply>
detail::ReadReply<Glx::GetQueryObjectivARBReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetQueryObjectivARBReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Glx::GetQueryObjectuivARBReply>
Glx::GetQueryObjectuivARB(
    const Glx::GetQueryObjectuivARBRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& context_tag = request.context_tag;
  auto& id = request.id;
  auto& pname = request.pname;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 166;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // context_tag
  Write(&context_tag, &buf);

  // id
  Write(&id, &buf);

  // pname
  Write(&pname, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Glx::GetQueryObjectuivARBReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Glx::GetQueryObjectuivARBReply>
detail::ReadReply<Glx::GetQueryObjectuivARBReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Glx::GetQueryObjectuivARBReply>();

  auto& sequence = (*reply).sequence;
  uint32_t n{};
  auto& datum = (*reply).datum;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad1
  Pad(&buf, 4);

  // n
  Read(&n, &buf);

  // datum
  Read(&datum, &buf);

  // pad2
  Pad(&buf, 12);

  // data
  data.resize(n);
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

}  // namespace x11
