#include "render.h"

#include <xcb/xcb.h>
#include <xcb/xcbext.h>

#include "base/logging.h"
#include "ui/gfx/x/xproto_internal.h"

namespace x11 {

Render::Render(x11::Connection* connection,
    const x11::QueryExtensionReply& info)
    : connection_(connection), info_(info) {}

Future<Render::QueryVersionReply>
Render::QueryVersion(
    const Render::QueryVersionRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& client_major_version = request.client_major_version;
  auto& client_minor_version = request.client_minor_version;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 0;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // client_major_version
  Write(&client_major_version, &buf);

  // client_minor_version
  Write(&client_minor_version, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Render::QueryVersionReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Render::QueryVersionReply>
detail::ReadReply<Render::QueryVersionReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Render::QueryVersionReply>();

  auto& sequence = (*reply).sequence;
  auto& major_version = (*reply).major_version;
  auto& minor_version = (*reply).minor_version;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // major_version
  Read(&major_version, &buf);

  // minor_version
  Read(&minor_version, &buf);

  // pad1
  Pad(&buf, 16);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Render::QueryPictFormatsReply>
Render::QueryPictFormats(
    const Render::QueryPictFormatsRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 1;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<Render::QueryPictFormatsReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Render::QueryPictFormatsReply>
detail::ReadReply<Render::QueryPictFormatsReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Render::QueryPictFormatsReply>();

  auto& sequence = (*reply).sequence;
  uint32_t num_formats{};
  uint32_t num_screens{};
  auto& num_depths = (*reply).num_depths;
  auto& num_visuals = (*reply).num_visuals;
  uint32_t num_subpixel{};
  auto& formats = (*reply).formats;
  size_t formats_len = formats.size();
  auto& screens = (*reply).screens;
  size_t screens_len = screens.size();
  auto& subpixels = (*reply).subpixels;
  size_t subpixels_len = subpixels.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // num_formats
  Read(&num_formats, &buf);

  // num_screens
  Read(&num_screens, &buf);

  // num_depths
  Read(&num_depths, &buf);

  // num_visuals
  Read(&num_visuals, &buf);

  // num_subpixel
  Read(&num_subpixel, &buf);

  // pad1
  Pad(&buf, 4);

  // formats
  formats.resize(num_formats);
  for (auto& formats_elem : formats) {
    // formats_elem
    {
      auto& id = formats_elem.id;
      auto& type = formats_elem.type;
      auto& depth = formats_elem.depth;
      auto& direct = formats_elem.direct;
      auto& colormap = formats_elem.colormap;

      // id
      Read(&id, &buf);

      // type
      uint8_t tmp0;
      Read(&tmp0, &buf);
      type = static_cast<Render::PictType>(tmp0);

      // depth
      Read(&depth, &buf);

      // pad0
      Pad(&buf, 2);

      // direct
      {
        auto& red_shift = direct.red_shift;
        auto& red_mask = direct.red_mask;
        auto& green_shift = direct.green_shift;
        auto& green_mask = direct.green_mask;
        auto& blue_shift = direct.blue_shift;
        auto& blue_mask = direct.blue_mask;
        auto& alpha_shift = direct.alpha_shift;
        auto& alpha_mask = direct.alpha_mask;

        // red_shift
        Read(&red_shift, &buf);

        // red_mask
        Read(&red_mask, &buf);

        // green_shift
        Read(&green_shift, &buf);

        // green_mask
        Read(&green_mask, &buf);

        // blue_shift
        Read(&blue_shift, &buf);

        // blue_mask
        Read(&blue_mask, &buf);

        // alpha_shift
        Read(&alpha_shift, &buf);

        // alpha_mask
        Read(&alpha_mask, &buf);

      }

      // colormap
      Read(&colormap, &buf);

    }

  }

  // screens
  screens.resize(num_screens);
  for (auto& screens_elem : screens) {
    // screens_elem
    {
      uint32_t num_depths{};
      auto& fallback = screens_elem.fallback;
      auto& depths = screens_elem.depths;
      size_t depths_len = depths.size();

      // num_depths
      Read(&num_depths, &buf);

      // fallback
      Read(&fallback, &buf);

      // depths
      depths.resize(num_depths);
      for (auto& depths_elem : depths) {
        // depths_elem
        {
          auto& depth = depths_elem.depth;
          uint16_t num_visuals{};
          auto& visuals = depths_elem.visuals;
          size_t visuals_len = visuals.size();

          // depth
          Read(&depth, &buf);

          // pad0
          Pad(&buf, 1);

          // num_visuals
          Read(&num_visuals, &buf);

          // pad1
          Pad(&buf, 4);

          // visuals
          visuals.resize(num_visuals);
          for (auto& visuals_elem : visuals) {
            // visuals_elem
            {
              auto& visual = visuals_elem.visual;
              auto& format = visuals_elem.format;

              // visual
              Read(&visual, &buf);

              // format
              Read(&format, &buf);

            }

          }

        }

      }

    }

  }

  // subpixels
  subpixels.resize(num_subpixel);
  for (auto& subpixels_elem : subpixels) {
    // subpixels_elem
    uint32_t tmp1;
    Read(&tmp1, &buf);
    subpixels_elem = static_cast<Render::SubPixel>(tmp1);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<Render::QueryPictIndexValuesReply>
Render::QueryPictIndexValues(
    const Render::QueryPictIndexValuesRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& format = request.format;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 2;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // format
  Write(&format, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Render::QueryPictIndexValuesReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Render::QueryPictIndexValuesReply>
detail::ReadReply<Render::QueryPictIndexValuesReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Render::QueryPictIndexValuesReply>();

  auto& sequence = (*reply).sequence;
  uint32_t num_values{};
  auto& values = (*reply).values;
  size_t values_len = values.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // num_values
  Read(&num_values, &buf);

  // pad1
  Pad(&buf, 20);

  // values
  values.resize(num_values);
  for (auto& values_elem : values) {
    // values_elem
    {
      auto& pixel = values_elem.pixel;
      auto& red = values_elem.red;
      auto& green = values_elem.green;
      auto& blue = values_elem.blue;
      auto& alpha = values_elem.alpha;

      // pixel
      Read(&pixel, &buf);

      // red
      Read(&red, &buf);

      // green
      Read(&green, &buf);

      // blue
      Read(&blue, &buf);

      // alpha
      Read(&alpha, &buf);

    }

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
Render::CreatePicture(
    const Render::CreatePictureRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& pid = request.pid;
  auto& drawable = request.drawable;
  auto& format = request.format;
  CreatePictureAttribute value_mask{};
  auto& value_list = request;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 4;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // pid
  Write(&pid, &buf);

  // drawable
  Write(&drawable, &buf);

  // format
  Write(&format, &buf);

  // value_mask
  SwitchVar(CreatePictureAttribute::Repeat, value_list.repeat.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::AlphaMap, value_list.alphamap.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::AlphaXOrigin, value_list.alphaxorigin.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::AlphaYOrigin, value_list.alphayorigin.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::ClipXOrigin, value_list.clipxorigin.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::ClipYOrigin, value_list.clipyorigin.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::ClipMask, value_list.clipmask.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::GraphicsExposure, value_list.graphicsexposure.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::SubwindowMode, value_list.subwindowmode.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::PolyEdge, value_list.polyedge.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::PolyMode, value_list.polymode.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::Dither, value_list.dither.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::ComponentAlpha, value_list.componentalpha.has_value(), true, &value_mask);
  uint32_t tmp2;
  tmp2 = static_cast<uint32_t>(value_mask);
  Write(&tmp2, &buf);

  // value_list
  auto value_list_expr = value_mask;
  if (CaseAnd(value_list_expr, CreatePictureAttribute::Repeat)) {
    auto& repeat = *value_list.repeat;

    // repeat
    uint32_t tmp3;
    tmp3 = static_cast<uint32_t>(repeat);
    Write(&tmp3, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::AlphaMap)) {
    auto& alphamap = *value_list.alphamap;

    // alphamap
    Write(&alphamap, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::AlphaXOrigin)) {
    auto& alphaxorigin = *value_list.alphaxorigin;

    // alphaxorigin
    Write(&alphaxorigin, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::AlphaYOrigin)) {
    auto& alphayorigin = *value_list.alphayorigin;

    // alphayorigin
    Write(&alphayorigin, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::ClipXOrigin)) {
    auto& clipxorigin = *value_list.clipxorigin;

    // clipxorigin
    Write(&clipxorigin, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::ClipYOrigin)) {
    auto& clipyorigin = *value_list.clipyorigin;

    // clipyorigin
    Write(&clipyorigin, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::ClipMask)) {
    auto& clipmask = *value_list.clipmask;

    // clipmask
    Write(&clipmask, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::GraphicsExposure)) {
    auto& graphicsexposure = *value_list.graphicsexposure;

    // graphicsexposure
    Write(&graphicsexposure, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::SubwindowMode)) {
    auto& subwindowmode = *value_list.subwindowmode;

    // subwindowmode
    uint32_t tmp4;
    tmp4 = static_cast<uint32_t>(subwindowmode);
    Write(&tmp4, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::PolyEdge)) {
    auto& polyedge = *value_list.polyedge;

    // polyedge
    uint32_t tmp5;
    tmp5 = static_cast<uint32_t>(polyedge);
    Write(&tmp5, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::PolyMode)) {
    auto& polymode = *value_list.polymode;

    // polymode
    uint32_t tmp6;
    tmp6 = static_cast<uint32_t>(polymode);
    Write(&tmp6, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::Dither)) {
    auto& dither = *value_list.dither;

    // dither
    Write(&dither, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::ComponentAlpha)) {
    auto& componentalpha = *value_list.componentalpha;

    // componentalpha
    Write(&componentalpha, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::ChangePicture(
    const Render::ChangePictureRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& picture = request.picture;
  CreatePictureAttribute value_mask{};
  auto& value_list = request;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 5;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // picture
  Write(&picture, &buf);

  // value_mask
  SwitchVar(CreatePictureAttribute::Repeat, value_list.repeat.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::AlphaMap, value_list.alphamap.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::AlphaXOrigin, value_list.alphaxorigin.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::AlphaYOrigin, value_list.alphayorigin.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::ClipXOrigin, value_list.clipxorigin.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::ClipYOrigin, value_list.clipyorigin.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::ClipMask, value_list.clipmask.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::GraphicsExposure, value_list.graphicsexposure.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::SubwindowMode, value_list.subwindowmode.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::PolyEdge, value_list.polyedge.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::PolyMode, value_list.polymode.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::Dither, value_list.dither.has_value(), true, &value_mask);
  SwitchVar(CreatePictureAttribute::ComponentAlpha, value_list.componentalpha.has_value(), true, &value_mask);
  uint32_t tmp7;
  tmp7 = static_cast<uint32_t>(value_mask);
  Write(&tmp7, &buf);

  // value_list
  auto value_list_expr = value_mask;
  if (CaseAnd(value_list_expr, CreatePictureAttribute::Repeat)) {
    auto& repeat = *value_list.repeat;

    // repeat
    uint32_t tmp8;
    tmp8 = static_cast<uint32_t>(repeat);
    Write(&tmp8, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::AlphaMap)) {
    auto& alphamap = *value_list.alphamap;

    // alphamap
    Write(&alphamap, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::AlphaXOrigin)) {
    auto& alphaxorigin = *value_list.alphaxorigin;

    // alphaxorigin
    Write(&alphaxorigin, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::AlphaYOrigin)) {
    auto& alphayorigin = *value_list.alphayorigin;

    // alphayorigin
    Write(&alphayorigin, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::ClipXOrigin)) {
    auto& clipxorigin = *value_list.clipxorigin;

    // clipxorigin
    Write(&clipxorigin, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::ClipYOrigin)) {
    auto& clipyorigin = *value_list.clipyorigin;

    // clipyorigin
    Write(&clipyorigin, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::ClipMask)) {
    auto& clipmask = *value_list.clipmask;

    // clipmask
    Write(&clipmask, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::GraphicsExposure)) {
    auto& graphicsexposure = *value_list.graphicsexposure;

    // graphicsexposure
    Write(&graphicsexposure, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::SubwindowMode)) {
    auto& subwindowmode = *value_list.subwindowmode;

    // subwindowmode
    uint32_t tmp9;
    tmp9 = static_cast<uint32_t>(subwindowmode);
    Write(&tmp9, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::PolyEdge)) {
    auto& polyedge = *value_list.polyedge;

    // polyedge
    uint32_t tmp10;
    tmp10 = static_cast<uint32_t>(polyedge);
    Write(&tmp10, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::PolyMode)) {
    auto& polymode = *value_list.polymode;

    // polymode
    uint32_t tmp11;
    tmp11 = static_cast<uint32_t>(polymode);
    Write(&tmp11, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::Dither)) {
    auto& dither = *value_list.dither;

    // dither
    Write(&dither, &buf);

  }
  if (CaseAnd(value_list_expr, CreatePictureAttribute::ComponentAlpha)) {
    auto& componentalpha = *value_list.componentalpha;

    // componentalpha
    Write(&componentalpha, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::SetPictureClipRectangles(
    const Render::SetPictureClipRectanglesRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& picture = request.picture;
  auto& clip_x_origin = request.clip_x_origin;
  auto& clip_y_origin = request.clip_y_origin;
  auto& rectangles = request.rectangles;
  size_t rectangles_len = rectangles.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 6;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // picture
  Write(&picture, &buf);

  // clip_x_origin
  Write(&clip_x_origin, &buf);

  // clip_y_origin
  Write(&clip_y_origin, &buf);

  // rectangles
  DCHECK_EQ(static_cast<size_t>(rectangles_len), rectangles.size());
  for (auto& rectangles_elem : rectangles) {
    // rectangles_elem
    {
      auto& x = rectangles_elem.x;
      auto& y = rectangles_elem.y;
      auto& width = rectangles_elem.width;
      auto& height = rectangles_elem.height;

      // x
      Write(&x, &buf);

      // y
      Write(&y, &buf);

      // width
      Write(&width, &buf);

      // height
      Write(&height, &buf);

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::FreePicture(
    const Render::FreePictureRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& picture = request.picture;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 7;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // picture
  Write(&picture, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::Composite(
    const Render::CompositeRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& op = request.op;
  auto& src = request.src;
  auto& mask = request.mask;
  auto& dst = request.dst;
  auto& src_x = request.src_x;
  auto& src_y = request.src_y;
  auto& mask_x = request.mask_x;
  auto& mask_y = request.mask_y;
  auto& dst_x = request.dst_x;
  auto& dst_y = request.dst_y;
  auto& width = request.width;
  auto& height = request.height;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 8;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // op
  uint8_t tmp12;
  tmp12 = static_cast<uint8_t>(op);
  Write(&tmp12, &buf);

  // pad0
  Pad(&buf, 3);

  // src
  Write(&src, &buf);

  // mask
  Write(&mask, &buf);

  // dst
  Write(&dst, &buf);

  // src_x
  Write(&src_x, &buf);

  // src_y
  Write(&src_y, &buf);

  // mask_x
  Write(&mask_x, &buf);

  // mask_y
  Write(&mask_y, &buf);

  // dst_x
  Write(&dst_x, &buf);

  // dst_y
  Write(&dst_y, &buf);

  // width
  Write(&width, &buf);

  // height
  Write(&height, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::Trapezoids(
    const Render::TrapezoidsRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& op = request.op;
  auto& src = request.src;
  auto& dst = request.dst;
  auto& mask_format = request.mask_format;
  auto& src_x = request.src_x;
  auto& src_y = request.src_y;
  auto& traps = request.traps;
  size_t traps_len = traps.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 10;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // op
  uint8_t tmp13;
  tmp13 = static_cast<uint8_t>(op);
  Write(&tmp13, &buf);

  // pad0
  Pad(&buf, 3);

  // src
  Write(&src, &buf);

  // dst
  Write(&dst, &buf);

  // mask_format
  Write(&mask_format, &buf);

  // src_x
  Write(&src_x, &buf);

  // src_y
  Write(&src_y, &buf);

  // traps
  DCHECK_EQ(static_cast<size_t>(traps_len), traps.size());
  for (auto& traps_elem : traps) {
    // traps_elem
    {
      auto& top = traps_elem.top;
      auto& bottom = traps_elem.bottom;
      auto& left = traps_elem.left;
      auto& right = traps_elem.right;

      // top
      Write(&top, &buf);

      // bottom
      Write(&bottom, &buf);

      // left
      {
        auto& p1 = left.p1;
        auto& p2 = left.p2;

        // p1
        {
          auto& x = p1.x;
          auto& y = p1.y;

          // x
          Write(&x, &buf);

          // y
          Write(&y, &buf);

        }

        // p2
        {
          auto& x = p2.x;
          auto& y = p2.y;

          // x
          Write(&x, &buf);

          // y
          Write(&y, &buf);

        }

      }

      // right
      {
        auto& p1 = right.p1;
        auto& p2 = right.p2;

        // p1
        {
          auto& x = p1.x;
          auto& y = p1.y;

          // x
          Write(&x, &buf);

          // y
          Write(&y, &buf);

        }

        // p2
        {
          auto& x = p2.x;
          auto& y = p2.y;

          // x
          Write(&x, &buf);

          // y
          Write(&y, &buf);

        }

      }

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::Triangles(
    const Render::TrianglesRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& op = request.op;
  auto& src = request.src;
  auto& dst = request.dst;
  auto& mask_format = request.mask_format;
  auto& src_x = request.src_x;
  auto& src_y = request.src_y;
  auto& triangles = request.triangles;
  size_t triangles_len = triangles.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 11;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // op
  uint8_t tmp14;
  tmp14 = static_cast<uint8_t>(op);
  Write(&tmp14, &buf);

  // pad0
  Pad(&buf, 3);

  // src
  Write(&src, &buf);

  // dst
  Write(&dst, &buf);

  // mask_format
  Write(&mask_format, &buf);

  // src_x
  Write(&src_x, &buf);

  // src_y
  Write(&src_y, &buf);

  // triangles
  DCHECK_EQ(static_cast<size_t>(triangles_len), triangles.size());
  for (auto& triangles_elem : triangles) {
    // triangles_elem
    {
      auto& p1 = triangles_elem.p1;
      auto& p2 = triangles_elem.p2;
      auto& p3 = triangles_elem.p3;

      // p1
      {
        auto& x = p1.x;
        auto& y = p1.y;

        // x
        Write(&x, &buf);

        // y
        Write(&y, &buf);

      }

      // p2
      {
        auto& x = p2.x;
        auto& y = p2.y;

        // x
        Write(&x, &buf);

        // y
        Write(&y, &buf);

      }

      // p3
      {
        auto& x = p3.x;
        auto& y = p3.y;

        // x
        Write(&x, &buf);

        // y
        Write(&y, &buf);

      }

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::TriStrip(
    const Render::TriStripRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& op = request.op;
  auto& src = request.src;
  auto& dst = request.dst;
  auto& mask_format = request.mask_format;
  auto& src_x = request.src_x;
  auto& src_y = request.src_y;
  auto& points = request.points;
  size_t points_len = points.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 12;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // op
  uint8_t tmp15;
  tmp15 = static_cast<uint8_t>(op);
  Write(&tmp15, &buf);

  // pad0
  Pad(&buf, 3);

  // src
  Write(&src, &buf);

  // dst
  Write(&dst, &buf);

  // mask_format
  Write(&mask_format, &buf);

  // src_x
  Write(&src_x, &buf);

  // src_y
  Write(&src_y, &buf);

  // points
  DCHECK_EQ(static_cast<size_t>(points_len), points.size());
  for (auto& points_elem : points) {
    // points_elem
    {
      auto& x = points_elem.x;
      auto& y = points_elem.y;

      // x
      Write(&x, &buf);

      // y
      Write(&y, &buf);

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::TriFan(
    const Render::TriFanRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& op = request.op;
  auto& src = request.src;
  auto& dst = request.dst;
  auto& mask_format = request.mask_format;
  auto& src_x = request.src_x;
  auto& src_y = request.src_y;
  auto& points = request.points;
  size_t points_len = points.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 13;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // op
  uint8_t tmp16;
  tmp16 = static_cast<uint8_t>(op);
  Write(&tmp16, &buf);

  // pad0
  Pad(&buf, 3);

  // src
  Write(&src, &buf);

  // dst
  Write(&dst, &buf);

  // mask_format
  Write(&mask_format, &buf);

  // src_x
  Write(&src_x, &buf);

  // src_y
  Write(&src_y, &buf);

  // points
  DCHECK_EQ(static_cast<size_t>(points_len), points.size());
  for (auto& points_elem : points) {
    // points_elem
    {
      auto& x = points_elem.x;
      auto& y = points_elem.y;

      // x
      Write(&x, &buf);

      // y
      Write(&y, &buf);

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::CreateGlyphSet(
    const Render::CreateGlyphSetRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& gsid = request.gsid;
  auto& format = request.format;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 17;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // gsid
  Write(&gsid, &buf);

  // format
  Write(&format, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::ReferenceGlyphSet(
    const Render::ReferenceGlyphSetRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& gsid = request.gsid;
  auto& existing = request.existing;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 18;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // gsid
  Write(&gsid, &buf);

  // existing
  Write(&existing, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::FreeGlyphSet(
    const Render::FreeGlyphSetRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& glyphset = request.glyphset;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 19;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // glyphset
  Write(&glyphset, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::AddGlyphs(
    const Render::AddGlyphsRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& glyphset = request.glyphset;
  uint32_t glyphs_len{};
  auto& glyphids = request.glyphids;
  size_t glyphids_len = glyphids.size();
  auto& glyphs = request.glyphs;
  auto& data = request.data;
  size_t data_len = data.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 20;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // glyphset
  Write(&glyphset, &buf);

  // glyphs_len
  glyphs_len = glyphs.size();
  Write(&glyphs_len, &buf);

  // glyphids
  DCHECK_EQ(static_cast<size_t>(glyphs_len), glyphids.size());
  for (auto& glyphids_elem : glyphids) {
    // glyphids_elem
    Write(&glyphids_elem, &buf);

  }

  // glyphs
  DCHECK_EQ(static_cast<size_t>(glyphs_len), glyphs.size());
  for (auto& glyphs_elem : glyphs) {
    // glyphs_elem
    {
      auto& width = glyphs_elem.width;
      auto& height = glyphs_elem.height;
      auto& x = glyphs_elem.x;
      auto& y = glyphs_elem.y;
      auto& x_off = glyphs_elem.x_off;
      auto& y_off = glyphs_elem.y_off;

      // width
      Write(&width, &buf);

      // height
      Write(&height, &buf);

      // x
      Write(&x, &buf);

      // y
      Write(&y, &buf);

      // x_off
      Write(&x_off, &buf);

      // y_off
      Write(&y_off, &buf);

    }

  }

  // data
  DCHECK_EQ(static_cast<size_t>(data_len), data.size());
  for (auto& data_elem : data) {
    // data_elem
    Write(&data_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::FreeGlyphs(
    const Render::FreeGlyphsRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& glyphset = request.glyphset;
  auto& glyphs = request.glyphs;
  size_t glyphs_len = glyphs.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 22;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // glyphset
  Write(&glyphset, &buf);

  // glyphs
  DCHECK_EQ(static_cast<size_t>(glyphs_len), glyphs.size());
  for (auto& glyphs_elem : glyphs) {
    // glyphs_elem
    Write(&glyphs_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::CompositeGlyphs8(
    const Render::CompositeGlyphs8Request& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& op = request.op;
  auto& src = request.src;
  auto& dst = request.dst;
  auto& mask_format = request.mask_format;
  auto& glyphset = request.glyphset;
  auto& src_x = request.src_x;
  auto& src_y = request.src_y;
  auto& glyphcmds = request.glyphcmds;
  size_t glyphcmds_len = glyphcmds.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 23;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // op
  uint8_t tmp17;
  tmp17 = static_cast<uint8_t>(op);
  Write(&tmp17, &buf);

  // pad0
  Pad(&buf, 3);

  // src
  Write(&src, &buf);

  // dst
  Write(&dst, &buf);

  // mask_format
  Write(&mask_format, &buf);

  // glyphset
  Write(&glyphset, &buf);

  // src_x
  Write(&src_x, &buf);

  // src_y
  Write(&src_y, &buf);

  // glyphcmds
  DCHECK_EQ(static_cast<size_t>(glyphcmds_len), glyphcmds.size());
  for (auto& glyphcmds_elem : glyphcmds) {
    // glyphcmds_elem
    Write(&glyphcmds_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::CompositeGlyphs16(
    const Render::CompositeGlyphs16Request& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& op = request.op;
  auto& src = request.src;
  auto& dst = request.dst;
  auto& mask_format = request.mask_format;
  auto& glyphset = request.glyphset;
  auto& src_x = request.src_x;
  auto& src_y = request.src_y;
  auto& glyphcmds = request.glyphcmds;
  size_t glyphcmds_len = glyphcmds.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 24;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // op
  uint8_t tmp18;
  tmp18 = static_cast<uint8_t>(op);
  Write(&tmp18, &buf);

  // pad0
  Pad(&buf, 3);

  // src
  Write(&src, &buf);

  // dst
  Write(&dst, &buf);

  // mask_format
  Write(&mask_format, &buf);

  // glyphset
  Write(&glyphset, &buf);

  // src_x
  Write(&src_x, &buf);

  // src_y
  Write(&src_y, &buf);

  // glyphcmds
  DCHECK_EQ(static_cast<size_t>(glyphcmds_len), glyphcmds.size());
  for (auto& glyphcmds_elem : glyphcmds) {
    // glyphcmds_elem
    Write(&glyphcmds_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::CompositeGlyphs32(
    const Render::CompositeGlyphs32Request& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& op = request.op;
  auto& src = request.src;
  auto& dst = request.dst;
  auto& mask_format = request.mask_format;
  auto& glyphset = request.glyphset;
  auto& src_x = request.src_x;
  auto& src_y = request.src_y;
  auto& glyphcmds = request.glyphcmds;
  size_t glyphcmds_len = glyphcmds.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 25;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // op
  uint8_t tmp19;
  tmp19 = static_cast<uint8_t>(op);
  Write(&tmp19, &buf);

  // pad0
  Pad(&buf, 3);

  // src
  Write(&src, &buf);

  // dst
  Write(&dst, &buf);

  // mask_format
  Write(&mask_format, &buf);

  // glyphset
  Write(&glyphset, &buf);

  // src_x
  Write(&src_x, &buf);

  // src_y
  Write(&src_y, &buf);

  // glyphcmds
  DCHECK_EQ(static_cast<size_t>(glyphcmds_len), glyphcmds.size());
  for (auto& glyphcmds_elem : glyphcmds) {
    // glyphcmds_elem
    Write(&glyphcmds_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::FillRectangles(
    const Render::FillRectanglesRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& op = request.op;
  auto& dst = request.dst;
  auto& color = request.color;
  auto& rects = request.rects;
  size_t rects_len = rects.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 26;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // op
  uint8_t tmp20;
  tmp20 = static_cast<uint8_t>(op);
  Write(&tmp20, &buf);

  // pad0
  Pad(&buf, 3);

  // dst
  Write(&dst, &buf);

  // color
  {
    auto& red = color.red;
    auto& green = color.green;
    auto& blue = color.blue;
    auto& alpha = color.alpha;

    // red
    Write(&red, &buf);

    // green
    Write(&green, &buf);

    // blue
    Write(&blue, &buf);

    // alpha
    Write(&alpha, &buf);

  }

  // rects
  DCHECK_EQ(static_cast<size_t>(rects_len), rects.size());
  for (auto& rects_elem : rects) {
    // rects_elem
    {
      auto& x = rects_elem.x;
      auto& y = rects_elem.y;
      auto& width = rects_elem.width;
      auto& height = rects_elem.height;

      // x
      Write(&x, &buf);

      // y
      Write(&y, &buf);

      // width
      Write(&width, &buf);

      // height
      Write(&height, &buf);

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::CreateCursor(
    const Render::CreateCursorRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& cid = request.cid;
  auto& source = request.source;
  auto& x = request.x;
  auto& y = request.y;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 27;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cid
  Write(&cid, &buf);

  // source
  Write(&source, &buf);

  // x
  Write(&x, &buf);

  // y
  Write(&y, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::SetPictureTransform(
    const Render::SetPictureTransformRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& picture = request.picture;
  auto& transform = request.transform;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 28;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // picture
  Write(&picture, &buf);

  // transform
  {
    auto& matrix11 = transform.matrix11;
    auto& matrix12 = transform.matrix12;
    auto& matrix13 = transform.matrix13;
    auto& matrix21 = transform.matrix21;
    auto& matrix22 = transform.matrix22;
    auto& matrix23 = transform.matrix23;
    auto& matrix31 = transform.matrix31;
    auto& matrix32 = transform.matrix32;
    auto& matrix33 = transform.matrix33;

    // matrix11
    Write(&matrix11, &buf);

    // matrix12
    Write(&matrix12, &buf);

    // matrix13
    Write(&matrix13, &buf);

    // matrix21
    Write(&matrix21, &buf);

    // matrix22
    Write(&matrix22, &buf);

    // matrix23
    Write(&matrix23, &buf);

    // matrix31
    Write(&matrix31, &buf);

    // matrix32
    Write(&matrix32, &buf);

    // matrix33
    Write(&matrix33, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<Render::QueryFiltersReply>
Render::QueryFilters(
    const Render::QueryFiltersRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& drawable = request.drawable;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 29;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  Align(&buf, 4);

  return x11::SendRequest<Render::QueryFiltersReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<Render::QueryFiltersReply>
detail::ReadReply<Render::QueryFiltersReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<Render::QueryFiltersReply>();

  auto& sequence = (*reply).sequence;
  uint32_t num_aliases{};
  uint32_t num_filters{};
  auto& aliases = (*reply).aliases;
  size_t aliases_len = aliases.size();
  auto& filters = (*reply).filters;
  size_t filters_len = filters.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // num_aliases
  Read(&num_aliases, &buf);

  // num_filters
  Read(&num_filters, &buf);

  // pad1
  Pad(&buf, 16);

  // aliases
  aliases.resize(num_aliases);
  for (auto& aliases_elem : aliases) {
    // aliases_elem
    Read(&aliases_elem, &buf);

  }

  // filters
  filters.resize(num_filters);
  for (auto& filters_elem : filters) {
    // filters_elem
    {
      uint8_t name_len{};
      auto& name = filters_elem.name;

      // name_len
      Read(&name_len, &buf);

      // name
      name.resize(name_len);
      for (auto& name_elem : name) {
        // name_elem
        Read(&name_elem, &buf);

      }

    }

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
Render::SetPictureFilter(
    const Render::SetPictureFilterRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& picture = request.picture;
  uint16_t filter_len{};
  auto& filter = request.filter;
  auto& values = request.values;
  size_t values_len = values.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 30;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // picture
  Write(&picture, &buf);

  // filter_len
  filter_len = filter.size();
  Write(&filter_len, &buf);

  // pad0
  Pad(&buf, 2);

  // filter
  DCHECK_EQ(static_cast<size_t>(filter_len), filter.size());
  for (auto& filter_elem : filter) {
    // filter_elem
    Write(&filter_elem, &buf);

  }

  // pad1
  Align(&buf, 4);

  // values
  DCHECK_EQ(static_cast<size_t>(values_len), values.size());
  for (auto& values_elem : values) {
    // values_elem
    Write(&values_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::CreateAnimCursor(
    const Render::CreateAnimCursorRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& cid = request.cid;
  auto& cursors = request.cursors;
  size_t cursors_len = cursors.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 31;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cid
  Write(&cid, &buf);

  // cursors
  DCHECK_EQ(static_cast<size_t>(cursors_len), cursors.size());
  for (auto& cursors_elem : cursors) {
    // cursors_elem
    {
      auto& cursor = cursors_elem.cursor;
      auto& delay = cursors_elem.delay;

      // cursor
      Write(&cursor, &buf);

      // delay
      Write(&delay, &buf);

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::AddTraps(
    const Render::AddTrapsRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& picture = request.picture;
  auto& x_off = request.x_off;
  auto& y_off = request.y_off;
  auto& traps = request.traps;
  size_t traps_len = traps.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 32;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // picture
  Write(&picture, &buf);

  // x_off
  Write(&x_off, &buf);

  // y_off
  Write(&y_off, &buf);

  // traps
  DCHECK_EQ(static_cast<size_t>(traps_len), traps.size());
  for (auto& traps_elem : traps) {
    // traps_elem
    {
      auto& top = traps_elem.top;
      auto& bot = traps_elem.bot;

      // top
      {
        auto& l = top.l;
        auto& r = top.r;
        auto& y = top.y;

        // l
        Write(&l, &buf);

        // r
        Write(&r, &buf);

        // y
        Write(&y, &buf);

      }

      // bot
      {
        auto& l = bot.l;
        auto& r = bot.r;
        auto& y = bot.y;

        // l
        Write(&l, &buf);

        // r
        Write(&r, &buf);

        // y
        Write(&y, &buf);

      }

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::CreateSolidFill(
    const Render::CreateSolidFillRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& picture = request.picture;
  auto& color = request.color;

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 33;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // picture
  Write(&picture, &buf);

  // color
  {
    auto& red = color.red;
    auto& green = color.green;
    auto& blue = color.blue;
    auto& alpha = color.alpha;

    // red
    Write(&red, &buf);

    // green
    Write(&green, &buf);

    // blue
    Write(&blue, &buf);

    // alpha
    Write(&alpha, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::CreateLinearGradient(
    const Render::CreateLinearGradientRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& picture = request.picture;
  auto& p1 = request.p1;
  auto& p2 = request.p2;
  uint32_t num_stops{};
  auto& stops = request.stops;
  size_t stops_len = stops.size();
  auto& colors = request.colors;
  size_t colors_len = colors.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 34;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // picture
  Write(&picture, &buf);

  // p1
  {
    auto& x = p1.x;
    auto& y = p1.y;

    // x
    Write(&x, &buf);

    // y
    Write(&y, &buf);

  }

  // p2
  {
    auto& x = p2.x;
    auto& y = p2.y;

    // x
    Write(&x, &buf);

    // y
    Write(&y, &buf);

  }

  // num_stops
  num_stops = stops.size();
  Write(&num_stops, &buf);

  // stops
  DCHECK_EQ(static_cast<size_t>(num_stops), stops.size());
  for (auto& stops_elem : stops) {
    // stops_elem
    Write(&stops_elem, &buf);

  }

  // colors
  DCHECK_EQ(static_cast<size_t>(num_stops), colors.size());
  for (auto& colors_elem : colors) {
    // colors_elem
    {
      auto& red = colors_elem.red;
      auto& green = colors_elem.green;
      auto& blue = colors_elem.blue;
      auto& alpha = colors_elem.alpha;

      // red
      Write(&red, &buf);

      // green
      Write(&green, &buf);

      // blue
      Write(&blue, &buf);

      // alpha
      Write(&alpha, &buf);

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::CreateRadialGradient(
    const Render::CreateRadialGradientRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& picture = request.picture;
  auto& inner = request.inner;
  auto& outer = request.outer;
  auto& inner_radius = request.inner_radius;
  auto& outer_radius = request.outer_radius;
  uint32_t num_stops{};
  auto& stops = request.stops;
  size_t stops_len = stops.size();
  auto& colors = request.colors;
  size_t colors_len = colors.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 35;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // picture
  Write(&picture, &buf);

  // inner
  {
    auto& x = inner.x;
    auto& y = inner.y;

    // x
    Write(&x, &buf);

    // y
    Write(&y, &buf);

  }

  // outer
  {
    auto& x = outer.x;
    auto& y = outer.y;

    // x
    Write(&x, &buf);

    // y
    Write(&y, &buf);

  }

  // inner_radius
  Write(&inner_radius, &buf);

  // outer_radius
  Write(&outer_radius, &buf);

  // num_stops
  num_stops = stops.size();
  Write(&num_stops, &buf);

  // stops
  DCHECK_EQ(static_cast<size_t>(num_stops), stops.size());
  for (auto& stops_elem : stops) {
    // stops_elem
    Write(&stops_elem, &buf);

  }

  // colors
  DCHECK_EQ(static_cast<size_t>(num_stops), colors.size());
  for (auto& colors_elem : colors) {
    // colors_elem
    {
      auto& red = colors_elem.red;
      auto& green = colors_elem.green;
      auto& blue = colors_elem.blue;
      auto& alpha = colors_elem.alpha;

      // red
      Write(&red, &buf);

      // green
      Write(&green, &buf);

      // blue
      Write(&blue, &buf);

      // alpha
      Write(&alpha, &buf);

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
Render::CreateConicalGradient(
    const Render::CreateConicalGradientRequest& request) {
  if (!connection_->Ready() || !present())
    return {};

  WriteBuffer buf;

  auto& picture = request.picture;
  auto& center = request.center;
  auto& angle = request.angle;
  uint32_t num_stops{};
  auto& stops = request.stops;
  size_t stops_len = stops.size();
  auto& colors = request.colors;
  size_t colors_len = colors.size();

  // major_opcode
  uint8_t major_opcode = info_.major_opcode;
  Write(&major_opcode, &buf);

  // minor_opcode
  uint8_t minor_opcode = 36;
  Write(&minor_opcode, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // picture
  Write(&picture, &buf);

  // center
  {
    auto& x = center.x;
    auto& y = center.y;

    // x
    Write(&x, &buf);

    // y
    Write(&y, &buf);

  }

  // angle
  Write(&angle, &buf);

  // num_stops
  num_stops = stops.size();
  Write(&num_stops, &buf);

  // stops
  DCHECK_EQ(static_cast<size_t>(num_stops), stops.size());
  for (auto& stops_elem : stops) {
    // stops_elem
    Write(&stops_elem, &buf);

  }

  // colors
  DCHECK_EQ(static_cast<size_t>(num_stops), colors.size());
  for (auto& colors_elem : colors) {
    // colors_elem
    {
      auto& red = colors_elem.red;
      auto& green = colors_elem.green;
      auto& blue = colors_elem.blue;
      auto& alpha = colors_elem.alpha;

      // red
      Write(&red, &buf);

      // green
      Write(&green, &buf);

      // blue
      Write(&blue, &buf);

      // alpha
      Write(&alpha, &buf);

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

}  // namespace x11
