#include "xproto.h"

#include <xcb/xcb.h>
#include <xcb/xcbext.h>

#include "base/logging.h"
#include "ui/gfx/x/xproto_internal.h"

namespace x11 {

XProto::XProto(Connection* connection) : connection_(connection) {}

template <> COMPONENT_EXPORT(X11)
Setup Read<Setup>(
    const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  Setup obj;

  auto& status = obj.status;
  auto& protocol_major_version = obj.protocol_major_version;
  auto& protocol_minor_version = obj.protocol_minor_version;
  auto& length = obj.length;
  auto& release_number = obj.release_number;
  auto& resource_id_base = obj.resource_id_base;
  auto& resource_id_mask = obj.resource_id_mask;
  auto& motion_buffer_size = obj.motion_buffer_size;
  uint16_t vendor_len{};
  auto& maximum_request_length = obj.maximum_request_length;
  uint8_t roots_len{};
  uint8_t pixmap_formats_len{};
  auto& image_byte_order = obj.image_byte_order;
  auto& bitmap_format_bit_order = obj.bitmap_format_bit_order;
  auto& bitmap_format_scanline_unit = obj.bitmap_format_scanline_unit;
  auto& bitmap_format_scanline_pad = obj.bitmap_format_scanline_pad;
  auto& min_keycode = obj.min_keycode;
  auto& max_keycode = obj.max_keycode;
  auto& vendor = obj.vendor;
  auto& pixmap_formats = obj.pixmap_formats;
  auto& roots = obj.roots;

  // status
  Read(&status, &buf);

  // pad0
  Pad(&buf, 1);

  // protocol_major_version
  Read(&protocol_major_version, &buf);

  // protocol_minor_version
  Read(&protocol_minor_version, &buf);

  // length
  Read(&length, &buf);

  // release_number
  Read(&release_number, &buf);

  // resource_id_base
  Read(&resource_id_base, &buf);

  // resource_id_mask
  Read(&resource_id_mask, &buf);

  // motion_buffer_size
  Read(&motion_buffer_size, &buf);

  // vendor_len
  Read(&vendor_len, &buf);

  // maximum_request_length
  Read(&maximum_request_length, &buf);

  // roots_len
  Read(&roots_len, &buf);

  // pixmap_formats_len
  Read(&pixmap_formats_len, &buf);

  // image_byte_order
  uint8_t tmp0;
  Read(&tmp0, &buf);
  image_byte_order = static_cast<ImageOrder>(tmp0);

  // bitmap_format_bit_order
  uint8_t tmp1;
  Read(&tmp1, &buf);
  bitmap_format_bit_order = static_cast<ImageOrder>(tmp1);

  // bitmap_format_scanline_unit
  Read(&bitmap_format_scanline_unit, &buf);

  // bitmap_format_scanline_pad
  Read(&bitmap_format_scanline_pad, &buf);

  // min_keycode
  Read(&min_keycode, &buf);

  // max_keycode
  Read(&max_keycode, &buf);

  // pad1
  Pad(&buf, 4);

  // vendor
  vendor.resize(vendor_len);
  for (auto& vendor_elem : vendor) {
    // vendor_elem
    Read(&vendor_elem, &buf);

  }

  // pad2
  Align(&buf, 4);

  // pixmap_formats
  pixmap_formats.resize(pixmap_formats_len);
  for (auto& pixmap_formats_elem : pixmap_formats) {
    // pixmap_formats_elem
    {
      auto& depth = pixmap_formats_elem.depth;
      auto& bits_per_pixel = pixmap_formats_elem.bits_per_pixel;
      auto& scanline_pad = pixmap_formats_elem.scanline_pad;

      // depth
      Read(&depth, &buf);

      // bits_per_pixel
      Read(&bits_per_pixel, &buf);

      // scanline_pad
      Read(&scanline_pad, &buf);

      // pad0
      Pad(&buf, 5);

    }

  }

  // roots
  roots.resize(roots_len);
  for (auto& roots_elem : roots) {
    // roots_elem
    {
      auto& root = roots_elem.root;
      auto& default_colormap = roots_elem.default_colormap;
      auto& white_pixel = roots_elem.white_pixel;
      auto& black_pixel = roots_elem.black_pixel;
      auto& current_input_masks = roots_elem.current_input_masks;
      auto& width_in_pixels = roots_elem.width_in_pixels;
      auto& height_in_pixels = roots_elem.height_in_pixels;
      auto& width_in_millimeters = roots_elem.width_in_millimeters;
      auto& height_in_millimeters = roots_elem.height_in_millimeters;
      auto& min_installed_maps = roots_elem.min_installed_maps;
      auto& max_installed_maps = roots_elem.max_installed_maps;
      auto& root_visual = roots_elem.root_visual;
      auto& backing_stores = roots_elem.backing_stores;
      auto& save_unders = roots_elem.save_unders;
      auto& root_depth = roots_elem.root_depth;
      uint8_t allowed_depths_len{};
      auto& allowed_depths = roots_elem.allowed_depths;

      // root
      Read(&root, &buf);

      // default_colormap
      Read(&default_colormap, &buf);

      // white_pixel
      Read(&white_pixel, &buf);

      // black_pixel
      Read(&black_pixel, &buf);

      // current_input_masks
      uint32_t tmp2;
      Read(&tmp2, &buf);
      current_input_masks = static_cast<EventMask>(tmp2);

      // width_in_pixels
      Read(&width_in_pixels, &buf);

      // height_in_pixels
      Read(&height_in_pixels, &buf);

      // width_in_millimeters
      Read(&width_in_millimeters, &buf);

      // height_in_millimeters
      Read(&height_in_millimeters, &buf);

      // min_installed_maps
      Read(&min_installed_maps, &buf);

      // max_installed_maps
      Read(&max_installed_maps, &buf);

      // root_visual
      Read(&root_visual, &buf);

      // backing_stores
      uint8_t tmp3;
      Read(&tmp3, &buf);
      backing_stores = static_cast<BackingStore>(tmp3);

      // save_unders
      Read(&save_unders, &buf);

      // root_depth
      Read(&root_depth, &buf);

      // allowed_depths_len
      Read(&allowed_depths_len, &buf);

      // allowed_depths
      allowed_depths.resize(allowed_depths_len);
      for (auto& allowed_depths_elem : allowed_depths) {
        // allowed_depths_elem
        {
          auto& depth = allowed_depths_elem.depth;
          uint16_t visuals_len{};
          auto& visuals = allowed_depths_elem.visuals;

          // depth
          Read(&depth, &buf);

          // pad0
          Pad(&buf, 1);

          // visuals_len
          Read(&visuals_len, &buf);

          // pad1
          Pad(&buf, 4);

          // visuals
          visuals.resize(visuals_len);
          for (auto& visuals_elem : visuals) {
            // visuals_elem
            {
              auto& visual_id = visuals_elem.visual_id;
              auto& c_class = visuals_elem.c_class;
              auto& bits_per_rgb_value = visuals_elem.bits_per_rgb_value;
              auto& colormap_entries = visuals_elem.colormap_entries;
              auto& red_mask = visuals_elem.red_mask;
              auto& green_mask = visuals_elem.green_mask;
              auto& blue_mask = visuals_elem.blue_mask;

              // visual_id
              Read(&visual_id, &buf);

              // c_class
              uint8_t tmp4;
              Read(&tmp4, &buf);
              c_class = static_cast<VisualClass>(tmp4);

              // bits_per_rgb_value
              Read(&bits_per_rgb_value, &buf);

              // colormap_entries
              Read(&colormap_entries, &buf);

              // red_mask
              Read(&red_mask, &buf);

              // green_mask
              Read(&green_mask, &buf);

              // blue_mask
              Read(&blue_mask, &buf);

              // pad0
              Pad(&buf, 4);

            }

          }

        }

      }

    }

  }

  return obj;
}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<KeyEvent>(
    KeyEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& detail = (*event_).detail;
  auto& sequence = (*event_).sequence;
  auto& time = (*event_).time;
  auto& root = (*event_).root;
  auto& event = (*event_).event;
  auto& child = (*event_).child;
  auto& root_x = (*event_).root_x;
  auto& root_y = (*event_).root_y;
  auto& event_x = (*event_).event_x;
  auto& event_y = (*event_).event_y;
  auto& state = (*event_).state;
  auto& same_screen = (*event_).same_screen;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // detail
  Read(&detail, &buf);

  // sequence
  Read(&sequence, &buf);

  // time
  Read(&time, &buf);

  // root
  Read(&root, &buf);

  // event
  Read(&event, &buf);

  // child
  Read(&child, &buf);

  // root_x
  Read(&root_x, &buf);

  // root_y
  Read(&root_y, &buf);

  // event_x
  Read(&event_x, &buf);

  // event_y
  Read(&event_y, &buf);

  // state
  uint16_t tmp5;
  Read(&tmp5, &buf);
  state = static_cast<KeyButMask>(tmp5);

  // same_screen
  Read(&same_screen, &buf);

  // pad0
  Pad(&buf, 1);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<ButtonEvent>(
    ButtonEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& detail = (*event_).detail;
  auto& sequence = (*event_).sequence;
  auto& time = (*event_).time;
  auto& root = (*event_).root;
  auto& event = (*event_).event;
  auto& child = (*event_).child;
  auto& root_x = (*event_).root_x;
  auto& root_y = (*event_).root_y;
  auto& event_x = (*event_).event_x;
  auto& event_y = (*event_).event_y;
  auto& state = (*event_).state;
  auto& same_screen = (*event_).same_screen;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // detail
  Read(&detail, &buf);

  // sequence
  Read(&sequence, &buf);

  // time
  Read(&time, &buf);

  // root
  Read(&root, &buf);

  // event
  Read(&event, &buf);

  // child
  Read(&child, &buf);

  // root_x
  Read(&root_x, &buf);

  // root_y
  Read(&root_y, &buf);

  // event_x
  Read(&event_x, &buf);

  // event_y
  Read(&event_y, &buf);

  // state
  uint16_t tmp6;
  Read(&tmp6, &buf);
  state = static_cast<KeyButMask>(tmp6);

  // same_screen
  Read(&same_screen, &buf);

  // pad0
  Pad(&buf, 1);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<MotionNotifyEvent>(
    MotionNotifyEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& detail = (*event_).detail;
  auto& sequence = (*event_).sequence;
  auto& time = (*event_).time;
  auto& root = (*event_).root;
  auto& event = (*event_).event;
  auto& child = (*event_).child;
  auto& root_x = (*event_).root_x;
  auto& root_y = (*event_).root_y;
  auto& event_x = (*event_).event_x;
  auto& event_y = (*event_).event_y;
  auto& state = (*event_).state;
  auto& same_screen = (*event_).same_screen;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // detail
  uint8_t tmp7;
  Read(&tmp7, &buf);
  detail = static_cast<Motion>(tmp7);

  // sequence
  Read(&sequence, &buf);

  // time
  Read(&time, &buf);

  // root
  Read(&root, &buf);

  // event
  Read(&event, &buf);

  // child
  Read(&child, &buf);

  // root_x
  Read(&root_x, &buf);

  // root_y
  Read(&root_y, &buf);

  // event_x
  Read(&event_x, &buf);

  // event_y
  Read(&event_y, &buf);

  // state
  uint16_t tmp8;
  Read(&tmp8, &buf);
  state = static_cast<KeyButMask>(tmp8);

  // same_screen
  Read(&same_screen, &buf);

  // pad0
  Pad(&buf, 1);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<CrossingEvent>(
    CrossingEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& detail = (*event_).detail;
  auto& sequence = (*event_).sequence;
  auto& time = (*event_).time;
  auto& root = (*event_).root;
  auto& event = (*event_).event;
  auto& child = (*event_).child;
  auto& root_x = (*event_).root_x;
  auto& root_y = (*event_).root_y;
  auto& event_x = (*event_).event_x;
  auto& event_y = (*event_).event_y;
  auto& state = (*event_).state;
  auto& mode = (*event_).mode;
  auto& same_screen_focus = (*event_).same_screen_focus;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // detail
  uint8_t tmp9;
  Read(&tmp9, &buf);
  detail = static_cast<NotifyDetail>(tmp9);

  // sequence
  Read(&sequence, &buf);

  // time
  Read(&time, &buf);

  // root
  Read(&root, &buf);

  // event
  Read(&event, &buf);

  // child
  Read(&child, &buf);

  // root_x
  Read(&root_x, &buf);

  // root_y
  Read(&root_y, &buf);

  // event_x
  Read(&event_x, &buf);

  // event_y
  Read(&event_y, &buf);

  // state
  uint16_t tmp10;
  Read(&tmp10, &buf);
  state = static_cast<KeyButMask>(tmp10);

  // mode
  uint8_t tmp11;
  Read(&tmp11, &buf);
  mode = static_cast<NotifyMode>(tmp11);

  // same_screen_focus
  Read(&same_screen_focus, &buf);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<FocusEvent>(
    FocusEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& detail = (*event_).detail;
  auto& sequence = (*event_).sequence;
  auto& event = (*event_).event;
  auto& mode = (*event_).mode;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // detail
  uint8_t tmp12;
  Read(&tmp12, &buf);
  detail = static_cast<NotifyDetail>(tmp12);

  // sequence
  Read(&sequence, &buf);

  // event
  Read(&event, &buf);

  // mode
  uint8_t tmp13;
  Read(&tmp13, &buf);
  mode = static_cast<NotifyMode>(tmp13);

  // pad0
  Pad(&buf, 3);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<KeymapNotifyEvent>(
    KeymapNotifyEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& keys = (*event_).keys;
  size_t keys_len = keys.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // keys
  for (auto& keys_elem : keys) {
    // keys_elem
    Read(&keys_elem, &buf);

  }

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<ExposeEvent>(
    ExposeEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& window = (*event_).window;
  auto& x = (*event_).x;
  auto& y = (*event_).y;
  auto& width = (*event_).width;
  auto& height = (*event_).height;
  auto& count = (*event_).count;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // window
  Read(&window, &buf);

  // x
  Read(&x, &buf);

  // y
  Read(&y, &buf);

  // width
  Read(&width, &buf);

  // height
  Read(&height, &buf);

  // count
  Read(&count, &buf);

  // pad1
  Pad(&buf, 2);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<GraphicsExposureEvent>(
    GraphicsExposureEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& drawable = (*event_).drawable;
  auto& x = (*event_).x;
  auto& y = (*event_).y;
  auto& width = (*event_).width;
  auto& height = (*event_).height;
  auto& minor_opcode = (*event_).minor_opcode;
  auto& count = (*event_).count;
  auto& major_opcode = (*event_).major_opcode;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // drawable
  Read(&drawable, &buf);

  // x
  Read(&x, &buf);

  // y
  Read(&y, &buf);

  // width
  Read(&width, &buf);

  // height
  Read(&height, &buf);

  // minor_opcode
  Read(&minor_opcode, &buf);

  // count
  Read(&count, &buf);

  // major_opcode
  Read(&major_opcode, &buf);

  // pad1
  Pad(&buf, 3);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<NoExposureEvent>(
    NoExposureEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& drawable = (*event_).drawable;
  auto& minor_opcode = (*event_).minor_opcode;
  auto& major_opcode = (*event_).major_opcode;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // drawable
  Read(&drawable, &buf);

  // minor_opcode
  Read(&minor_opcode, &buf);

  // major_opcode
  Read(&major_opcode, &buf);

  // pad1
  Pad(&buf, 1);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<VisibilityNotifyEvent>(
    VisibilityNotifyEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& window = (*event_).window;
  auto& state = (*event_).state;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // window
  Read(&window, &buf);

  // state
  uint8_t tmp14;
  Read(&tmp14, &buf);
  state = static_cast<Visibility>(tmp14);

  // pad1
  Pad(&buf, 3);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<CreateNotifyEvent>(
    CreateNotifyEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& parent = (*event_).parent;
  auto& window = (*event_).window;
  auto& x = (*event_).x;
  auto& y = (*event_).y;
  auto& width = (*event_).width;
  auto& height = (*event_).height;
  auto& border_width = (*event_).border_width;
  auto& override_redirect = (*event_).override_redirect;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // parent
  Read(&parent, &buf);

  // window
  Read(&window, &buf);

  // x
  Read(&x, &buf);

  // y
  Read(&y, &buf);

  // width
  Read(&width, &buf);

  // height
  Read(&height, &buf);

  // border_width
  Read(&border_width, &buf);

  // override_redirect
  Read(&override_redirect, &buf);

  // pad1
  Pad(&buf, 1);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<DestroyNotifyEvent>(
    DestroyNotifyEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& event = (*event_).event;
  auto& window = (*event_).window;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // event
  Read(&event, &buf);

  // window
  Read(&window, &buf);

}

template <> COMPONENT_EXPORT(X11)
std::vector<uint8_t> Write<UnmapNotifyEvent>(
    const UnmapNotifyEvent& obj) {
  WriteBuffer buf;

  auto& sequence = obj.sequence;
  auto& event = obj.event;
  auto& window = obj.window;
  auto& from_configure = obj.from_configure;

  // response_type
  uint8_t response_type = 18;
  Write(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Write(&sequence, &buf);

  // event
  Write(&event, &buf);

  // window
  Write(&window, &buf);

  // from_configure
  Write(&from_configure, &buf);

  // pad1
  Pad(&buf, 3);

  return buf;
}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<UnmapNotifyEvent>(
    UnmapNotifyEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& event = (*event_).event;
  auto& window = (*event_).window;
  auto& from_configure = (*event_).from_configure;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // event
  Read(&event, &buf);

  // window
  Read(&window, &buf);

  // from_configure
  Read(&from_configure, &buf);

  // pad1
  Pad(&buf, 3);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<MapNotifyEvent>(
    MapNotifyEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& event = (*event_).event;
  auto& window = (*event_).window;
  auto& override_redirect = (*event_).override_redirect;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // event
  Read(&event, &buf);

  // window
  Read(&window, &buf);

  // override_redirect
  Read(&override_redirect, &buf);

  // pad1
  Pad(&buf, 3);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<MapRequestEvent>(
    MapRequestEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& parent = (*event_).parent;
  auto& window = (*event_).window;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // parent
  Read(&parent, &buf);

  // window
  Read(&window, &buf);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<ReparentNotifyEvent>(
    ReparentNotifyEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& event = (*event_).event;
  auto& window = (*event_).window;
  auto& parent = (*event_).parent;
  auto& x = (*event_).x;
  auto& y = (*event_).y;
  auto& override_redirect = (*event_).override_redirect;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // event
  Read(&event, &buf);

  // window
  Read(&window, &buf);

  // parent
  Read(&parent, &buf);

  // x
  Read(&x, &buf);

  // y
  Read(&y, &buf);

  // override_redirect
  Read(&override_redirect, &buf);

  // pad1
  Pad(&buf, 3);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<ConfigureNotifyEvent>(
    ConfigureNotifyEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& event = (*event_).event;
  auto& window = (*event_).window;
  auto& above_sibling = (*event_).above_sibling;
  auto& x = (*event_).x;
  auto& y = (*event_).y;
  auto& width = (*event_).width;
  auto& height = (*event_).height;
  auto& border_width = (*event_).border_width;
  auto& override_redirect = (*event_).override_redirect;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // event
  Read(&event, &buf);

  // window
  Read(&window, &buf);

  // above_sibling
  Read(&above_sibling, &buf);

  // x
  Read(&x, &buf);

  // y
  Read(&y, &buf);

  // width
  Read(&width, &buf);

  // height
  Read(&height, &buf);

  // border_width
  Read(&border_width, &buf);

  // override_redirect
  Read(&override_redirect, &buf);

  // pad1
  Pad(&buf, 1);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<ConfigureRequestEvent>(
    ConfigureRequestEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& stack_mode = (*event_).stack_mode;
  auto& sequence = (*event_).sequence;
  auto& parent = (*event_).parent;
  auto& window = (*event_).window;
  auto& sibling = (*event_).sibling;
  auto& x = (*event_).x;
  auto& y = (*event_).y;
  auto& width = (*event_).width;
  auto& height = (*event_).height;
  auto& border_width = (*event_).border_width;
  auto& value_mask = (*event_).value_mask;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // stack_mode
  uint8_t tmp15;
  Read(&tmp15, &buf);
  stack_mode = static_cast<StackMode>(tmp15);

  // sequence
  Read(&sequence, &buf);

  // parent
  Read(&parent, &buf);

  // window
  Read(&window, &buf);

  // sibling
  Read(&sibling, &buf);

  // x
  Read(&x, &buf);

  // y
  Read(&y, &buf);

  // width
  Read(&width, &buf);

  // height
  Read(&height, &buf);

  // border_width
  Read(&border_width, &buf);

  // value_mask
  uint16_t tmp16;
  Read(&tmp16, &buf);
  value_mask = static_cast<ConfigWindow>(tmp16);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<GravityNotifyEvent>(
    GravityNotifyEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& event = (*event_).event;
  auto& window = (*event_).window;
  auto& x = (*event_).x;
  auto& y = (*event_).y;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // event
  Read(&event, &buf);

  // window
  Read(&window, &buf);

  // x
  Read(&x, &buf);

  // y
  Read(&y, &buf);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<ResizeRequestEvent>(
    ResizeRequestEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& window = (*event_).window;
  auto& width = (*event_).width;
  auto& height = (*event_).height;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // window
  Read(&window, &buf);

  // width
  Read(&width, &buf);

  // height
  Read(&height, &buf);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<CirculateEvent>(
    CirculateEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& event = (*event_).event;
  auto& window = (*event_).window;
  auto& place = (*event_).place;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // event
  Read(&event, &buf);

  // window
  Read(&window, &buf);

  // pad1
  Pad(&buf, 4);

  // place
  uint8_t tmp17;
  Read(&tmp17, &buf);
  place = static_cast<Place>(tmp17);

  // pad2
  Pad(&buf, 3);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<PropertyNotifyEvent>(
    PropertyNotifyEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& window = (*event_).window;
  auto& atom = (*event_).atom;
  auto& time = (*event_).time;
  auto& state = (*event_).state;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // window
  Read(&window, &buf);

  // atom
  Read(&atom, &buf);

  // time
  Read(&time, &buf);

  // state
  uint8_t tmp18;
  Read(&tmp18, &buf);
  state = static_cast<Property>(tmp18);

  // pad1
  Pad(&buf, 3);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<SelectionClearEvent>(
    SelectionClearEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& time = (*event_).time;
  auto& owner = (*event_).owner;
  auto& selection = (*event_).selection;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // time
  Read(&time, &buf);

  // owner
  Read(&owner, &buf);

  // selection
  Read(&selection, &buf);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<SelectionRequestEvent>(
    SelectionRequestEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& time = (*event_).time;
  auto& owner = (*event_).owner;
  auto& requestor = (*event_).requestor;
  auto& selection = (*event_).selection;
  auto& target = (*event_).target;
  auto& property = (*event_).property;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // time
  Read(&time, &buf);

  // owner
  Read(&owner, &buf);

  // requestor
  Read(&requestor, &buf);

  // selection
  Read(&selection, &buf);

  // target
  Read(&target, &buf);

  // property
  Read(&property, &buf);

}

template <> COMPONENT_EXPORT(X11)
std::vector<uint8_t> Write<SelectionNotifyEvent>(
    const SelectionNotifyEvent& obj) {
  WriteBuffer buf;

  auto& sequence = obj.sequence;
  auto& time = obj.time;
  auto& requestor = obj.requestor;
  auto& selection = obj.selection;
  auto& target = obj.target;
  auto& property = obj.property;

  // response_type
  uint8_t response_type = 31;
  Write(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Write(&sequence, &buf);

  // time
  Write(&time, &buf);

  // requestor
  Write(&requestor, &buf);

  // selection
  Write(&selection, &buf);

  // target
  Write(&target, &buf);

  // property
  Write(&property, &buf);

  return buf;
}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<SelectionNotifyEvent>(
    SelectionNotifyEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& time = (*event_).time;
  auto& requestor = (*event_).requestor;
  auto& selection = (*event_).selection;
  auto& target = (*event_).target;
  auto& property = (*event_).property;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // time
  Read(&time, &buf);

  // requestor
  Read(&requestor, &buf);

  // selection
  Read(&selection, &buf);

  // target
  Read(&target, &buf);

  // property
  Read(&property, &buf);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<ColormapNotifyEvent>(
    ColormapNotifyEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& window = (*event_).window;
  auto& colormap = (*event_).colormap;
  auto& c_new = (*event_).c_new;
  auto& state = (*event_).state;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // window
  Read(&window, &buf);

  // colormap
  Read(&colormap, &buf);

  // c_new
  Read(&c_new, &buf);

  // state
  uint8_t tmp19;
  Read(&tmp19, &buf);
  state = static_cast<ColormapState>(tmp19);

  // pad1
  Pad(&buf, 2);

}

template <> COMPONENT_EXPORT(X11)
std::vector<uint8_t> Write<ClientMessageEvent>(
    const ClientMessageEvent& obj) {
  WriteBuffer buf;

  auto& format = obj.format;
  auto& sequence = obj.sequence;
  auto& window = obj.window;
  auto& type = obj.type;
  auto& data = obj.data;

  // response_type
  uint8_t response_type = 33;
  Write(&response_type, &buf);

  // format
  Write(&format, &buf);

  // sequence
  Write(&sequence, &buf);

  // window
  Write(&window, &buf);

  // type
  Write(&type, &buf);

  // data
  Write(&data, &buf);

  return buf;
}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<ClientMessageEvent>(
    ClientMessageEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& format = (*event_).format;
  auto& sequence = (*event_).sequence;
  auto& window = (*event_).window;
  auto& type = (*event_).type;
  auto& data = (*event_).data;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // format
  Read(&format, &buf);

  // sequence
  Read(&sequence, &buf);

  // window
  Read(&window, &buf);

  // type
  Read(&type, &buf);

  // data
  Read(&data, &buf);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<MappingNotifyEvent>(
    MappingNotifyEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;
  auto& request = (*event_).request;
  auto& first_keycode = (*event_).first_keycode;
  auto& count = (*event_).count;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // request
  uint8_t tmp20;
  Read(&tmp20, &buf);
  request = static_cast<Mapping>(tmp20);

  // first_keycode
  Read(&first_keycode, &buf);

  // count
  Read(&count, &buf);

  // pad1
  Pad(&buf, 1);

}

template <> COMPONENT_EXPORT(X11)
void ReadEvent<GeGenericEvent>(
    GeGenericEvent* event_, const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};

  auto& sequence = (*event_).sequence;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // extension
  uint8_t extension;
  Read(&extension, &buf);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // event_type
  uint16_t event_type;
  Read(&event_type, &buf);

  // pad0
  Pad(&buf, 22);

}

Future<void>
XProto::CreateWindow(
    const CreateWindowRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& depth = request.depth;
  auto& wid = request.wid;
  auto& parent = request.parent;
  auto& x = request.x;
  auto& y = request.y;
  auto& width = request.width;
  auto& height = request.height;
  auto& border_width = request.border_width;
  auto& c_class = request.c_class;
  auto& visual = request.visual;
  CreateWindowAttribute value_mask{};
  auto& value_list = request;

  // major_opcode
  uint8_t major_opcode = 1;
  Write(&major_opcode, &buf);

  // depth
  Write(&depth, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // wid
  Write(&wid, &buf);

  // parent
  Write(&parent, &buf);

  // x
  Write(&x, &buf);

  // y
  Write(&y, &buf);

  // width
  Write(&width, &buf);

  // height
  Write(&height, &buf);

  // border_width
  Write(&border_width, &buf);

  // c_class
  uint16_t tmp21;
  tmp21 = static_cast<uint16_t>(c_class);
  Write(&tmp21, &buf);

  // visual
  Write(&visual, &buf);

  // value_mask
  SwitchVar(CreateWindowAttribute::BackPixmap, value_list.background_pixmap.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::BackPixel, value_list.background_pixel.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::BorderPixmap, value_list.border_pixmap.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::BorderPixel, value_list.border_pixel.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::BitGravity, value_list.bit_gravity.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::WinGravity, value_list.win_gravity.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::BackingStore, value_list.backing_store.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::BackingPlanes, value_list.backing_planes.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::BackingPixel, value_list.backing_pixel.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::OverrideRedirect, value_list.override_redirect.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::SaveUnder, value_list.save_under.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::EventMask, value_list.event_mask.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::DontPropagate, value_list.do_not_propogate_mask.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::Colormap, value_list.colormap.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::Cursor, value_list.cursor.has_value(), true, &value_mask);
  uint32_t tmp22;
  tmp22 = static_cast<uint32_t>(value_mask);
  Write(&tmp22, &buf);

  // value_list
  auto value_list_expr = value_mask;
  if (CaseAnd(value_list_expr, CreateWindowAttribute::BackPixmap)) {
    auto& background_pixmap = *value_list.background_pixmap;

    // background_pixmap
    Write(&background_pixmap, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::BackPixel)) {
    auto& background_pixel = *value_list.background_pixel;

    // background_pixel
    Write(&background_pixel, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::BorderPixmap)) {
    auto& border_pixmap = *value_list.border_pixmap;

    // border_pixmap
    Write(&border_pixmap, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::BorderPixel)) {
    auto& border_pixel = *value_list.border_pixel;

    // border_pixel
    Write(&border_pixel, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::BitGravity)) {
    auto& bit_gravity = *value_list.bit_gravity;

    // bit_gravity
    uint32_t tmp23;
    tmp23 = static_cast<uint32_t>(bit_gravity);
    Write(&tmp23, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::WinGravity)) {
    auto& win_gravity = *value_list.win_gravity;

    // win_gravity
    uint32_t tmp24;
    tmp24 = static_cast<uint32_t>(win_gravity);
    Write(&tmp24, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::BackingStore)) {
    auto& backing_store = *value_list.backing_store;

    // backing_store
    uint32_t tmp25;
    tmp25 = static_cast<uint32_t>(backing_store);
    Write(&tmp25, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::BackingPlanes)) {
    auto& backing_planes = *value_list.backing_planes;

    // backing_planes
    Write(&backing_planes, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::BackingPixel)) {
    auto& backing_pixel = *value_list.backing_pixel;

    // backing_pixel
    Write(&backing_pixel, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::OverrideRedirect)) {
    auto& override_redirect = *value_list.override_redirect;

    // override_redirect
    Write(&override_redirect, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::SaveUnder)) {
    auto& save_under = *value_list.save_under;

    // save_under
    Write(&save_under, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::EventMask)) {
    auto& event_mask = *value_list.event_mask;

    // event_mask
    uint32_t tmp26;
    tmp26 = static_cast<uint32_t>(event_mask);
    Write(&tmp26, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::DontPropagate)) {
    auto& do_not_propogate_mask = *value_list.do_not_propogate_mask;

    // do_not_propogate_mask
    uint32_t tmp27;
    tmp27 = static_cast<uint32_t>(do_not_propogate_mask);
    Write(&tmp27, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::Colormap)) {
    auto& colormap = *value_list.colormap;

    // colormap
    Write(&colormap, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::Cursor)) {
    auto& cursor = *value_list.cursor;

    // cursor
    Write(&cursor, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::ChangeWindowAttributes(
    const ChangeWindowAttributesRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& window = request.window;
  CreateWindowAttribute value_mask{};
  auto& value_list = request;

  // major_opcode
  uint8_t major_opcode = 2;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  // value_mask
  SwitchVar(CreateWindowAttribute::BackPixmap, value_list.background_pixmap.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::BackPixel, value_list.background_pixel.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::BorderPixmap, value_list.border_pixmap.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::BorderPixel, value_list.border_pixel.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::BitGravity, value_list.bit_gravity.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::WinGravity, value_list.win_gravity.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::BackingStore, value_list.backing_store.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::BackingPlanes, value_list.backing_planes.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::BackingPixel, value_list.backing_pixel.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::OverrideRedirect, value_list.override_redirect.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::SaveUnder, value_list.save_under.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::EventMask, value_list.event_mask.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::DontPropagate, value_list.do_not_propogate_mask.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::Colormap, value_list.colormap.has_value(), true, &value_mask);
  SwitchVar(CreateWindowAttribute::Cursor, value_list.cursor.has_value(), true, &value_mask);
  uint32_t tmp28;
  tmp28 = static_cast<uint32_t>(value_mask);
  Write(&tmp28, &buf);

  // value_list
  auto value_list_expr = value_mask;
  if (CaseAnd(value_list_expr, CreateWindowAttribute::BackPixmap)) {
    auto& background_pixmap = *value_list.background_pixmap;

    // background_pixmap
    Write(&background_pixmap, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::BackPixel)) {
    auto& background_pixel = *value_list.background_pixel;

    // background_pixel
    Write(&background_pixel, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::BorderPixmap)) {
    auto& border_pixmap = *value_list.border_pixmap;

    // border_pixmap
    Write(&border_pixmap, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::BorderPixel)) {
    auto& border_pixel = *value_list.border_pixel;

    // border_pixel
    Write(&border_pixel, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::BitGravity)) {
    auto& bit_gravity = *value_list.bit_gravity;

    // bit_gravity
    uint32_t tmp29;
    tmp29 = static_cast<uint32_t>(bit_gravity);
    Write(&tmp29, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::WinGravity)) {
    auto& win_gravity = *value_list.win_gravity;

    // win_gravity
    uint32_t tmp30;
    tmp30 = static_cast<uint32_t>(win_gravity);
    Write(&tmp30, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::BackingStore)) {
    auto& backing_store = *value_list.backing_store;

    // backing_store
    uint32_t tmp31;
    tmp31 = static_cast<uint32_t>(backing_store);
    Write(&tmp31, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::BackingPlanes)) {
    auto& backing_planes = *value_list.backing_planes;

    // backing_planes
    Write(&backing_planes, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::BackingPixel)) {
    auto& backing_pixel = *value_list.backing_pixel;

    // backing_pixel
    Write(&backing_pixel, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::OverrideRedirect)) {
    auto& override_redirect = *value_list.override_redirect;

    // override_redirect
    Write(&override_redirect, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::SaveUnder)) {
    auto& save_under = *value_list.save_under;

    // save_under
    Write(&save_under, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::EventMask)) {
    auto& event_mask = *value_list.event_mask;

    // event_mask
    uint32_t tmp32;
    tmp32 = static_cast<uint32_t>(event_mask);
    Write(&tmp32, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::DontPropagate)) {
    auto& do_not_propogate_mask = *value_list.do_not_propogate_mask;

    // do_not_propogate_mask
    uint32_t tmp33;
    tmp33 = static_cast<uint32_t>(do_not_propogate_mask);
    Write(&tmp33, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::Colormap)) {
    auto& colormap = *value_list.colormap;

    // colormap
    Write(&colormap, &buf);

  }
  if (CaseAnd(value_list_expr, CreateWindowAttribute::Cursor)) {
    auto& cursor = *value_list.cursor;

    // cursor
    Write(&cursor, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<GetWindowAttributesReply>
XProto::GetWindowAttributes(
    const GetWindowAttributesRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 3;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<GetWindowAttributesReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<GetWindowAttributesReply>
detail::ReadReply<GetWindowAttributesReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<GetWindowAttributesReply>();

  auto& backing_store = (*reply).backing_store;
  auto& sequence = (*reply).sequence;
  auto& visual = (*reply).visual;
  auto& c_class = (*reply).c_class;
  auto& bit_gravity = (*reply).bit_gravity;
  auto& win_gravity = (*reply).win_gravity;
  auto& backing_planes = (*reply).backing_planes;
  auto& backing_pixel = (*reply).backing_pixel;
  auto& save_under = (*reply).save_under;
  auto& map_is_installed = (*reply).map_is_installed;
  auto& map_state = (*reply).map_state;
  auto& override_redirect = (*reply).override_redirect;
  auto& colormap = (*reply).colormap;
  auto& all_event_masks = (*reply).all_event_masks;
  auto& your_event_mask = (*reply).your_event_mask;
  auto& do_not_propagate_mask = (*reply).do_not_propagate_mask;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // backing_store
  uint8_t tmp34;
  Read(&tmp34, &buf);
  backing_store = static_cast<BackingStore>(tmp34);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // visual
  Read(&visual, &buf);

  // c_class
  uint16_t tmp35;
  Read(&tmp35, &buf);
  c_class = static_cast<WindowClass>(tmp35);

  // bit_gravity
  uint8_t tmp36;
  Read(&tmp36, &buf);
  bit_gravity = static_cast<Gravity>(tmp36);

  // win_gravity
  uint8_t tmp37;
  Read(&tmp37, &buf);
  win_gravity = static_cast<Gravity>(tmp37);

  // backing_planes
  Read(&backing_planes, &buf);

  // backing_pixel
  Read(&backing_pixel, &buf);

  // save_under
  Read(&save_under, &buf);

  // map_is_installed
  Read(&map_is_installed, &buf);

  // map_state
  uint8_t tmp38;
  Read(&tmp38, &buf);
  map_state = static_cast<MapState>(tmp38);

  // override_redirect
  Read(&override_redirect, &buf);

  // colormap
  Read(&colormap, &buf);

  // all_event_masks
  uint32_t tmp39;
  Read(&tmp39, &buf);
  all_event_masks = static_cast<EventMask>(tmp39);

  // your_event_mask
  uint32_t tmp40;
  Read(&tmp40, &buf);
  your_event_mask = static_cast<EventMask>(tmp40);

  // do_not_propagate_mask
  uint16_t tmp41;
  Read(&tmp41, &buf);
  do_not_propagate_mask = static_cast<EventMask>(tmp41);

  // pad0
  Pad(&buf, 2);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::DestroyWindow(
    const DestroyWindowRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 4;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::DestroySubwindows(
    const DestroySubwindowsRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 5;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::ChangeSaveSet(
    const ChangeSaveSetRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& mode = request.mode;
  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 6;
  Write(&major_opcode, &buf);

  // mode
  uint8_t tmp42;
  tmp42 = static_cast<uint8_t>(mode);
  Write(&tmp42, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::ReparentWindow(
    const ReparentWindowRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& window = request.window;
  auto& parent = request.parent;
  auto& x = request.x;
  auto& y = request.y;

  // major_opcode
  uint8_t major_opcode = 7;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  // parent
  Write(&parent, &buf);

  // x
  Write(&x, &buf);

  // y
  Write(&y, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::MapWindow(
    const MapWindowRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 8;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::MapSubwindows(
    const MapSubwindowsRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 9;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::UnmapWindow(
    const UnmapWindowRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 10;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::UnmapSubwindows(
    const UnmapSubwindowsRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 11;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::ConfigureWindow(
    const ConfigureWindowRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& window = request.window;
  ConfigWindow value_mask{};
  auto& value_list = request;

  // major_opcode
  uint8_t major_opcode = 12;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  // value_mask
  SwitchVar(ConfigWindow::X, value_list.x.has_value(), true, &value_mask);
  SwitchVar(ConfigWindow::Y, value_list.y.has_value(), true, &value_mask);
  SwitchVar(ConfigWindow::Width, value_list.width.has_value(), true, &value_mask);
  SwitchVar(ConfigWindow::Height, value_list.height.has_value(), true, &value_mask);
  SwitchVar(ConfigWindow::BorderWidth, value_list.border_width.has_value(), true, &value_mask);
  SwitchVar(ConfigWindow::Sibling, value_list.sibling.has_value(), true, &value_mask);
  SwitchVar(ConfigWindow::StackMode, value_list.stack_mode.has_value(), true, &value_mask);
  uint16_t tmp43;
  tmp43 = static_cast<uint16_t>(value_mask);
  Write(&tmp43, &buf);

  // pad1
  Pad(&buf, 2);

  // value_list
  auto value_list_expr = value_mask;
  if (CaseAnd(value_list_expr, ConfigWindow::X)) {
    auto& x = *value_list.x;

    // x
    Write(&x, &buf);

  }
  if (CaseAnd(value_list_expr, ConfigWindow::Y)) {
    auto& y = *value_list.y;

    // y
    Write(&y, &buf);

  }
  if (CaseAnd(value_list_expr, ConfigWindow::Width)) {
    auto& width = *value_list.width;

    // width
    Write(&width, &buf);

  }
  if (CaseAnd(value_list_expr, ConfigWindow::Height)) {
    auto& height = *value_list.height;

    // height
    Write(&height, &buf);

  }
  if (CaseAnd(value_list_expr, ConfigWindow::BorderWidth)) {
    auto& border_width = *value_list.border_width;

    // border_width
    Write(&border_width, &buf);

  }
  if (CaseAnd(value_list_expr, ConfigWindow::Sibling)) {
    auto& sibling = *value_list.sibling;

    // sibling
    Write(&sibling, &buf);

  }
  if (CaseAnd(value_list_expr, ConfigWindow::StackMode)) {
    auto& stack_mode = *value_list.stack_mode;

    // stack_mode
    uint32_t tmp44;
    tmp44 = static_cast<uint32_t>(stack_mode);
    Write(&tmp44, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::CirculateWindow(
    const CirculateWindowRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& direction = request.direction;
  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 13;
  Write(&major_opcode, &buf);

  // direction
  uint8_t tmp45;
  tmp45 = static_cast<uint8_t>(direction);
  Write(&tmp45, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<GetGeometryReply>
XProto::GetGeometry(
    const GetGeometryRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& drawable = request.drawable;

  // major_opcode
  uint8_t major_opcode = 14;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  Align(&buf, 4);

  return x11::SendRequest<GetGeometryReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<GetGeometryReply>
detail::ReadReply<GetGeometryReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<GetGeometryReply>();

  auto& depth = (*reply).depth;
  auto& sequence = (*reply).sequence;
  auto& root = (*reply).root;
  auto& x = (*reply).x;
  auto& y = (*reply).y;
  auto& width = (*reply).width;
  auto& height = (*reply).height;
  auto& border_width = (*reply).border_width;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // depth
  Read(&depth, &buf);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // root
  Read(&root, &buf);

  // x
  Read(&x, &buf);

  // y
  Read(&y, &buf);

  // width
  Read(&width, &buf);

  // height
  Read(&height, &buf);

  // border_width
  Read(&border_width, &buf);

  // pad0
  Pad(&buf, 2);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<QueryTreeReply>
XProto::QueryTree(
    const QueryTreeRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 15;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<QueryTreeReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<QueryTreeReply>
detail::ReadReply<QueryTreeReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<QueryTreeReply>();

  auto& sequence = (*reply).sequence;
  auto& root = (*reply).root;
  auto& parent = (*reply).parent;
  uint16_t children_len{};
  auto& children = (*reply).children;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // root
  Read(&root, &buf);

  // parent
  Read(&parent, &buf);

  // children_len
  Read(&children_len, &buf);

  // pad1
  Pad(&buf, 14);

  // children
  children.resize(children_len);
  for (auto& children_elem : children) {
    // children_elem
    Read(&children_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<InternAtomReply>
XProto::InternAtom(
    const InternAtomRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& only_if_exists = request.only_if_exists;
  uint16_t name_len{};
  auto& name = request.name;

  // major_opcode
  uint8_t major_opcode = 16;
  Write(&major_opcode, &buf);

  // only_if_exists
  Write(&only_if_exists, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // name_len
  name_len = name.size();
  Write(&name_len, &buf);

  // pad0
  Pad(&buf, 2);

  // name
  DCHECK_EQ(static_cast<size_t>(name_len), name.size());
  for (auto& name_elem : name) {
    // name_elem
    Write(&name_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<InternAtomReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<InternAtomReply>
detail::ReadReply<InternAtomReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<InternAtomReply>();

  auto& sequence = (*reply).sequence;
  auto& atom = (*reply).atom;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // atom
  Read(&atom, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<GetAtomNameReply>
XProto::GetAtomName(
    const GetAtomNameRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& atom = request.atom;

  // major_opcode
  uint8_t major_opcode = 17;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // atom
  Write(&atom, &buf);

  Align(&buf, 4);

  return x11::SendRequest<GetAtomNameReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<GetAtomNameReply>
detail::ReadReply<GetAtomNameReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<GetAtomNameReply>();

  auto& sequence = (*reply).sequence;
  uint16_t name_len{};
  auto& name = (*reply).name;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // name_len
  Read(&name_len, &buf);

  // pad1
  Pad(&buf, 22);

  // name
  name.resize(name_len);
  for (auto& name_elem : name) {
    // name_elem
    Read(&name_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::ChangeProperty(
    const ChangePropertyRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& mode = request.mode;
  auto& window = request.window;
  auto& property = request.property;
  auto& type = request.type;
  auto& format = request.format;
  auto& data_len = request.data_len;
  auto& data = request.data;

  // major_opcode
  uint8_t major_opcode = 18;
  Write(&major_opcode, &buf);

  // mode
  uint8_t tmp46;
  tmp46 = static_cast<uint8_t>(mode);
  Write(&tmp46, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  // property
  Write(&property, &buf);

  // type
  Write(&type, &buf);

  // format
  Write(&format, &buf);

  // pad0
  Pad(&buf, 3);

  // data_len
  Write(&data_len, &buf);

  // data
  DCHECK_EQ(static_cast<size_t>(((data_len) * (format)) / (8)), data.size());
  for (auto& data_elem : data) {
    // data_elem
    Write(&data_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::DeleteProperty(
    const DeletePropertyRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& window = request.window;
  auto& property = request.property;

  // major_opcode
  uint8_t major_opcode = 19;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  // property
  Write(&property, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<GetPropertyReply>
XProto::GetProperty(
    const GetPropertyRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& c_delete = request.c_delete;
  auto& window = request.window;
  auto& property = request.property;
  auto& type = request.type;
  auto& long_offset = request.long_offset;
  auto& long_length = request.long_length;

  // major_opcode
  uint8_t major_opcode = 20;
  Write(&major_opcode, &buf);

  // c_delete
  Write(&c_delete, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  // property
  Write(&property, &buf);

  // type
  Write(&type, &buf);

  // long_offset
  Write(&long_offset, &buf);

  // long_length
  Write(&long_length, &buf);

  Align(&buf, 4);

  return x11::SendRequest<GetPropertyReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<GetPropertyReply>
detail::ReadReply<GetPropertyReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<GetPropertyReply>();

  auto& format = (*reply).format;
  auto& sequence = (*reply).sequence;
  auto& type = (*reply).type;
  auto& bytes_after = (*reply).bytes_after;
  auto& value_len = (*reply).value_len;
  auto& value = (*reply).value;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // format
  Read(&format, &buf);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // type
  Read(&type, &buf);

  // bytes_after
  Read(&bytes_after, &buf);

  // value_len
  Read(&value_len, &buf);

  // pad0
  Pad(&buf, 12);

  // value
  value.resize((value_len) * ((format) / (8)));
  for (auto& value_elem : value) {
    // value_elem
    Read(&value_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<ListPropertiesReply>
XProto::ListProperties(
    const ListPropertiesRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 21;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<ListPropertiesReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<ListPropertiesReply>
detail::ReadReply<ListPropertiesReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<ListPropertiesReply>();

  auto& sequence = (*reply).sequence;
  uint16_t atoms_len{};
  auto& atoms = (*reply).atoms;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // atoms_len
  Read(&atoms_len, &buf);

  // pad1
  Pad(&buf, 22);

  // atoms
  atoms.resize(atoms_len);
  for (auto& atoms_elem : atoms) {
    // atoms_elem
    Read(&atoms_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::SetSelectionOwner(
    const SetSelectionOwnerRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& owner = request.owner;
  auto& selection = request.selection;
  auto& time = request.time;

  // major_opcode
  uint8_t major_opcode = 22;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // owner
  Write(&owner, &buf);

  // selection
  Write(&selection, &buf);

  // time
  Write(&time, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<GetSelectionOwnerReply>
XProto::GetSelectionOwner(
    const GetSelectionOwnerRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& selection = request.selection;

  // major_opcode
  uint8_t major_opcode = 23;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // selection
  Write(&selection, &buf);

  Align(&buf, 4);

  return x11::SendRequest<GetSelectionOwnerReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<GetSelectionOwnerReply>
detail::ReadReply<GetSelectionOwnerReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<GetSelectionOwnerReply>();

  auto& sequence = (*reply).sequence;
  auto& owner = (*reply).owner;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // owner
  Read(&owner, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::ConvertSelection(
    const ConvertSelectionRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& requestor = request.requestor;
  auto& selection = request.selection;
  auto& target = request.target;
  auto& property = request.property;
  auto& time = request.time;

  // major_opcode
  uint8_t major_opcode = 24;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // requestor
  Write(&requestor, &buf);

  // selection
  Write(&selection, &buf);

  // target
  Write(&target, &buf);

  // property
  Write(&property, &buf);

  // time
  Write(&time, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::SendEvent(
    const SendEventRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& propagate = request.propagate;
  auto& destination = request.destination;
  auto& event_mask = request.event_mask;
  auto& event = request.event;
  size_t event_len = event.size();

  // major_opcode
  uint8_t major_opcode = 25;
  Write(&major_opcode, &buf);

  // propagate
  Write(&propagate, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // destination
  Write(&destination, &buf);

  // event_mask
  uint32_t tmp47;
  tmp47 = static_cast<uint32_t>(event_mask);
  Write(&tmp47, &buf);

  // event
  for (auto& event_elem : event) {
    // event_elem
    Write(&event_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<GrabPointerReply>
XProto::GrabPointer(
    const GrabPointerRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& owner_events = request.owner_events;
  auto& grab_window = request.grab_window;
  auto& event_mask = request.event_mask;
  auto& pointer_mode = request.pointer_mode;
  auto& keyboard_mode = request.keyboard_mode;
  auto& confine_to = request.confine_to;
  auto& cursor = request.cursor;
  auto& time = request.time;

  // major_opcode
  uint8_t major_opcode = 26;
  Write(&major_opcode, &buf);

  // owner_events
  Write(&owner_events, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // grab_window
  Write(&grab_window, &buf);

  // event_mask
  uint16_t tmp48;
  tmp48 = static_cast<uint16_t>(event_mask);
  Write(&tmp48, &buf);

  // pointer_mode
  uint8_t tmp49;
  tmp49 = static_cast<uint8_t>(pointer_mode);
  Write(&tmp49, &buf);

  // keyboard_mode
  uint8_t tmp50;
  tmp50 = static_cast<uint8_t>(keyboard_mode);
  Write(&tmp50, &buf);

  // confine_to
  Write(&confine_to, &buf);

  // cursor
  Write(&cursor, &buf);

  // time
  Write(&time, &buf);

  Align(&buf, 4);

  return x11::SendRequest<GrabPointerReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<GrabPointerReply>
detail::ReadReply<GrabPointerReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<GrabPointerReply>();

  auto& status = (*reply).status;
  auto& sequence = (*reply).sequence;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // status
  uint8_t tmp51;
  Read(&tmp51, &buf);
  status = static_cast<GrabStatus>(tmp51);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::UngrabPointer(
    const UngrabPointerRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& time = request.time;

  // major_opcode
  uint8_t major_opcode = 27;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // time
  Write(&time, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::GrabButton(
    const GrabButtonRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& owner_events = request.owner_events;
  auto& grab_window = request.grab_window;
  auto& event_mask = request.event_mask;
  auto& pointer_mode = request.pointer_mode;
  auto& keyboard_mode = request.keyboard_mode;
  auto& confine_to = request.confine_to;
  auto& cursor = request.cursor;
  auto& button = request.button;
  auto& modifiers = request.modifiers;

  // major_opcode
  uint8_t major_opcode = 28;
  Write(&major_opcode, &buf);

  // owner_events
  Write(&owner_events, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // grab_window
  Write(&grab_window, &buf);

  // event_mask
  uint16_t tmp52;
  tmp52 = static_cast<uint16_t>(event_mask);
  Write(&tmp52, &buf);

  // pointer_mode
  uint8_t tmp53;
  tmp53 = static_cast<uint8_t>(pointer_mode);
  Write(&tmp53, &buf);

  // keyboard_mode
  uint8_t tmp54;
  tmp54 = static_cast<uint8_t>(keyboard_mode);
  Write(&tmp54, &buf);

  // confine_to
  Write(&confine_to, &buf);

  // cursor
  Write(&cursor, &buf);

  // button
  uint8_t tmp55;
  tmp55 = static_cast<uint8_t>(button);
  Write(&tmp55, &buf);

  // pad0
  Pad(&buf, 1);

  // modifiers
  uint16_t tmp56;
  tmp56 = static_cast<uint16_t>(modifiers);
  Write(&tmp56, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::UngrabButton(
    const UngrabButtonRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& button = request.button;
  auto& grab_window = request.grab_window;
  auto& modifiers = request.modifiers;

  // major_opcode
  uint8_t major_opcode = 29;
  Write(&major_opcode, &buf);

  // button
  uint8_t tmp57;
  tmp57 = static_cast<uint8_t>(button);
  Write(&tmp57, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // grab_window
  Write(&grab_window, &buf);

  // modifiers
  uint16_t tmp58;
  tmp58 = static_cast<uint16_t>(modifiers);
  Write(&tmp58, &buf);

  // pad0
  Pad(&buf, 2);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::ChangeActivePointerGrab(
    const ChangeActivePointerGrabRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& cursor = request.cursor;
  auto& time = request.time;
  auto& event_mask = request.event_mask;

  // major_opcode
  uint8_t major_opcode = 30;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cursor
  Write(&cursor, &buf);

  // time
  Write(&time, &buf);

  // event_mask
  uint16_t tmp59;
  tmp59 = static_cast<uint16_t>(event_mask);
  Write(&tmp59, &buf);

  // pad1
  Pad(&buf, 2);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<GrabKeyboardReply>
XProto::GrabKeyboard(
    const GrabKeyboardRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& owner_events = request.owner_events;
  auto& grab_window = request.grab_window;
  auto& time = request.time;
  auto& pointer_mode = request.pointer_mode;
  auto& keyboard_mode = request.keyboard_mode;

  // major_opcode
  uint8_t major_opcode = 31;
  Write(&major_opcode, &buf);

  // owner_events
  Write(&owner_events, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // grab_window
  Write(&grab_window, &buf);

  // time
  Write(&time, &buf);

  // pointer_mode
  uint8_t tmp60;
  tmp60 = static_cast<uint8_t>(pointer_mode);
  Write(&tmp60, &buf);

  // keyboard_mode
  uint8_t tmp61;
  tmp61 = static_cast<uint8_t>(keyboard_mode);
  Write(&tmp61, &buf);

  // pad0
  Pad(&buf, 2);

  Align(&buf, 4);

  return x11::SendRequest<GrabKeyboardReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<GrabKeyboardReply>
detail::ReadReply<GrabKeyboardReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<GrabKeyboardReply>();

  auto& status = (*reply).status;
  auto& sequence = (*reply).sequence;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // status
  uint8_t tmp62;
  Read(&tmp62, &buf);
  status = static_cast<GrabStatus>(tmp62);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::UngrabKeyboard(
    const UngrabKeyboardRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& time = request.time;

  // major_opcode
  uint8_t major_opcode = 32;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // time
  Write(&time, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::GrabKey(
    const GrabKeyRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& owner_events = request.owner_events;
  auto& grab_window = request.grab_window;
  auto& modifiers = request.modifiers;
  auto& key = request.key;
  auto& pointer_mode = request.pointer_mode;
  auto& keyboard_mode = request.keyboard_mode;

  // major_opcode
  uint8_t major_opcode = 33;
  Write(&major_opcode, &buf);

  // owner_events
  Write(&owner_events, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // grab_window
  Write(&grab_window, &buf);

  // modifiers
  uint16_t tmp63;
  tmp63 = static_cast<uint16_t>(modifiers);
  Write(&tmp63, &buf);

  // key
  Write(&key, &buf);

  // pointer_mode
  uint8_t tmp64;
  tmp64 = static_cast<uint8_t>(pointer_mode);
  Write(&tmp64, &buf);

  // keyboard_mode
  uint8_t tmp65;
  tmp65 = static_cast<uint8_t>(keyboard_mode);
  Write(&tmp65, &buf);

  // pad0
  Pad(&buf, 3);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::UngrabKey(
    const UngrabKeyRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& key = request.key;
  auto& grab_window = request.grab_window;
  auto& modifiers = request.modifiers;

  // major_opcode
  uint8_t major_opcode = 34;
  Write(&major_opcode, &buf);

  // key
  Write(&key, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // grab_window
  Write(&grab_window, &buf);

  // modifiers
  uint16_t tmp66;
  tmp66 = static_cast<uint16_t>(modifiers);
  Write(&tmp66, &buf);

  // pad0
  Pad(&buf, 2);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::AllowEvents(
    const AllowEventsRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& mode = request.mode;
  auto& time = request.time;

  // major_opcode
  uint8_t major_opcode = 35;
  Write(&major_opcode, &buf);

  // mode
  uint8_t tmp67;
  tmp67 = static_cast<uint8_t>(mode);
  Write(&tmp67, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // time
  Write(&time, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::GrabServer(
    const GrabServerRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 36;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::UngrabServer(
    const UngrabServerRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 37;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<QueryPointerReply>
XProto::QueryPointer(
    const QueryPointerRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 38;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<QueryPointerReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<QueryPointerReply>
detail::ReadReply<QueryPointerReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<QueryPointerReply>();

  auto& same_screen = (*reply).same_screen;
  auto& sequence = (*reply).sequence;
  auto& root = (*reply).root;
  auto& child = (*reply).child;
  auto& root_x = (*reply).root_x;
  auto& root_y = (*reply).root_y;
  auto& win_x = (*reply).win_x;
  auto& win_y = (*reply).win_y;
  auto& mask = (*reply).mask;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // same_screen
  Read(&same_screen, &buf);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // root
  Read(&root, &buf);

  // child
  Read(&child, &buf);

  // root_x
  Read(&root_x, &buf);

  // root_y
  Read(&root_y, &buf);

  // win_x
  Read(&win_x, &buf);

  // win_y
  Read(&win_y, &buf);

  // mask
  uint16_t tmp68;
  Read(&tmp68, &buf);
  mask = static_cast<KeyButMask>(tmp68);

  // pad0
  Pad(&buf, 2);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<GetMotionEventsReply>
XProto::GetMotionEvents(
    const GetMotionEventsRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& window = request.window;
  auto& start = request.start;
  auto& stop = request.stop;

  // major_opcode
  uint8_t major_opcode = 39;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  // start
  Write(&start, &buf);

  // stop
  Write(&stop, &buf);

  Align(&buf, 4);

  return x11::SendRequest<GetMotionEventsReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<GetMotionEventsReply>
detail::ReadReply<GetMotionEventsReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<GetMotionEventsReply>();

  auto& sequence = (*reply).sequence;
  uint32_t events_len{};
  auto& events = (*reply).events;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // events_len
  Read(&events_len, &buf);

  // pad1
  Pad(&buf, 20);

  // events
  events.resize(events_len);
  for (auto& events_elem : events) {
    // events_elem
    {
      auto& time = events_elem.time;
      auto& x = events_elem.x;
      auto& y = events_elem.y;

      // time
      Read(&time, &buf);

      // x
      Read(&x, &buf);

      // y
      Read(&y, &buf);

    }

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<TranslateCoordinatesReply>
XProto::TranslateCoordinates(
    const TranslateCoordinatesRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& src_window = request.src_window;
  auto& dst_window = request.dst_window;
  auto& src_x = request.src_x;
  auto& src_y = request.src_y;

  // major_opcode
  uint8_t major_opcode = 40;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // src_window
  Write(&src_window, &buf);

  // dst_window
  Write(&dst_window, &buf);

  // src_x
  Write(&src_x, &buf);

  // src_y
  Write(&src_y, &buf);

  Align(&buf, 4);

  return x11::SendRequest<TranslateCoordinatesReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<TranslateCoordinatesReply>
detail::ReadReply<TranslateCoordinatesReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<TranslateCoordinatesReply>();

  auto& same_screen = (*reply).same_screen;
  auto& sequence = (*reply).sequence;
  auto& child = (*reply).child;
  auto& dst_x = (*reply).dst_x;
  auto& dst_y = (*reply).dst_y;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // same_screen
  Read(&same_screen, &buf);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // child
  Read(&child, &buf);

  // dst_x
  Read(&dst_x, &buf);

  // dst_y
  Read(&dst_y, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::WarpPointer(
    const WarpPointerRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& src_window = request.src_window;
  auto& dst_window = request.dst_window;
  auto& src_x = request.src_x;
  auto& src_y = request.src_y;
  auto& src_width = request.src_width;
  auto& src_height = request.src_height;
  auto& dst_x = request.dst_x;
  auto& dst_y = request.dst_y;

  // major_opcode
  uint8_t major_opcode = 41;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // src_window
  Write(&src_window, &buf);

  // dst_window
  Write(&dst_window, &buf);

  // src_x
  Write(&src_x, &buf);

  // src_y
  Write(&src_y, &buf);

  // src_width
  Write(&src_width, &buf);

  // src_height
  Write(&src_height, &buf);

  // dst_x
  Write(&dst_x, &buf);

  // dst_y
  Write(&dst_y, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::SetInputFocus(
    const SetInputFocusRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& revert_to = request.revert_to;
  auto& focus = request.focus;
  auto& time = request.time;

  // major_opcode
  uint8_t major_opcode = 42;
  Write(&major_opcode, &buf);

  // revert_to
  uint8_t tmp69;
  tmp69 = static_cast<uint8_t>(revert_to);
  Write(&tmp69, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // focus
  Write(&focus, &buf);

  // time
  Write(&time, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<GetInputFocusReply>
XProto::GetInputFocus(
    const GetInputFocusRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 43;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<GetInputFocusReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<GetInputFocusReply>
detail::ReadReply<GetInputFocusReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<GetInputFocusReply>();

  auto& revert_to = (*reply).revert_to;
  auto& sequence = (*reply).sequence;
  auto& focus = (*reply).focus;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // revert_to
  uint8_t tmp70;
  Read(&tmp70, &buf);
  revert_to = static_cast<InputFocus>(tmp70);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // focus
  Read(&focus, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<QueryKeymapReply>
XProto::QueryKeymap(
    const QueryKeymapRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 44;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<QueryKeymapReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<QueryKeymapReply>
detail::ReadReply<QueryKeymapReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<QueryKeymapReply>();

  auto& sequence = (*reply).sequence;
  auto& keys = (*reply).keys;
  size_t keys_len = keys.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // keys
  for (auto& keys_elem : keys) {
    // keys_elem
    Read(&keys_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::OpenFont(
    const OpenFontRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& fid = request.fid;
  uint16_t name_len{};
  auto& name = request.name;

  // major_opcode
  uint8_t major_opcode = 45;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // fid
  Write(&fid, &buf);

  // name_len
  name_len = name.size();
  Write(&name_len, &buf);

  // pad1
  Pad(&buf, 2);

  // name
  DCHECK_EQ(static_cast<size_t>(name_len), name.size());
  for (auto& name_elem : name) {
    // name_elem
    Write(&name_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::CloseFont(
    const CloseFontRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& font = request.font;

  // major_opcode
  uint8_t major_opcode = 46;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // font
  Write(&font, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<QueryFontReply>
XProto::QueryFont(
    const QueryFontRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& font = request.font;

  // major_opcode
  uint8_t major_opcode = 47;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // font
  Write(&font, &buf);

  Align(&buf, 4);

  return x11::SendRequest<QueryFontReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<QueryFontReply>
detail::ReadReply<QueryFontReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<QueryFontReply>();

  auto& sequence = (*reply).sequence;
  auto& min_bounds = (*reply).min_bounds;
  auto& max_bounds = (*reply).max_bounds;
  auto& min_char_or_byte2 = (*reply).min_char_or_byte2;
  auto& max_char_or_byte2 = (*reply).max_char_or_byte2;
  auto& default_char = (*reply).default_char;
  uint16_t properties_len{};
  auto& draw_direction = (*reply).draw_direction;
  auto& min_byte1 = (*reply).min_byte1;
  auto& max_byte1 = (*reply).max_byte1;
  auto& all_chars_exist = (*reply).all_chars_exist;
  auto& font_ascent = (*reply).font_ascent;
  auto& font_descent = (*reply).font_descent;
  uint32_t char_infos_len{};
  auto& properties = (*reply).properties;
  auto& char_infos = (*reply).char_infos;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // min_bounds
  {
    auto& left_side_bearing = min_bounds.left_side_bearing;
    auto& right_side_bearing = min_bounds.right_side_bearing;
    auto& character_width = min_bounds.character_width;
    auto& ascent = min_bounds.ascent;
    auto& descent = min_bounds.descent;
    auto& attributes = min_bounds.attributes;

    // left_side_bearing
    Read(&left_side_bearing, &buf);

    // right_side_bearing
    Read(&right_side_bearing, &buf);

    // character_width
    Read(&character_width, &buf);

    // ascent
    Read(&ascent, &buf);

    // descent
    Read(&descent, &buf);

    // attributes
    Read(&attributes, &buf);

  }

  // pad1
  Pad(&buf, 4);

  // max_bounds
  {
    auto& left_side_bearing = max_bounds.left_side_bearing;
    auto& right_side_bearing = max_bounds.right_side_bearing;
    auto& character_width = max_bounds.character_width;
    auto& ascent = max_bounds.ascent;
    auto& descent = max_bounds.descent;
    auto& attributes = max_bounds.attributes;

    // left_side_bearing
    Read(&left_side_bearing, &buf);

    // right_side_bearing
    Read(&right_side_bearing, &buf);

    // character_width
    Read(&character_width, &buf);

    // ascent
    Read(&ascent, &buf);

    // descent
    Read(&descent, &buf);

    // attributes
    Read(&attributes, &buf);

  }

  // pad2
  Pad(&buf, 4);

  // min_char_or_byte2
  Read(&min_char_or_byte2, &buf);

  // max_char_or_byte2
  Read(&max_char_or_byte2, &buf);

  // default_char
  Read(&default_char, &buf);

  // properties_len
  Read(&properties_len, &buf);

  // draw_direction
  uint8_t tmp71;
  Read(&tmp71, &buf);
  draw_direction = static_cast<FontDraw>(tmp71);

  // min_byte1
  Read(&min_byte1, &buf);

  // max_byte1
  Read(&max_byte1, &buf);

  // all_chars_exist
  Read(&all_chars_exist, &buf);

  // font_ascent
  Read(&font_ascent, &buf);

  // font_descent
  Read(&font_descent, &buf);

  // char_infos_len
  Read(&char_infos_len, &buf);

  // properties
  properties.resize(properties_len);
  for (auto& properties_elem : properties) {
    // properties_elem
    {
      auto& name = properties_elem.name;
      auto& value = properties_elem.value;

      // name
      Read(&name, &buf);

      // value
      Read(&value, &buf);

    }

  }

  // char_infos
  char_infos.resize(char_infos_len);
  for (auto& char_infos_elem : char_infos) {
    // char_infos_elem
    {
      auto& left_side_bearing = char_infos_elem.left_side_bearing;
      auto& right_side_bearing = char_infos_elem.right_side_bearing;
      auto& character_width = char_infos_elem.character_width;
      auto& ascent = char_infos_elem.ascent;
      auto& descent = char_infos_elem.descent;
      auto& attributes = char_infos_elem.attributes;

      // left_side_bearing
      Read(&left_side_bearing, &buf);

      // right_side_bearing
      Read(&right_side_bearing, &buf);

      // character_width
      Read(&character_width, &buf);

      // ascent
      Read(&ascent, &buf);

      // descent
      Read(&descent, &buf);

      // attributes
      Read(&attributes, &buf);

    }

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<QueryTextExtentsReply>
XProto::QueryTextExtents(
    const QueryTextExtentsRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& font = request.font;
  auto& string = request.string;
  size_t string_len = string.size();

  // major_opcode
  uint8_t major_opcode = 48;
  Write(&major_opcode, &buf);

  // odd_length
  uint8_t odd_length = BitAnd(string_len, 1);
  Write(&odd_length, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // font
  Write(&font, &buf);

  // string
  DCHECK_EQ(static_cast<size_t>(string_len), string.size());
  for (auto& string_elem : string) {
    // string_elem
    {
      auto& byte1 = string_elem.byte1;
      auto& byte2 = string_elem.byte2;

      // byte1
      Write(&byte1, &buf);

      // byte2
      Write(&byte2, &buf);

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<QueryTextExtentsReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<QueryTextExtentsReply>
detail::ReadReply<QueryTextExtentsReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<QueryTextExtentsReply>();

  auto& draw_direction = (*reply).draw_direction;
  auto& sequence = (*reply).sequence;
  auto& font_ascent = (*reply).font_ascent;
  auto& font_descent = (*reply).font_descent;
  auto& overall_ascent = (*reply).overall_ascent;
  auto& overall_descent = (*reply).overall_descent;
  auto& overall_width = (*reply).overall_width;
  auto& overall_left = (*reply).overall_left;
  auto& overall_right = (*reply).overall_right;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // draw_direction
  uint8_t tmp72;
  Read(&tmp72, &buf);
  draw_direction = static_cast<FontDraw>(tmp72);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // font_ascent
  Read(&font_ascent, &buf);

  // font_descent
  Read(&font_descent, &buf);

  // overall_ascent
  Read(&overall_ascent, &buf);

  // overall_descent
  Read(&overall_descent, &buf);

  // overall_width
  Read(&overall_width, &buf);

  // overall_left
  Read(&overall_left, &buf);

  // overall_right
  Read(&overall_right, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<ListFontsReply>
XProto::ListFonts(
    const ListFontsRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& max_names = request.max_names;
  uint16_t pattern_len{};
  auto& pattern = request.pattern;

  // major_opcode
  uint8_t major_opcode = 49;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // max_names
  Write(&max_names, &buf);

  // pattern_len
  pattern_len = pattern.size();
  Write(&pattern_len, &buf);

  // pattern
  DCHECK_EQ(static_cast<size_t>(pattern_len), pattern.size());
  for (auto& pattern_elem : pattern) {
    // pattern_elem
    Write(&pattern_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<ListFontsReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<ListFontsReply>
detail::ReadReply<ListFontsReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<ListFontsReply>();

  auto& sequence = (*reply).sequence;
  uint16_t names_len{};
  auto& names = (*reply).names;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // names_len
  Read(&names_len, &buf);

  // pad1
  Pad(&buf, 22);

  // names
  names.resize(names_len);
  for (auto& names_elem : names) {
    // names_elem
    {
      uint8_t name_len{};
      auto& name = names_elem.name;

      // name_len
      Read(&name_len, &buf);

      // name
      name.resize(name_len);
      for (auto& name_elem : name) {
        // name_elem
        Read(&name_elem, &buf);

      }

    }

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<ListFontsWithInfoReply>
XProto::ListFontsWithInfo(
    const ListFontsWithInfoRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& max_names = request.max_names;
  uint16_t pattern_len{};
  auto& pattern = request.pattern;

  // major_opcode
  uint8_t major_opcode = 50;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // max_names
  Write(&max_names, &buf);

  // pattern_len
  pattern_len = pattern.size();
  Write(&pattern_len, &buf);

  // pattern
  DCHECK_EQ(static_cast<size_t>(pattern_len), pattern.size());
  for (auto& pattern_elem : pattern) {
    // pattern_elem
    Write(&pattern_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<ListFontsWithInfoReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<ListFontsWithInfoReply>
detail::ReadReply<ListFontsWithInfoReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<ListFontsWithInfoReply>();

  uint8_t name_len{};
  auto& sequence = (*reply).sequence;
  auto& min_bounds = (*reply).min_bounds;
  auto& max_bounds = (*reply).max_bounds;
  auto& min_char_or_byte2 = (*reply).min_char_or_byte2;
  auto& max_char_or_byte2 = (*reply).max_char_or_byte2;
  auto& default_char = (*reply).default_char;
  uint16_t properties_len{};
  auto& draw_direction = (*reply).draw_direction;
  auto& min_byte1 = (*reply).min_byte1;
  auto& max_byte1 = (*reply).max_byte1;
  auto& all_chars_exist = (*reply).all_chars_exist;
  auto& font_ascent = (*reply).font_ascent;
  auto& font_descent = (*reply).font_descent;
  auto& replies_hint = (*reply).replies_hint;
  auto& properties = (*reply).properties;
  auto& name = (*reply).name;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // name_len
  Read(&name_len, &buf);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // min_bounds
  {
    auto& left_side_bearing = min_bounds.left_side_bearing;
    auto& right_side_bearing = min_bounds.right_side_bearing;
    auto& character_width = min_bounds.character_width;
    auto& ascent = min_bounds.ascent;
    auto& descent = min_bounds.descent;
    auto& attributes = min_bounds.attributes;

    // left_side_bearing
    Read(&left_side_bearing, &buf);

    // right_side_bearing
    Read(&right_side_bearing, &buf);

    // character_width
    Read(&character_width, &buf);

    // ascent
    Read(&ascent, &buf);

    // descent
    Read(&descent, &buf);

    // attributes
    Read(&attributes, &buf);

  }

  // pad0
  Pad(&buf, 4);

  // max_bounds
  {
    auto& left_side_bearing = max_bounds.left_side_bearing;
    auto& right_side_bearing = max_bounds.right_side_bearing;
    auto& character_width = max_bounds.character_width;
    auto& ascent = max_bounds.ascent;
    auto& descent = max_bounds.descent;
    auto& attributes = max_bounds.attributes;

    // left_side_bearing
    Read(&left_side_bearing, &buf);

    // right_side_bearing
    Read(&right_side_bearing, &buf);

    // character_width
    Read(&character_width, &buf);

    // ascent
    Read(&ascent, &buf);

    // descent
    Read(&descent, &buf);

    // attributes
    Read(&attributes, &buf);

  }

  // pad1
  Pad(&buf, 4);

  // min_char_or_byte2
  Read(&min_char_or_byte2, &buf);

  // max_char_or_byte2
  Read(&max_char_or_byte2, &buf);

  // default_char
  Read(&default_char, &buf);

  // properties_len
  Read(&properties_len, &buf);

  // draw_direction
  uint8_t tmp73;
  Read(&tmp73, &buf);
  draw_direction = static_cast<FontDraw>(tmp73);

  // min_byte1
  Read(&min_byte1, &buf);

  // max_byte1
  Read(&max_byte1, &buf);

  // all_chars_exist
  Read(&all_chars_exist, &buf);

  // font_ascent
  Read(&font_ascent, &buf);

  // font_descent
  Read(&font_descent, &buf);

  // replies_hint
  Read(&replies_hint, &buf);

  // properties
  properties.resize(properties_len);
  for (auto& properties_elem : properties) {
    // properties_elem
    {
      auto& name = properties_elem.name;
      auto& value = properties_elem.value;

      // name
      Read(&name, &buf);

      // value
      Read(&value, &buf);

    }

  }

  // name
  name.resize(name_len);
  for (auto& name_elem : name) {
    // name_elem
    Read(&name_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::SetFontPath(
    const SetFontPathRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  uint16_t font_qty{};
  auto& font = request.font;
  size_t font_len = font.size();

  // major_opcode
  uint8_t major_opcode = 51;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // font_qty
  font_qty = font.size();
  Write(&font_qty, &buf);

  // pad1
  Pad(&buf, 2);

  // font
  DCHECK_EQ(static_cast<size_t>(font_qty), font.size());
  for (auto& font_elem : font) {
    // font_elem
    {
      uint8_t name_len{};
      auto& name = font_elem.name;

      // name_len
      name_len = name.size();
      Write(&name_len, &buf);

      // name
      DCHECK_EQ(static_cast<size_t>(name_len), name.size());
      for (auto& name_elem : name) {
        // name_elem
        Write(&name_elem, &buf);

      }

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<GetFontPathReply>
XProto::GetFontPath(
    const GetFontPathRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 52;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<GetFontPathReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<GetFontPathReply>
detail::ReadReply<GetFontPathReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<GetFontPathReply>();

  auto& sequence = (*reply).sequence;
  uint16_t path_len{};
  auto& path = (*reply).path;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // path_len
  Read(&path_len, &buf);

  // pad1
  Pad(&buf, 22);

  // path
  path.resize(path_len);
  for (auto& path_elem : path) {
    // path_elem
    {
      uint8_t name_len{};
      auto& name = path_elem.name;

      // name_len
      Read(&name_len, &buf);

      // name
      name.resize(name_len);
      for (auto& name_elem : name) {
        // name_elem
        Read(&name_elem, &buf);

      }

    }

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::CreatePixmap(
    const CreatePixmapRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& depth = request.depth;
  auto& pid = request.pid;
  auto& drawable = request.drawable;
  auto& width = request.width;
  auto& height = request.height;

  // major_opcode
  uint8_t major_opcode = 53;
  Write(&major_opcode, &buf);

  // depth
  Write(&depth, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // pid
  Write(&pid, &buf);

  // drawable
  Write(&drawable, &buf);

  // width
  Write(&width, &buf);

  // height
  Write(&height, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::FreePixmap(
    const FreePixmapRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& pixmap = request.pixmap;

  // major_opcode
  uint8_t major_opcode = 54;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // pixmap
  Write(&pixmap, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::CreateGC(
    const CreateGCRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& cid = request.cid;
  auto& drawable = request.drawable;
  GraphicsContextAttribute value_mask{};
  auto& value_list = request;

  // major_opcode
  uint8_t major_opcode = 55;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cid
  Write(&cid, &buf);

  // drawable
  Write(&drawable, &buf);

  // value_mask
  SwitchVar(GraphicsContextAttribute::Function, value_list.function.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::PlaneMask, value_list.plane_mask.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::Foreground, value_list.foreground.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::Background, value_list.background.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::LineWidth, value_list.line_width.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::LineStyle, value_list.line_style.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::CapStyle, value_list.cap_style.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::JoinStyle, value_list.join_style.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::FillStyle, value_list.fill_style.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::FillRule, value_list.fill_rule.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::Tile, value_list.tile.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::Stipple, value_list.stipple.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::TileStippleOriginX, value_list.tile_stipple_x_origin.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::TileStippleOriginY, value_list.tile_stipple_y_origin.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::Font, value_list.font.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::SubwindowMode, value_list.subwindow_mode.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::GraphicsExposures, value_list.graphics_exposures.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::ClipOriginX, value_list.clip_x_origin.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::ClipOriginY, value_list.clip_y_origin.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::ClipMask, value_list.clip_mask.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::DashOffset, value_list.dash_offset.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::DashList, value_list.dashes.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::ArcMode, value_list.arc_mode.has_value(), true, &value_mask);
  uint32_t tmp74;
  tmp74 = static_cast<uint32_t>(value_mask);
  Write(&tmp74, &buf);

  // value_list
  auto value_list_expr = value_mask;
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::Function)) {
    auto& function = *value_list.function;

    // function
    uint32_t tmp75;
    tmp75 = static_cast<uint32_t>(function);
    Write(&tmp75, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::PlaneMask)) {
    auto& plane_mask = *value_list.plane_mask;

    // plane_mask
    Write(&plane_mask, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::Foreground)) {
    auto& foreground = *value_list.foreground;

    // foreground
    Write(&foreground, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::Background)) {
    auto& background = *value_list.background;

    // background
    Write(&background, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::LineWidth)) {
    auto& line_width = *value_list.line_width;

    // line_width
    Write(&line_width, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::LineStyle)) {
    auto& line_style = *value_list.line_style;

    // line_style
    uint32_t tmp76;
    tmp76 = static_cast<uint32_t>(line_style);
    Write(&tmp76, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::CapStyle)) {
    auto& cap_style = *value_list.cap_style;

    // cap_style
    uint32_t tmp77;
    tmp77 = static_cast<uint32_t>(cap_style);
    Write(&tmp77, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::JoinStyle)) {
    auto& join_style = *value_list.join_style;

    // join_style
    uint32_t tmp78;
    tmp78 = static_cast<uint32_t>(join_style);
    Write(&tmp78, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::FillStyle)) {
    auto& fill_style = *value_list.fill_style;

    // fill_style
    uint32_t tmp79;
    tmp79 = static_cast<uint32_t>(fill_style);
    Write(&tmp79, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::FillRule)) {
    auto& fill_rule = *value_list.fill_rule;

    // fill_rule
    uint32_t tmp80;
    tmp80 = static_cast<uint32_t>(fill_rule);
    Write(&tmp80, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::Tile)) {
    auto& tile = *value_list.tile;

    // tile
    Write(&tile, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::Stipple)) {
    auto& stipple = *value_list.stipple;

    // stipple
    Write(&stipple, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::TileStippleOriginX)) {
    auto& tile_stipple_x_origin = *value_list.tile_stipple_x_origin;

    // tile_stipple_x_origin
    Write(&tile_stipple_x_origin, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::TileStippleOriginY)) {
    auto& tile_stipple_y_origin = *value_list.tile_stipple_y_origin;

    // tile_stipple_y_origin
    Write(&tile_stipple_y_origin, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::Font)) {
    auto& font = *value_list.font;

    // font
    Write(&font, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::SubwindowMode)) {
    auto& subwindow_mode = *value_list.subwindow_mode;

    // subwindow_mode
    uint32_t tmp81;
    tmp81 = static_cast<uint32_t>(subwindow_mode);
    Write(&tmp81, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::GraphicsExposures)) {
    auto& graphics_exposures = *value_list.graphics_exposures;

    // graphics_exposures
    Write(&graphics_exposures, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::ClipOriginX)) {
    auto& clip_x_origin = *value_list.clip_x_origin;

    // clip_x_origin
    Write(&clip_x_origin, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::ClipOriginY)) {
    auto& clip_y_origin = *value_list.clip_y_origin;

    // clip_y_origin
    Write(&clip_y_origin, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::ClipMask)) {
    auto& clip_mask = *value_list.clip_mask;

    // clip_mask
    Write(&clip_mask, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::DashOffset)) {
    auto& dash_offset = *value_list.dash_offset;

    // dash_offset
    Write(&dash_offset, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::DashList)) {
    auto& dashes = *value_list.dashes;

    // dashes
    Write(&dashes, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::ArcMode)) {
    auto& arc_mode = *value_list.arc_mode;

    // arc_mode
    uint32_t tmp82;
    tmp82 = static_cast<uint32_t>(arc_mode);
    Write(&tmp82, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::ChangeGC(
    const ChangeGCRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& gc = request.gc;
  GraphicsContextAttribute value_mask{};
  auto& value_list = request;

  // major_opcode
  uint8_t major_opcode = 56;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // gc
  Write(&gc, &buf);

  // value_mask
  SwitchVar(GraphicsContextAttribute::Function, value_list.function.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::PlaneMask, value_list.plane_mask.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::Foreground, value_list.foreground.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::Background, value_list.background.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::LineWidth, value_list.line_width.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::LineStyle, value_list.line_style.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::CapStyle, value_list.cap_style.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::JoinStyle, value_list.join_style.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::FillStyle, value_list.fill_style.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::FillRule, value_list.fill_rule.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::Tile, value_list.tile.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::Stipple, value_list.stipple.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::TileStippleOriginX, value_list.tile_stipple_x_origin.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::TileStippleOriginY, value_list.tile_stipple_y_origin.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::Font, value_list.font.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::SubwindowMode, value_list.subwindow_mode.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::GraphicsExposures, value_list.graphics_exposures.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::ClipOriginX, value_list.clip_x_origin.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::ClipOriginY, value_list.clip_y_origin.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::ClipMask, value_list.clip_mask.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::DashOffset, value_list.dash_offset.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::DashList, value_list.dashes.has_value(), true, &value_mask);
  SwitchVar(GraphicsContextAttribute::ArcMode, value_list.arc_mode.has_value(), true, &value_mask);
  uint32_t tmp83;
  tmp83 = static_cast<uint32_t>(value_mask);
  Write(&tmp83, &buf);

  // value_list
  auto value_list_expr = value_mask;
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::Function)) {
    auto& function = *value_list.function;

    // function
    uint32_t tmp84;
    tmp84 = static_cast<uint32_t>(function);
    Write(&tmp84, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::PlaneMask)) {
    auto& plane_mask = *value_list.plane_mask;

    // plane_mask
    Write(&plane_mask, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::Foreground)) {
    auto& foreground = *value_list.foreground;

    // foreground
    Write(&foreground, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::Background)) {
    auto& background = *value_list.background;

    // background
    Write(&background, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::LineWidth)) {
    auto& line_width = *value_list.line_width;

    // line_width
    Write(&line_width, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::LineStyle)) {
    auto& line_style = *value_list.line_style;

    // line_style
    uint32_t tmp85;
    tmp85 = static_cast<uint32_t>(line_style);
    Write(&tmp85, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::CapStyle)) {
    auto& cap_style = *value_list.cap_style;

    // cap_style
    uint32_t tmp86;
    tmp86 = static_cast<uint32_t>(cap_style);
    Write(&tmp86, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::JoinStyle)) {
    auto& join_style = *value_list.join_style;

    // join_style
    uint32_t tmp87;
    tmp87 = static_cast<uint32_t>(join_style);
    Write(&tmp87, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::FillStyle)) {
    auto& fill_style = *value_list.fill_style;

    // fill_style
    uint32_t tmp88;
    tmp88 = static_cast<uint32_t>(fill_style);
    Write(&tmp88, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::FillRule)) {
    auto& fill_rule = *value_list.fill_rule;

    // fill_rule
    uint32_t tmp89;
    tmp89 = static_cast<uint32_t>(fill_rule);
    Write(&tmp89, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::Tile)) {
    auto& tile = *value_list.tile;

    // tile
    Write(&tile, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::Stipple)) {
    auto& stipple = *value_list.stipple;

    // stipple
    Write(&stipple, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::TileStippleOriginX)) {
    auto& tile_stipple_x_origin = *value_list.tile_stipple_x_origin;

    // tile_stipple_x_origin
    Write(&tile_stipple_x_origin, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::TileStippleOriginY)) {
    auto& tile_stipple_y_origin = *value_list.tile_stipple_y_origin;

    // tile_stipple_y_origin
    Write(&tile_stipple_y_origin, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::Font)) {
    auto& font = *value_list.font;

    // font
    Write(&font, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::SubwindowMode)) {
    auto& subwindow_mode = *value_list.subwindow_mode;

    // subwindow_mode
    uint32_t tmp90;
    tmp90 = static_cast<uint32_t>(subwindow_mode);
    Write(&tmp90, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::GraphicsExposures)) {
    auto& graphics_exposures = *value_list.graphics_exposures;

    // graphics_exposures
    Write(&graphics_exposures, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::ClipOriginX)) {
    auto& clip_x_origin = *value_list.clip_x_origin;

    // clip_x_origin
    Write(&clip_x_origin, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::ClipOriginY)) {
    auto& clip_y_origin = *value_list.clip_y_origin;

    // clip_y_origin
    Write(&clip_y_origin, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::ClipMask)) {
    auto& clip_mask = *value_list.clip_mask;

    // clip_mask
    Write(&clip_mask, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::DashOffset)) {
    auto& dash_offset = *value_list.dash_offset;

    // dash_offset
    Write(&dash_offset, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::DashList)) {
    auto& dashes = *value_list.dashes;

    // dashes
    Write(&dashes, &buf);

  }
  if (CaseAnd(value_list_expr, GraphicsContextAttribute::ArcMode)) {
    auto& arc_mode = *value_list.arc_mode;

    // arc_mode
    uint32_t tmp91;
    tmp91 = static_cast<uint32_t>(arc_mode);
    Write(&tmp91, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::CopyGC(
    const CopyGCRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& src_gc = request.src_gc;
  auto& dst_gc = request.dst_gc;
  auto& value_mask = request.value_mask;

  // major_opcode
  uint8_t major_opcode = 57;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // src_gc
  Write(&src_gc, &buf);

  // dst_gc
  Write(&dst_gc, &buf);

  // value_mask
  uint32_t tmp92;
  tmp92 = static_cast<uint32_t>(value_mask);
  Write(&tmp92, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::SetDashes(
    const SetDashesRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& gc = request.gc;
  auto& dash_offset = request.dash_offset;
  uint16_t dashes_len{};
  auto& dashes = request.dashes;

  // major_opcode
  uint8_t major_opcode = 58;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // gc
  Write(&gc, &buf);

  // dash_offset
  Write(&dash_offset, &buf);

  // dashes_len
  dashes_len = dashes.size();
  Write(&dashes_len, &buf);

  // dashes
  DCHECK_EQ(static_cast<size_t>(dashes_len), dashes.size());
  for (auto& dashes_elem : dashes) {
    // dashes_elem
    Write(&dashes_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::SetClipRectangles(
    const SetClipRectanglesRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& ordering = request.ordering;
  auto& gc = request.gc;
  auto& clip_x_origin = request.clip_x_origin;
  auto& clip_y_origin = request.clip_y_origin;
  auto& rectangles = request.rectangles;
  size_t rectangles_len = rectangles.size();

  // major_opcode
  uint8_t major_opcode = 59;
  Write(&major_opcode, &buf);

  // ordering
  uint8_t tmp93;
  tmp93 = static_cast<uint8_t>(ordering);
  Write(&tmp93, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // gc
  Write(&gc, &buf);

  // clip_x_origin
  Write(&clip_x_origin, &buf);

  // clip_y_origin
  Write(&clip_y_origin, &buf);

  // rectangles
  DCHECK_EQ(static_cast<size_t>(rectangles_len), rectangles.size());
  for (auto& rectangles_elem : rectangles) {
    // rectangles_elem
    {
      auto& x = rectangles_elem.x;
      auto& y = rectangles_elem.y;
      auto& width = rectangles_elem.width;
      auto& height = rectangles_elem.height;

      // x
      Write(&x, &buf);

      // y
      Write(&y, &buf);

      // width
      Write(&width, &buf);

      // height
      Write(&height, &buf);

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::FreeGC(
    const FreeGCRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& gc = request.gc;

  // major_opcode
  uint8_t major_opcode = 60;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // gc
  Write(&gc, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::ClearArea(
    const ClearAreaRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& exposures = request.exposures;
  auto& window = request.window;
  auto& x = request.x;
  auto& y = request.y;
  auto& width = request.width;
  auto& height = request.height;

  // major_opcode
  uint8_t major_opcode = 61;
  Write(&major_opcode, &buf);

  // exposures
  Write(&exposures, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  // x
  Write(&x, &buf);

  // y
  Write(&y, &buf);

  // width
  Write(&width, &buf);

  // height
  Write(&height, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::CopyArea(
    const CopyAreaRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& src_drawable = request.src_drawable;
  auto& dst_drawable = request.dst_drawable;
  auto& gc = request.gc;
  auto& src_x = request.src_x;
  auto& src_y = request.src_y;
  auto& dst_x = request.dst_x;
  auto& dst_y = request.dst_y;
  auto& width = request.width;
  auto& height = request.height;

  // major_opcode
  uint8_t major_opcode = 62;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // src_drawable
  Write(&src_drawable, &buf);

  // dst_drawable
  Write(&dst_drawable, &buf);

  // gc
  Write(&gc, &buf);

  // src_x
  Write(&src_x, &buf);

  // src_y
  Write(&src_y, &buf);

  // dst_x
  Write(&dst_x, &buf);

  // dst_y
  Write(&dst_y, &buf);

  // width
  Write(&width, &buf);

  // height
  Write(&height, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::CopyPlane(
    const CopyPlaneRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& src_drawable = request.src_drawable;
  auto& dst_drawable = request.dst_drawable;
  auto& gc = request.gc;
  auto& src_x = request.src_x;
  auto& src_y = request.src_y;
  auto& dst_x = request.dst_x;
  auto& dst_y = request.dst_y;
  auto& width = request.width;
  auto& height = request.height;
  auto& bit_plane = request.bit_plane;

  // major_opcode
  uint8_t major_opcode = 63;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // src_drawable
  Write(&src_drawable, &buf);

  // dst_drawable
  Write(&dst_drawable, &buf);

  // gc
  Write(&gc, &buf);

  // src_x
  Write(&src_x, &buf);

  // src_y
  Write(&src_y, &buf);

  // dst_x
  Write(&dst_x, &buf);

  // dst_y
  Write(&dst_y, &buf);

  // width
  Write(&width, &buf);

  // height
  Write(&height, &buf);

  // bit_plane
  Write(&bit_plane, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::PolyPoint(
    const PolyPointRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& coordinate_mode = request.coordinate_mode;
  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& points = request.points;
  size_t points_len = points.size();

  // major_opcode
  uint8_t major_opcode = 64;
  Write(&major_opcode, &buf);

  // coordinate_mode
  uint8_t tmp94;
  tmp94 = static_cast<uint8_t>(coordinate_mode);
  Write(&tmp94, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // points
  DCHECK_EQ(static_cast<size_t>(points_len), points.size());
  for (auto& points_elem : points) {
    // points_elem
    {
      auto& x = points_elem.x;
      auto& y = points_elem.y;

      // x
      Write(&x, &buf);

      // y
      Write(&y, &buf);

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::PolyLine(
    const PolyLineRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& coordinate_mode = request.coordinate_mode;
  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& points = request.points;
  size_t points_len = points.size();

  // major_opcode
  uint8_t major_opcode = 65;
  Write(&major_opcode, &buf);

  // coordinate_mode
  uint8_t tmp95;
  tmp95 = static_cast<uint8_t>(coordinate_mode);
  Write(&tmp95, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // points
  DCHECK_EQ(static_cast<size_t>(points_len), points.size());
  for (auto& points_elem : points) {
    // points_elem
    {
      auto& x = points_elem.x;
      auto& y = points_elem.y;

      // x
      Write(&x, &buf);

      // y
      Write(&y, &buf);

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::PolySegment(
    const PolySegmentRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& segments = request.segments;
  size_t segments_len = segments.size();

  // major_opcode
  uint8_t major_opcode = 66;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // segments
  DCHECK_EQ(static_cast<size_t>(segments_len), segments.size());
  for (auto& segments_elem : segments) {
    // segments_elem
    {
      auto& x1 = segments_elem.x1;
      auto& y1 = segments_elem.y1;
      auto& x2 = segments_elem.x2;
      auto& y2 = segments_elem.y2;

      // x1
      Write(&x1, &buf);

      // y1
      Write(&y1, &buf);

      // x2
      Write(&x2, &buf);

      // y2
      Write(&y2, &buf);

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::PolyRectangle(
    const PolyRectangleRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& rectangles = request.rectangles;
  size_t rectangles_len = rectangles.size();

  // major_opcode
  uint8_t major_opcode = 67;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // rectangles
  DCHECK_EQ(static_cast<size_t>(rectangles_len), rectangles.size());
  for (auto& rectangles_elem : rectangles) {
    // rectangles_elem
    {
      auto& x = rectangles_elem.x;
      auto& y = rectangles_elem.y;
      auto& width = rectangles_elem.width;
      auto& height = rectangles_elem.height;

      // x
      Write(&x, &buf);

      // y
      Write(&y, &buf);

      // width
      Write(&width, &buf);

      // height
      Write(&height, &buf);

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::PolyArc(
    const PolyArcRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& arcs = request.arcs;
  size_t arcs_len = arcs.size();

  // major_opcode
  uint8_t major_opcode = 68;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // arcs
  DCHECK_EQ(static_cast<size_t>(arcs_len), arcs.size());
  for (auto& arcs_elem : arcs) {
    // arcs_elem
    {
      auto& x = arcs_elem.x;
      auto& y = arcs_elem.y;
      auto& width = arcs_elem.width;
      auto& height = arcs_elem.height;
      auto& angle1 = arcs_elem.angle1;
      auto& angle2 = arcs_elem.angle2;

      // x
      Write(&x, &buf);

      // y
      Write(&y, &buf);

      // width
      Write(&width, &buf);

      // height
      Write(&height, &buf);

      // angle1
      Write(&angle1, &buf);

      // angle2
      Write(&angle2, &buf);

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::FillPoly(
    const FillPolyRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& shape = request.shape;
  auto& coordinate_mode = request.coordinate_mode;
  auto& points = request.points;
  size_t points_len = points.size();

  // major_opcode
  uint8_t major_opcode = 69;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // shape
  uint8_t tmp96;
  tmp96 = static_cast<uint8_t>(shape);
  Write(&tmp96, &buf);

  // coordinate_mode
  uint8_t tmp97;
  tmp97 = static_cast<uint8_t>(coordinate_mode);
  Write(&tmp97, &buf);

  // pad1
  Pad(&buf, 2);

  // points
  DCHECK_EQ(static_cast<size_t>(points_len), points.size());
  for (auto& points_elem : points) {
    // points_elem
    {
      auto& x = points_elem.x;
      auto& y = points_elem.y;

      // x
      Write(&x, &buf);

      // y
      Write(&y, &buf);

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::PolyFillRectangle(
    const PolyFillRectangleRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& rectangles = request.rectangles;
  size_t rectangles_len = rectangles.size();

  // major_opcode
  uint8_t major_opcode = 70;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // rectangles
  DCHECK_EQ(static_cast<size_t>(rectangles_len), rectangles.size());
  for (auto& rectangles_elem : rectangles) {
    // rectangles_elem
    {
      auto& x = rectangles_elem.x;
      auto& y = rectangles_elem.y;
      auto& width = rectangles_elem.width;
      auto& height = rectangles_elem.height;

      // x
      Write(&x, &buf);

      // y
      Write(&y, &buf);

      // width
      Write(&width, &buf);

      // height
      Write(&height, &buf);

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::PolyFillArc(
    const PolyFillArcRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& arcs = request.arcs;
  size_t arcs_len = arcs.size();

  // major_opcode
  uint8_t major_opcode = 71;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // arcs
  DCHECK_EQ(static_cast<size_t>(arcs_len), arcs.size());
  for (auto& arcs_elem : arcs) {
    // arcs_elem
    {
      auto& x = arcs_elem.x;
      auto& y = arcs_elem.y;
      auto& width = arcs_elem.width;
      auto& height = arcs_elem.height;
      auto& angle1 = arcs_elem.angle1;
      auto& angle2 = arcs_elem.angle2;

      // x
      Write(&x, &buf);

      // y
      Write(&y, &buf);

      // width
      Write(&width, &buf);

      // height
      Write(&height, &buf);

      // angle1
      Write(&angle1, &buf);

      // angle2
      Write(&angle2, &buf);

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::PutImage(
    const PutImageRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& format = request.format;
  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& width = request.width;
  auto& height = request.height;
  auto& dst_x = request.dst_x;
  auto& dst_y = request.dst_y;
  auto& left_pad = request.left_pad;
  auto& depth = request.depth;
  auto& data = request.data;
  size_t data_len = data.size();

  // major_opcode
  uint8_t major_opcode = 72;
  Write(&major_opcode, &buf);

  // format
  uint8_t tmp98;
  tmp98 = static_cast<uint8_t>(format);
  Write(&tmp98, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // width
  Write(&width, &buf);

  // height
  Write(&height, &buf);

  // dst_x
  Write(&dst_x, &buf);

  // dst_y
  Write(&dst_y, &buf);

  // left_pad
  Write(&left_pad, &buf);

  // depth
  Write(&depth, &buf);

  // pad0
  Pad(&buf, 2);

  // data
  DCHECK_EQ(static_cast<size_t>(data_len), data.size());
  for (auto& data_elem : data) {
    // data_elem
    Write(&data_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<GetImageReply>
XProto::GetImage(
    const GetImageRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& format = request.format;
  auto& drawable = request.drawable;
  auto& x = request.x;
  auto& y = request.y;
  auto& width = request.width;
  auto& height = request.height;
  auto& plane_mask = request.plane_mask;

  // major_opcode
  uint8_t major_opcode = 73;
  Write(&major_opcode, &buf);

  // format
  uint8_t tmp99;
  tmp99 = static_cast<uint8_t>(format);
  Write(&tmp99, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // x
  Write(&x, &buf);

  // y
  Write(&y, &buf);

  // width
  Write(&width, &buf);

  // height
  Write(&height, &buf);

  // plane_mask
  Write(&plane_mask, &buf);

  Align(&buf, 4);

  return x11::SendRequest<GetImageReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<GetImageReply>
detail::ReadReply<GetImageReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<GetImageReply>();

  auto& depth = (*reply).depth;
  auto& sequence = (*reply).sequence;
  auto& visual = (*reply).visual;
  auto& data = (*reply).data;
  size_t data_len = data.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // depth
  Read(&depth, &buf);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // visual
  Read(&visual, &buf);

  // pad0
  Pad(&buf, 20);

  // data
  data.resize((length) * (4));
  for (auto& data_elem : data) {
    // data_elem
    Read(&data_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::PolyText8(
    const PolyText8Request& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& x = request.x;
  auto& y = request.y;
  auto& items = request.items;
  size_t items_len = items.size();

  // major_opcode
  uint8_t major_opcode = 74;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // x
  Write(&x, &buf);

  // y
  Write(&y, &buf);

  // items
  DCHECK_EQ(static_cast<size_t>(items_len), items.size());
  for (auto& items_elem : items) {
    // items_elem
    Write(&items_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::PolyText16(
    const PolyText16Request& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& x = request.x;
  auto& y = request.y;
  auto& items = request.items;
  size_t items_len = items.size();

  // major_opcode
  uint8_t major_opcode = 75;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // x
  Write(&x, &buf);

  // y
  Write(&y, &buf);

  // items
  DCHECK_EQ(static_cast<size_t>(items_len), items.size());
  for (auto& items_elem : items) {
    // items_elem
    Write(&items_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::ImageText8(
    const ImageText8Request& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  uint8_t string_len{};
  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& x = request.x;
  auto& y = request.y;
  auto& string = request.string;

  // major_opcode
  uint8_t major_opcode = 76;
  Write(&major_opcode, &buf);

  // string_len
  string_len = string.size();
  Write(&string_len, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // x
  Write(&x, &buf);

  // y
  Write(&y, &buf);

  // string
  DCHECK_EQ(static_cast<size_t>(string_len), string.size());
  for (auto& string_elem : string) {
    // string_elem
    Write(&string_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::ImageText16(
    const ImageText16Request& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  uint8_t string_len{};
  auto& drawable = request.drawable;
  auto& gc = request.gc;
  auto& x = request.x;
  auto& y = request.y;
  auto& string = request.string;

  // major_opcode
  uint8_t major_opcode = 77;
  Write(&major_opcode, &buf);

  // string_len
  string_len = string.size();
  Write(&string_len, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // gc
  Write(&gc, &buf);

  // x
  Write(&x, &buf);

  // y
  Write(&y, &buf);

  // string
  DCHECK_EQ(static_cast<size_t>(string_len), string.size());
  for (auto& string_elem : string) {
    // string_elem
    {
      auto& byte1 = string_elem.byte1;
      auto& byte2 = string_elem.byte2;

      // byte1
      Write(&byte1, &buf);

      // byte2
      Write(&byte2, &buf);

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::CreateColormap(
    const CreateColormapRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& alloc = request.alloc;
  auto& mid = request.mid;
  auto& window = request.window;
  auto& visual = request.visual;

  // major_opcode
  uint8_t major_opcode = 78;
  Write(&major_opcode, &buf);

  // alloc
  uint8_t tmp100;
  tmp100 = static_cast<uint8_t>(alloc);
  Write(&tmp100, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // mid
  Write(&mid, &buf);

  // window
  Write(&window, &buf);

  // visual
  Write(&visual, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::FreeColormap(
    const FreeColormapRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& cmap = request.cmap;

  // major_opcode
  uint8_t major_opcode = 79;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::CopyColormapAndFree(
    const CopyColormapAndFreeRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& mid = request.mid;
  auto& src_cmap = request.src_cmap;

  // major_opcode
  uint8_t major_opcode = 80;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // mid
  Write(&mid, &buf);

  // src_cmap
  Write(&src_cmap, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::InstallColormap(
    const InstallColormapRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& cmap = request.cmap;

  // major_opcode
  uint8_t major_opcode = 81;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::UninstallColormap(
    const UninstallColormapRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& cmap = request.cmap;

  // major_opcode
  uint8_t major_opcode = 82;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<ListInstalledColormapsReply>
XProto::ListInstalledColormaps(
    const ListInstalledColormapsRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& window = request.window;

  // major_opcode
  uint8_t major_opcode = 83;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  Align(&buf, 4);

  return x11::SendRequest<ListInstalledColormapsReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<ListInstalledColormapsReply>
detail::ReadReply<ListInstalledColormapsReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<ListInstalledColormapsReply>();

  auto& sequence = (*reply).sequence;
  uint16_t cmaps_len{};
  auto& cmaps = (*reply).cmaps;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // cmaps_len
  Read(&cmaps_len, &buf);

  // pad1
  Pad(&buf, 22);

  // cmaps
  cmaps.resize(cmaps_len);
  for (auto& cmaps_elem : cmaps) {
    // cmaps_elem
    Read(&cmaps_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<AllocColorReply>
XProto::AllocColor(
    const AllocColorRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& cmap = request.cmap;
  auto& red = request.red;
  auto& green = request.green;
  auto& blue = request.blue;

  // major_opcode
  uint8_t major_opcode = 84;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  // red
  Write(&red, &buf);

  // green
  Write(&green, &buf);

  // blue
  Write(&blue, &buf);

  // pad1
  Pad(&buf, 2);

  Align(&buf, 4);

  return x11::SendRequest<AllocColorReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<AllocColorReply>
detail::ReadReply<AllocColorReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<AllocColorReply>();

  auto& sequence = (*reply).sequence;
  auto& red = (*reply).red;
  auto& green = (*reply).green;
  auto& blue = (*reply).blue;
  auto& pixel = (*reply).pixel;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // red
  Read(&red, &buf);

  // green
  Read(&green, &buf);

  // blue
  Read(&blue, &buf);

  // pad1
  Pad(&buf, 2);

  // pixel
  Read(&pixel, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<AllocNamedColorReply>
XProto::AllocNamedColor(
    const AllocNamedColorRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& cmap = request.cmap;
  uint16_t name_len{};
  auto& name = request.name;

  // major_opcode
  uint8_t major_opcode = 85;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  // name_len
  name_len = name.size();
  Write(&name_len, &buf);

  // pad1
  Pad(&buf, 2);

  // name
  DCHECK_EQ(static_cast<size_t>(name_len), name.size());
  for (auto& name_elem : name) {
    // name_elem
    Write(&name_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<AllocNamedColorReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<AllocNamedColorReply>
detail::ReadReply<AllocNamedColorReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<AllocNamedColorReply>();

  auto& sequence = (*reply).sequence;
  auto& pixel = (*reply).pixel;
  auto& exact_red = (*reply).exact_red;
  auto& exact_green = (*reply).exact_green;
  auto& exact_blue = (*reply).exact_blue;
  auto& visual_red = (*reply).visual_red;
  auto& visual_green = (*reply).visual_green;
  auto& visual_blue = (*reply).visual_blue;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pixel
  Read(&pixel, &buf);

  // exact_red
  Read(&exact_red, &buf);

  // exact_green
  Read(&exact_green, &buf);

  // exact_blue
  Read(&exact_blue, &buf);

  // visual_red
  Read(&visual_red, &buf);

  // visual_green
  Read(&visual_green, &buf);

  // visual_blue
  Read(&visual_blue, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<AllocColorCellsReply>
XProto::AllocColorCells(
    const AllocColorCellsRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& contiguous = request.contiguous;
  auto& cmap = request.cmap;
  auto& colors = request.colors;
  auto& planes = request.planes;

  // major_opcode
  uint8_t major_opcode = 86;
  Write(&major_opcode, &buf);

  // contiguous
  Write(&contiguous, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  // colors
  Write(&colors, &buf);

  // planes
  Write(&planes, &buf);

  Align(&buf, 4);

  return x11::SendRequest<AllocColorCellsReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<AllocColorCellsReply>
detail::ReadReply<AllocColorCellsReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<AllocColorCellsReply>();

  auto& sequence = (*reply).sequence;
  uint16_t pixels_len{};
  uint16_t masks_len{};
  auto& pixels = (*reply).pixels;
  auto& masks = (*reply).masks;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pixels_len
  Read(&pixels_len, &buf);

  // masks_len
  Read(&masks_len, &buf);

  // pad1
  Pad(&buf, 20);

  // pixels
  pixels.resize(pixels_len);
  for (auto& pixels_elem : pixels) {
    // pixels_elem
    Read(&pixels_elem, &buf);

  }

  // masks
  masks.resize(masks_len);
  for (auto& masks_elem : masks) {
    // masks_elem
    Read(&masks_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<AllocColorPlanesReply>
XProto::AllocColorPlanes(
    const AllocColorPlanesRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& contiguous = request.contiguous;
  auto& cmap = request.cmap;
  auto& colors = request.colors;
  auto& reds = request.reds;
  auto& greens = request.greens;
  auto& blues = request.blues;

  // major_opcode
  uint8_t major_opcode = 87;
  Write(&major_opcode, &buf);

  // contiguous
  Write(&contiguous, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  // colors
  Write(&colors, &buf);

  // reds
  Write(&reds, &buf);

  // greens
  Write(&greens, &buf);

  // blues
  Write(&blues, &buf);

  Align(&buf, 4);

  return x11::SendRequest<AllocColorPlanesReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<AllocColorPlanesReply>
detail::ReadReply<AllocColorPlanesReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<AllocColorPlanesReply>();

  auto& sequence = (*reply).sequence;
  uint16_t pixels_len{};
  auto& red_mask = (*reply).red_mask;
  auto& green_mask = (*reply).green_mask;
  auto& blue_mask = (*reply).blue_mask;
  auto& pixels = (*reply).pixels;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pixels_len
  Read(&pixels_len, &buf);

  // pad1
  Pad(&buf, 2);

  // red_mask
  Read(&red_mask, &buf);

  // green_mask
  Read(&green_mask, &buf);

  // blue_mask
  Read(&blue_mask, &buf);

  // pad2
  Pad(&buf, 8);

  // pixels
  pixels.resize(pixels_len);
  for (auto& pixels_elem : pixels) {
    // pixels_elem
    Read(&pixels_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::FreeColors(
    const FreeColorsRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& cmap = request.cmap;
  auto& plane_mask = request.plane_mask;
  auto& pixels = request.pixels;
  size_t pixels_len = pixels.size();

  // major_opcode
  uint8_t major_opcode = 88;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  // plane_mask
  Write(&plane_mask, &buf);

  // pixels
  DCHECK_EQ(static_cast<size_t>(pixels_len), pixels.size());
  for (auto& pixels_elem : pixels) {
    // pixels_elem
    Write(&pixels_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::StoreColors(
    const StoreColorsRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& cmap = request.cmap;
  auto& items = request.items;
  size_t items_len = items.size();

  // major_opcode
  uint8_t major_opcode = 89;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  // items
  DCHECK_EQ(static_cast<size_t>(items_len), items.size());
  for (auto& items_elem : items) {
    // items_elem
    {
      auto& pixel = items_elem.pixel;
      auto& red = items_elem.red;
      auto& green = items_elem.green;
      auto& blue = items_elem.blue;
      auto& flags = items_elem.flags;

      // pixel
      Write(&pixel, &buf);

      // red
      Write(&red, &buf);

      // green
      Write(&green, &buf);

      // blue
      Write(&blue, &buf);

      // flags
      uint8_t tmp101;
      tmp101 = static_cast<uint8_t>(flags);
      Write(&tmp101, &buf);

      // pad0
      Pad(&buf, 1);

    }

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::StoreNamedColor(
    const StoreNamedColorRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& flags = request.flags;
  auto& cmap = request.cmap;
  auto& pixel = request.pixel;
  uint16_t name_len{};
  auto& name = request.name;

  // major_opcode
  uint8_t major_opcode = 90;
  Write(&major_opcode, &buf);

  // flags
  uint8_t tmp102;
  tmp102 = static_cast<uint8_t>(flags);
  Write(&tmp102, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  // pixel
  Write(&pixel, &buf);

  // name_len
  name_len = name.size();
  Write(&name_len, &buf);

  // pad0
  Pad(&buf, 2);

  // name
  DCHECK_EQ(static_cast<size_t>(name_len), name.size());
  for (auto& name_elem : name) {
    // name_elem
    Write(&name_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<QueryColorsReply>
XProto::QueryColors(
    const QueryColorsRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& cmap = request.cmap;
  auto& pixels = request.pixels;
  size_t pixels_len = pixels.size();

  // major_opcode
  uint8_t major_opcode = 91;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  // pixels
  DCHECK_EQ(static_cast<size_t>(pixels_len), pixels.size());
  for (auto& pixels_elem : pixels) {
    // pixels_elem
    Write(&pixels_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<QueryColorsReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<QueryColorsReply>
detail::ReadReply<QueryColorsReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<QueryColorsReply>();

  auto& sequence = (*reply).sequence;
  uint16_t colors_len{};
  auto& colors = (*reply).colors;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // colors_len
  Read(&colors_len, &buf);

  // pad1
  Pad(&buf, 22);

  // colors
  colors.resize(colors_len);
  for (auto& colors_elem : colors) {
    // colors_elem
    {
      auto& red = colors_elem.red;
      auto& green = colors_elem.green;
      auto& blue = colors_elem.blue;

      // red
      Read(&red, &buf);

      // green
      Read(&green, &buf);

      // blue
      Read(&blue, &buf);

      // pad0
      Pad(&buf, 2);

    }

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<LookupColorReply>
XProto::LookupColor(
    const LookupColorRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& cmap = request.cmap;
  uint16_t name_len{};
  auto& name = request.name;

  // major_opcode
  uint8_t major_opcode = 92;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cmap
  Write(&cmap, &buf);

  // name_len
  name_len = name.size();
  Write(&name_len, &buf);

  // pad1
  Pad(&buf, 2);

  // name
  DCHECK_EQ(static_cast<size_t>(name_len), name.size());
  for (auto& name_elem : name) {
    // name_elem
    Write(&name_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<LookupColorReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<LookupColorReply>
detail::ReadReply<LookupColorReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<LookupColorReply>();

  auto& sequence = (*reply).sequence;
  auto& exact_red = (*reply).exact_red;
  auto& exact_green = (*reply).exact_green;
  auto& exact_blue = (*reply).exact_blue;
  auto& visual_red = (*reply).visual_red;
  auto& visual_green = (*reply).visual_green;
  auto& visual_blue = (*reply).visual_blue;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // exact_red
  Read(&exact_red, &buf);

  // exact_green
  Read(&exact_green, &buf);

  // exact_blue
  Read(&exact_blue, &buf);

  // visual_red
  Read(&visual_red, &buf);

  // visual_green
  Read(&visual_green, &buf);

  // visual_blue
  Read(&visual_blue, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::CreateCursor(
    const CreateCursorRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& cid = request.cid;
  auto& source = request.source;
  auto& mask = request.mask;
  auto& fore_red = request.fore_red;
  auto& fore_green = request.fore_green;
  auto& fore_blue = request.fore_blue;
  auto& back_red = request.back_red;
  auto& back_green = request.back_green;
  auto& back_blue = request.back_blue;
  auto& x = request.x;
  auto& y = request.y;

  // major_opcode
  uint8_t major_opcode = 93;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cid
  Write(&cid, &buf);

  // source
  Write(&source, &buf);

  // mask
  Write(&mask, &buf);

  // fore_red
  Write(&fore_red, &buf);

  // fore_green
  Write(&fore_green, &buf);

  // fore_blue
  Write(&fore_blue, &buf);

  // back_red
  Write(&back_red, &buf);

  // back_green
  Write(&back_green, &buf);

  // back_blue
  Write(&back_blue, &buf);

  // x
  Write(&x, &buf);

  // y
  Write(&y, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::CreateGlyphCursor(
    const CreateGlyphCursorRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& cid = request.cid;
  auto& source_font = request.source_font;
  auto& mask_font = request.mask_font;
  auto& source_char = request.source_char;
  auto& mask_char = request.mask_char;
  auto& fore_red = request.fore_red;
  auto& fore_green = request.fore_green;
  auto& fore_blue = request.fore_blue;
  auto& back_red = request.back_red;
  auto& back_green = request.back_green;
  auto& back_blue = request.back_blue;

  // major_opcode
  uint8_t major_opcode = 94;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cid
  Write(&cid, &buf);

  // source_font
  Write(&source_font, &buf);

  // mask_font
  Write(&mask_font, &buf);

  // source_char
  Write(&source_char, &buf);

  // mask_char
  Write(&mask_char, &buf);

  // fore_red
  Write(&fore_red, &buf);

  // fore_green
  Write(&fore_green, &buf);

  // fore_blue
  Write(&fore_blue, &buf);

  // back_red
  Write(&back_red, &buf);

  // back_green
  Write(&back_green, &buf);

  // back_blue
  Write(&back_blue, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::FreeCursor(
    const FreeCursorRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& cursor = request.cursor;

  // major_opcode
  uint8_t major_opcode = 95;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cursor
  Write(&cursor, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::RecolorCursor(
    const RecolorCursorRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& cursor = request.cursor;
  auto& fore_red = request.fore_red;
  auto& fore_green = request.fore_green;
  auto& fore_blue = request.fore_blue;
  auto& back_red = request.back_red;
  auto& back_green = request.back_green;
  auto& back_blue = request.back_blue;

  // major_opcode
  uint8_t major_opcode = 96;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // cursor
  Write(&cursor, &buf);

  // fore_red
  Write(&fore_red, &buf);

  // fore_green
  Write(&fore_green, &buf);

  // fore_blue
  Write(&fore_blue, &buf);

  // back_red
  Write(&back_red, &buf);

  // back_green
  Write(&back_green, &buf);

  // back_blue
  Write(&back_blue, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<QueryBestSizeReply>
XProto::QueryBestSize(
    const QueryBestSizeRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& c_class = request.c_class;
  auto& drawable = request.drawable;
  auto& width = request.width;
  auto& height = request.height;

  // major_opcode
  uint8_t major_opcode = 97;
  Write(&major_opcode, &buf);

  // c_class
  uint8_t tmp103;
  tmp103 = static_cast<uint8_t>(c_class);
  Write(&tmp103, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // drawable
  Write(&drawable, &buf);

  // width
  Write(&width, &buf);

  // height
  Write(&height, &buf);

  Align(&buf, 4);

  return x11::SendRequest<QueryBestSizeReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<QueryBestSizeReply>
detail::ReadReply<QueryBestSizeReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<QueryBestSizeReply>();

  auto& sequence = (*reply).sequence;
  auto& width = (*reply).width;
  auto& height = (*reply).height;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // width
  Read(&width, &buf);

  // height
  Read(&height, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<QueryExtensionReply>
XProto::QueryExtension(
    const QueryExtensionRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  uint16_t name_len{};
  auto& name = request.name;

  // major_opcode
  uint8_t major_opcode = 98;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // name_len
  name_len = name.size();
  Write(&name_len, &buf);

  // pad1
  Pad(&buf, 2);

  // name
  DCHECK_EQ(static_cast<size_t>(name_len), name.size());
  for (auto& name_elem : name) {
    // name_elem
    Write(&name_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<QueryExtensionReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<QueryExtensionReply>
detail::ReadReply<QueryExtensionReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<QueryExtensionReply>();

  auto& sequence = (*reply).sequence;
  auto& present = (*reply).present;
  auto& major_opcode = (*reply).major_opcode;
  auto& first_event = (*reply).first_event;
  auto& first_error = (*reply).first_error;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // present
  Read(&present, &buf);

  // major_opcode
  Read(&major_opcode, &buf);

  // first_event
  Read(&first_event, &buf);

  // first_error
  Read(&first_error, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<ListExtensionsReply>
XProto::ListExtensions(
    const ListExtensionsRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 99;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<ListExtensionsReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<ListExtensionsReply>
detail::ReadReply<ListExtensionsReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<ListExtensionsReply>();

  uint8_t names_len{};
  auto& sequence = (*reply).sequence;
  auto& names = (*reply).names;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // names_len
  Read(&names_len, &buf);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad0
  Pad(&buf, 24);

  // names
  names.resize(names_len);
  for (auto& names_elem : names) {
    // names_elem
    {
      uint8_t name_len{};
      auto& name = names_elem.name;

      // name_len
      Read(&name_len, &buf);

      // name
      name.resize(name_len);
      for (auto& name_elem : name) {
        // name_elem
        Read(&name_elem, &buf);

      }

    }

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::ChangeKeyboardMapping(
    const ChangeKeyboardMappingRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& keycode_count = request.keycode_count;
  auto& first_keycode = request.first_keycode;
  auto& keysyms_per_keycode = request.keysyms_per_keycode;
  auto& keysyms = request.keysyms;
  size_t keysyms_len = keysyms.size();

  // major_opcode
  uint8_t major_opcode = 100;
  Write(&major_opcode, &buf);

  // keycode_count
  Write(&keycode_count, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // first_keycode
  Write(&first_keycode, &buf);

  // keysyms_per_keycode
  Write(&keysyms_per_keycode, &buf);

  // pad0
  Pad(&buf, 2);

  // keysyms
  DCHECK_EQ(static_cast<size_t>((keycode_count) * (keysyms_per_keycode)), keysyms.size());
  for (auto& keysyms_elem : keysyms) {
    // keysyms_elem
    Write(&keysyms_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<GetKeyboardMappingReply>
XProto::GetKeyboardMapping(
    const GetKeyboardMappingRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& first_keycode = request.first_keycode;
  auto& count = request.count;

  // major_opcode
  uint8_t major_opcode = 101;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // first_keycode
  Write(&first_keycode, &buf);

  // count
  Write(&count, &buf);

  Align(&buf, 4);

  return x11::SendRequest<GetKeyboardMappingReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<GetKeyboardMappingReply>
detail::ReadReply<GetKeyboardMappingReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<GetKeyboardMappingReply>();

  auto& keysyms_per_keycode = (*reply).keysyms_per_keycode;
  auto& sequence = (*reply).sequence;
  auto& keysyms = (*reply).keysyms;
  size_t keysyms_len = keysyms.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // keysyms_per_keycode
  Read(&keysyms_per_keycode, &buf);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad0
  Pad(&buf, 24);

  // keysyms
  keysyms.resize(length);
  for (auto& keysyms_elem : keysyms) {
    // keysyms_elem
    Read(&keysyms_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::ChangeKeyboardControl(
    const ChangeKeyboardControlRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  Keyboard value_mask{};
  auto& value_list = request;

  // major_opcode
  uint8_t major_opcode = 102;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // value_mask
  SwitchVar(Keyboard::KeyClickPercent, value_list.key_click_percent.has_value(), true, &value_mask);
  SwitchVar(Keyboard::BellPercent, value_list.bell_percent.has_value(), true, &value_mask);
  SwitchVar(Keyboard::BellPitch, value_list.bell_pitch.has_value(), true, &value_mask);
  SwitchVar(Keyboard::BellDuration, value_list.bell_duration.has_value(), true, &value_mask);
  SwitchVar(Keyboard::Led, value_list.led.has_value(), true, &value_mask);
  SwitchVar(Keyboard::LedMode, value_list.led_mode.has_value(), true, &value_mask);
  SwitchVar(Keyboard::Key, value_list.key.has_value(), true, &value_mask);
  SwitchVar(Keyboard::AutoRepeatMode, value_list.auto_repeat_mode.has_value(), true, &value_mask);
  uint32_t tmp104;
  tmp104 = static_cast<uint32_t>(value_mask);
  Write(&tmp104, &buf);

  // value_list
  auto value_list_expr = value_mask;
  if (CaseAnd(value_list_expr, Keyboard::KeyClickPercent)) {
    auto& key_click_percent = *value_list.key_click_percent;

    // key_click_percent
    Write(&key_click_percent, &buf);

  }
  if (CaseAnd(value_list_expr, Keyboard::BellPercent)) {
    auto& bell_percent = *value_list.bell_percent;

    // bell_percent
    Write(&bell_percent, &buf);

  }
  if (CaseAnd(value_list_expr, Keyboard::BellPitch)) {
    auto& bell_pitch = *value_list.bell_pitch;

    // bell_pitch
    Write(&bell_pitch, &buf);

  }
  if (CaseAnd(value_list_expr, Keyboard::BellDuration)) {
    auto& bell_duration = *value_list.bell_duration;

    // bell_duration
    Write(&bell_duration, &buf);

  }
  if (CaseAnd(value_list_expr, Keyboard::Led)) {
    auto& led = *value_list.led;

    // led
    Write(&led, &buf);

  }
  if (CaseAnd(value_list_expr, Keyboard::LedMode)) {
    auto& led_mode = *value_list.led_mode;

    // led_mode
    uint32_t tmp105;
    tmp105 = static_cast<uint32_t>(led_mode);
    Write(&tmp105, &buf);

  }
  if (CaseAnd(value_list_expr, Keyboard::Key)) {
    auto& key = *value_list.key;

    // key
    Write(&key, &buf);

  }
  if (CaseAnd(value_list_expr, Keyboard::AutoRepeatMode)) {
    auto& auto_repeat_mode = *value_list.auto_repeat_mode;

    // auto_repeat_mode
    uint32_t tmp106;
    tmp106 = static_cast<uint32_t>(auto_repeat_mode);
    Write(&tmp106, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<GetKeyboardControlReply>
XProto::GetKeyboardControl(
    const GetKeyboardControlRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 103;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<GetKeyboardControlReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<GetKeyboardControlReply>
detail::ReadReply<GetKeyboardControlReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<GetKeyboardControlReply>();

  auto& global_auto_repeat = (*reply).global_auto_repeat;
  auto& sequence = (*reply).sequence;
  auto& led_mask = (*reply).led_mask;
  auto& key_click_percent = (*reply).key_click_percent;
  auto& bell_percent = (*reply).bell_percent;
  auto& bell_pitch = (*reply).bell_pitch;
  auto& bell_duration = (*reply).bell_duration;
  auto& auto_repeats = (*reply).auto_repeats;
  size_t auto_repeats_len = auto_repeats.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // global_auto_repeat
  uint8_t tmp107;
  Read(&tmp107, &buf);
  global_auto_repeat = static_cast<AutoRepeatMode>(tmp107);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // led_mask
  Read(&led_mask, &buf);

  // key_click_percent
  Read(&key_click_percent, &buf);

  // bell_percent
  Read(&bell_percent, &buf);

  // bell_pitch
  Read(&bell_pitch, &buf);

  // bell_duration
  Read(&bell_duration, &buf);

  // pad0
  Pad(&buf, 2);

  // auto_repeats
  for (auto& auto_repeats_elem : auto_repeats) {
    // auto_repeats_elem
    Read(&auto_repeats_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::Bell(
    const BellRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& percent = request.percent;

  // major_opcode
  uint8_t major_opcode = 104;
  Write(&major_opcode, &buf);

  // percent
  Write(&percent, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::ChangePointerControl(
    const ChangePointerControlRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& acceleration_numerator = request.acceleration_numerator;
  auto& acceleration_denominator = request.acceleration_denominator;
  auto& threshold = request.threshold;
  auto& do_acceleration = request.do_acceleration;
  auto& do_threshold = request.do_threshold;

  // major_opcode
  uint8_t major_opcode = 105;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // acceleration_numerator
  Write(&acceleration_numerator, &buf);

  // acceleration_denominator
  Write(&acceleration_denominator, &buf);

  // threshold
  Write(&threshold, &buf);

  // do_acceleration
  Write(&do_acceleration, &buf);

  // do_threshold
  Write(&do_threshold, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<GetPointerControlReply>
XProto::GetPointerControl(
    const GetPointerControlRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 106;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<GetPointerControlReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<GetPointerControlReply>
detail::ReadReply<GetPointerControlReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<GetPointerControlReply>();

  auto& sequence = (*reply).sequence;
  auto& acceleration_numerator = (*reply).acceleration_numerator;
  auto& acceleration_denominator = (*reply).acceleration_denominator;
  auto& threshold = (*reply).threshold;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // acceleration_numerator
  Read(&acceleration_numerator, &buf);

  // acceleration_denominator
  Read(&acceleration_denominator, &buf);

  // threshold
  Read(&threshold, &buf);

  // pad1
  Pad(&buf, 18);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::SetScreenSaver(
    const SetScreenSaverRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& timeout = request.timeout;
  auto& interval = request.interval;
  auto& prefer_blanking = request.prefer_blanking;
  auto& allow_exposures = request.allow_exposures;

  // major_opcode
  uint8_t major_opcode = 107;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // timeout
  Write(&timeout, &buf);

  // interval
  Write(&interval, &buf);

  // prefer_blanking
  uint8_t tmp108;
  tmp108 = static_cast<uint8_t>(prefer_blanking);
  Write(&tmp108, &buf);

  // allow_exposures
  uint8_t tmp109;
  tmp109 = static_cast<uint8_t>(allow_exposures);
  Write(&tmp109, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<GetScreenSaverReply>
XProto::GetScreenSaver(
    const GetScreenSaverRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 108;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<GetScreenSaverReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<GetScreenSaverReply>
detail::ReadReply<GetScreenSaverReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<GetScreenSaverReply>();

  auto& sequence = (*reply).sequence;
  auto& timeout = (*reply).timeout;
  auto& interval = (*reply).interval;
  auto& prefer_blanking = (*reply).prefer_blanking;
  auto& allow_exposures = (*reply).allow_exposures;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // pad0
  Pad(&buf, 1);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // timeout
  Read(&timeout, &buf);

  // interval
  Read(&interval, &buf);

  // prefer_blanking
  uint8_t tmp110;
  Read(&tmp110, &buf);
  prefer_blanking = static_cast<Blanking>(tmp110);

  // allow_exposures
  uint8_t tmp111;
  Read(&tmp111, &buf);
  allow_exposures = static_cast<Exposures>(tmp111);

  // pad1
  Pad(&buf, 18);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::ChangeHosts(
    const ChangeHostsRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& mode = request.mode;
  auto& family = request.family;
  uint16_t address_len{};
  auto& address = request.address;

  // major_opcode
  uint8_t major_opcode = 109;
  Write(&major_opcode, &buf);

  // mode
  uint8_t tmp112;
  tmp112 = static_cast<uint8_t>(mode);
  Write(&tmp112, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // family
  uint8_t tmp113;
  tmp113 = static_cast<uint8_t>(family);
  Write(&tmp113, &buf);

  // pad0
  Pad(&buf, 1);

  // address_len
  address_len = address.size();
  Write(&address_len, &buf);

  // address
  DCHECK_EQ(static_cast<size_t>(address_len), address.size());
  for (auto& address_elem : address) {
    // address_elem
    Write(&address_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<ListHostsReply>
XProto::ListHosts(
    const ListHostsRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 110;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<ListHostsReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<ListHostsReply>
detail::ReadReply<ListHostsReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<ListHostsReply>();

  auto& mode = (*reply).mode;
  auto& sequence = (*reply).sequence;
  uint16_t hosts_len{};
  auto& hosts = (*reply).hosts;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // mode
  uint8_t tmp114;
  Read(&tmp114, &buf);
  mode = static_cast<AccessControl>(tmp114);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // hosts_len
  Read(&hosts_len, &buf);

  // pad0
  Pad(&buf, 22);

  // hosts
  hosts.resize(hosts_len);
  for (auto& hosts_elem : hosts) {
    // hosts_elem
    {
      auto& family = hosts_elem.family;
      uint16_t address_len{};
      auto& address = hosts_elem.address;

      // family
      uint8_t tmp115;
      Read(&tmp115, &buf);
      family = static_cast<Family>(tmp115);

      // pad0
      Pad(&buf, 1);

      // address_len
      Read(&address_len, &buf);

      // address
      address.resize(address_len);
      for (auto& address_elem : address) {
        // address_elem
        Read(&address_elem, &buf);

      }

      // pad1
      Align(&buf, 4);

    }

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::SetAccessControl(
    const SetAccessControlRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& mode = request.mode;

  // major_opcode
  uint8_t major_opcode = 111;
  Write(&major_opcode, &buf);

  // mode
  uint8_t tmp116;
  tmp116 = static_cast<uint8_t>(mode);
  Write(&tmp116, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::SetCloseDownMode(
    const SetCloseDownModeRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& mode = request.mode;

  // major_opcode
  uint8_t major_opcode = 112;
  Write(&major_opcode, &buf);

  // mode
  uint8_t tmp117;
  tmp117 = static_cast<uint8_t>(mode);
  Write(&tmp117, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::KillClient(
    const KillClientRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& resource = request.resource;

  // major_opcode
  uint8_t major_opcode = 113;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // resource
  Write(&resource, &buf);

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::RotateProperties(
    const RotatePropertiesRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& window = request.window;
  uint16_t atoms_len{};
  auto& delta = request.delta;
  auto& atoms = request.atoms;

  // major_opcode
  uint8_t major_opcode = 114;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // window
  Write(&window, &buf);

  // atoms_len
  atoms_len = atoms.size();
  Write(&atoms_len, &buf);

  // delta
  Write(&delta, &buf);

  // atoms
  DCHECK_EQ(static_cast<size_t>(atoms_len), atoms.size());
  for (auto& atoms_elem : atoms) {
    // atoms_elem
    Write(&atoms_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<void>
XProto::ForceScreenSaver(
    const ForceScreenSaverRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& mode = request.mode;

  // major_opcode
  uint8_t major_opcode = 115;
  Write(&major_opcode, &buf);

  // mode
  uint8_t tmp118;
  tmp118 = static_cast<uint8_t>(mode);
  Write(&tmp118, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

Future<SetPointerMappingReply>
XProto::SetPointerMapping(
    const SetPointerMappingRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  uint8_t map_len{};
  auto& map = request.map;

  // major_opcode
  uint8_t major_opcode = 116;
  Write(&major_opcode, &buf);

  // map_len
  map_len = map.size();
  Write(&map_len, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // map
  DCHECK_EQ(static_cast<size_t>(map_len), map.size());
  for (auto& map_elem : map) {
    // map_elem
    Write(&map_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<SetPointerMappingReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<SetPointerMappingReply>
detail::ReadReply<SetPointerMappingReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<SetPointerMappingReply>();

  auto& status = (*reply).status;
  auto& sequence = (*reply).sequence;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // status
  uint8_t tmp119;
  Read(&tmp119, &buf);
  status = static_cast<MappingStatus>(tmp119);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<GetPointerMappingReply>
XProto::GetPointerMapping(
    const GetPointerMappingRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 117;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<GetPointerMappingReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<GetPointerMappingReply>
detail::ReadReply<GetPointerMappingReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<GetPointerMappingReply>();

  uint8_t map_len{};
  auto& sequence = (*reply).sequence;
  auto& map = (*reply).map;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // map_len
  Read(&map_len, &buf);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad0
  Pad(&buf, 24);

  // map
  map.resize(map_len);
  for (auto& map_elem : map) {
    // map_elem
    Read(&map_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<SetModifierMappingReply>
XProto::SetModifierMapping(
    const SetModifierMappingRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  auto& keycodes_per_modifier = request.keycodes_per_modifier;
  auto& keycodes = request.keycodes;
  size_t keycodes_len = keycodes.size();

  // major_opcode
  uint8_t major_opcode = 118;
  Write(&major_opcode, &buf);

  // keycodes_per_modifier
  Write(&keycodes_per_modifier, &buf);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  // keycodes
  DCHECK_EQ(static_cast<size_t>((keycodes_per_modifier) * (8)), keycodes.size());
  for (auto& keycodes_elem : keycodes) {
    // keycodes_elem
    Write(&keycodes_elem, &buf);

  }

  Align(&buf, 4);

  return x11::SendRequest<SetModifierMappingReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<SetModifierMappingReply>
detail::ReadReply<SetModifierMappingReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<SetModifierMappingReply>();

  auto& status = (*reply).status;
  auto& sequence = (*reply).sequence;

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // status
  uint8_t tmp120;
  Read(&tmp120, &buf);
  status = static_cast<MappingStatus>(tmp120);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<GetModifierMappingReply>
XProto::GetModifierMapping(
    const GetModifierMappingRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 119;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<GetModifierMappingReply>(connection_, &buf);
}

template<> COMPONENT_EXPORT(X11)
std::unique_ptr<GetModifierMappingReply>
detail::ReadReply<GetModifierMappingReply>(const uint8_t* buffer) {
  ReadBuffer buf{buffer, 0UL};
  auto reply = std::make_unique<GetModifierMappingReply>();

  auto& keycodes_per_modifier = (*reply).keycodes_per_modifier;
  auto& sequence = (*reply).sequence;
  auto& keycodes = (*reply).keycodes;
  size_t keycodes_len = keycodes.size();

  // response_type
  uint8_t response_type;
  Read(&response_type, &buf);

  // keycodes_per_modifier
  Read(&keycodes_per_modifier, &buf);

  // sequence
  Read(&sequence, &buf);

  // length
  uint32_t length;
  Read(&length, &buf);

  // pad0
  Pad(&buf, 24);

  // keycodes
  keycodes.resize((keycodes_per_modifier) * (8));
  for (auto& keycodes_elem : keycodes) {
    // keycodes_elem
    Read(&keycodes_elem, &buf);

  }

  Align(&buf, 4);
  DCHECK_EQ(buf.offset < 32 ? 0 : buf.offset - 32, 4 * length);

  return reply;
}

Future<void>
XProto::NoOperation(
    const NoOperationRequest& request) {
  if (!connection_->Ready())
    return {};

  WriteBuffer buf;

  // major_opcode
  uint8_t major_opcode = 127;
  Write(&major_opcode, &buf);

  // pad0
  Pad(&buf, 1);

  // length
  // Caller fills in length for writes.
  Pad(&buf, sizeof(uint16_t));

  Align(&buf, 4);

  return x11::SendRequest<void>(connection_, &buf);
}

}  // namespace x11
